﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Globalization;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Input;
using OfficeIVS;

namespace AssistToolDialog
{
    /// <summary>
    /// UserControl1.xaml の相互作用ロジック
    /// </summary>
    public partial class AssistControl : UserControl
    {
        private const float DEF_FONST_SIZE = 11;

        #region フィールド

        /// <summary>
        /// 確定ボタン押下イベント
        /// </summary>
        public event EventHandler DecisionClicked = null;

        /// <summary>
        /// ViewModel
        /// </summary>
        private MainWindowViewModel viewModel = new MainWindowViewModel();

        /// <summary>
        /// 外字
        /// </summary>
        private string _eudc;

        /// <summary>
        /// フォント名
        /// </summary>
        private string _fontName;

        /// <summary>
        /// キャレットインデックス（Excel用）
        /// </summary>
        private int _caretIndex;

        /// <summary>
        /// IMEで変換中か
        /// </summary>
        private bool _imeFlag = false;

        /// <summary>
        /// 入力文字保存用
        /// </summary>
        private string _prevInput = string.Empty;

        #endregion

        #region コンストラクタ

        /// <summary>
        /// コンストラクタ
        /// </summary>
        public AssistControl(bool cellVisibility)
        {
            InitializeComponent();

            TextCompositionManager.AddPreviewTextInputHandler(this.inputTextBox, OnPreviewTextInput);
            TextCompositionManager.AddPreviewTextInputStartHandler(this.inputTextBox, OnPreviewTextInputStart);
            TextCompositionManager.AddPreviewTextInputUpdateHandler(this.inputTextBox, OnPreviewTextInputUpdate);

            this.editButton.IsEnabled = false;
            this.message.Visibility = Visibility.Hidden;

            // 初期化
            cellTextGrid.Visibility = SetVisibility(cellVisibility);

            // 枠にはまるようにサイズ調整
            if (cellVisibility)
                dockPanel.Height = 336;
            else
                dockPanel.Height = 286;

            this.DataContext = viewModel;
            
        }

        #endregion

        #region プロパティ

        /// <summary>
        /// 外字
        /// </summary>
        public string Eudc
        {
            get { return _eudc; }
        }

        /// <summary>
        /// フォント名
        /// </summary>
        public string FontName
        {
            get { return _fontName; }
        }

        public class ExStringItem
        {
            /// <summary>
            /// 1文字
            /// </summary>
            public string OneChar { get; private set; }
            /// <summary>
            /// フォント名
            /// </summary>
            public string FontName { get; private set; }
            /// <summary>
            /// フォントサイズ
            /// </summary>
            public float FontSize { get; private set; }

            /// <summary>
            /// コンストラクタ
            /// </summary>
            /// <param name="oneChar"></param>
            /// <param name="fontName"></param>
            /// <param name="fontSize"></param>
            public ExStringItem(string oneChar, string fontName, float fontSize)
            {
                if (string.IsNullOrEmpty(oneChar) == true)
                {
                    throw new ArgumentNullException("oneChar");
                }
                StringInfo si = new StringInfo(oneChar);
                if (si.LengthInTextElements != 1)
                {
                    throw new ArgumentException("oneCharは必ず1文字を指定してください");
                }

                this.OneChar = oneChar;
                this.FontName = fontName;
                this.FontSize = fontSize;
            }
        }

        /// <summary>
        /// セルテキスト（Excel用）
        /// </summary>
        public List<ExStringItem> CellText
        {
            set
            {
                SetCellTextForOpen(value);
            }
        }

        /// <summary>
        /// キャレットインデックス（Excel用）
        /// </summary>
        public int CaretIndex
        {
            get { return _caretIndex; }
        }


        #endregion

        #region メソッド

        /// <summary>
        /// 入力補助欄にテキストをセットする
        /// </summary>
        /// <param name="value"></param>
        public void SetCellTextForOpen(List<ExStringItem> exStringList)
        {
            // AssistControlが呼び出された直後にSetCellTextForOpenが呼ばれると
            // SelectedItemはnullになってしまうのでDataContextで分岐するようにする
            if (((AssistToolDialog.MainWindowViewModel)(this.DataContext)).IvsFontInfos != null)
            {
                this.ExStringList2RichTextBox(this.cellText, exStringList);
            }
        }

        private void ExStringList2RichTextBox(System.Windows.Forms.RichTextBox rtb, List<ExStringItem> exStringList)
        {
            this.cellText.Clear();
            if (exStringList != null && 0 < exStringList.Count)
            {
                foreach (var item in exStringList)
                {
                    this.cellText.AppendText(item.OneChar);
                    this.cellText.Select(this.cellText.TextLength - 1, item.OneChar.Length);
                    this.cellText.SelectionFont = new Font(item.FontName, DEF_FONST_SIZE, this.cellText.Font.Style); //フォントの設定
                }
                this.cellText.SelectionStart = this.cellText.TextLength;
                this.cellText.SelectionLength = 0;
            }
        }

        /// <summary>
        /// 入力補助欄にテキストをセットする
        /// </summary>
        public void SetCellText()
        {
            // textにinsertしたあとはcaret位置が0になってしまうのでここで保持しておく
            int caretIndex = this.cellText.SelectionStart;

            this.cellText.Select(caretIndex, 0);    //length=0でその場所に挿入
            this.cellText.SelectedText = _eudc;
            this.cellText.Select(caretIndex, _eudc.Length); //選択しなおし
            this.cellText.SelectionFont = new Font(_fontName, DEF_FONST_SIZE, this.cellText.Font.Style);
            this.cellText.SelectionStart = caretIndex + _eudc.Length;
            this.cellText.SelectionLength = 0;
        }

        /// <summary>
        /// 指定文字列をコードポイント表記文字列に変換します（例：A -> U+0041）
        /// </summary>
        /// <param name="value">文字列</param>
        /// <returns>コードポイント表記文字列</returns>
        private static string ToCodePointString(string value)
        {
            if (string.IsNullOrEmpty(value))
            {
                return string.Empty;
            }

            StringBuilder sb = new StringBuilder();

            for (int i = 0; i < value.Length; i++)
            {
                if (!Char.IsSurrogate(value[i]))
                {
                    sb.AppendFormat("U+{0:X4} ", (int)value[i]);
                }
                else
                {
                    int codePoint = 0x10000
                                  + ((value[i++] - 0xD800) << 10)
                                  + (value[i] - 0xDC00);
                    sb.AppendFormat("U+{0:X5} ", codePoint);
                }
            }

            sb.Length = sb.Length - 1;

            return sb.ToString();
        }

        /// <summary>
        /// 外字セット
        /// </summary>
        private void SetEudc()
        {
            // 指定フォントの IVC⇔外字 マッピング情報を取得
            ICharMap charMap = FontManager.GetCharMap(((FontInfo)ivsFontList.SelectedItem).FamilyName);
            // 異体字セレクタ付き文字 -> 外字変換
            _eudc = charMap.ReserveEUDC(((TextBlock)((Grid)(((Border)((ListBoxItem)this.searchListBox.SelectedItem).Content).Child)).Children[1]).Text);
            // フォント名
            _fontName = ((FontInfo)ivsFontList.SelectedItem).FamilyName;
            // 割り当て内容を外字フォントに適用
            charMap.ApplyToFont();
            // キャレットを格納
            _caretIndex = this.cellText.SelectionStart;

            // 終了処理
            //            FontManager.Uninitialize();

            // イベント呼び出し
            DecisionClicked(this, null);
        }

        /// <summary>
        /// Excelの場合のみ表示させるコントロール部分の制御
        /// </summary>
        /// <param name="visibility"></param>
        /// <returns></returns>
        private Visibility SetVisibility(bool visibility)
        {
            if (visibility)
                return Visibility.Visible;
            else
                return Visibility.Collapsed;
        }
        #endregion

        #region イベント

        /// <summary>
        /// フォント変更時
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ivsFontList_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            if (string.IsNullOrEmpty(inputTextBox.Text))
            {
                return;
            }

            this.inputTextBox.FontFamily = new System.Windows.Media.FontFamily(((FontInfo)this.ivsFontList.SelectedItem).FamilyName);

            InputTextBox_TextInput(inputTextBox, null);
        }

        /// <summary>
        /// テキスト取得時
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void InputTextBox_TextInput(object sender, TextCompositionEventArgs e)
        {
            var font = this.ivsFontList.SelectedItem as FontInfo;

            if (font != null)
            {
                this.searchListBox.Items.Clear();
                string moji = (sender as TextBox).Text;

                if (0 < moji.Length)
                {
                    // サロゲート文字の判定
                    if (!Char.IsSurrogate(moji[0]))
                        this.inputTextBox.MaxLength = 1;
                    else
                        this.inputTextBox.MaxLength = 2;

                    // 検索
                    var ivcChars = FontManager.FindIvChar(font.FamilyName, (sender as TextBox).Text);

                    // ListBoxItemの中身を構築する
                    // Glyphs.Indicesでバインドしようとしたが、初期表示時に値がnullだとエラーがでてしまうのでソース側から構築することにした。
                    foreach (var ivcChar in ivcChars)
                    {
                        ListBoxItem item = new ListBoxItem();
                        Grid grid = new Grid();
                        Border border = new Border();
                        Glyphs glyphs = new Glyphs();
                        glyphs.Indices = ivcChar.GlyphIndex.ToString();
                        grid.Children.Add(glyphs);
                        // 値保持用
                        TextBlock hideTextBlock = new TextBlock();
                        // 裏で持たせるだけなので表示はさせない
                        // 異体字セレクタ付き文字 -> 外字変換のときに使用する
                        hideTextBlock.Visibility = Visibility.Collapsed;
                        hideTextBlock.Text = ivcChar.Code;
                        grid.Children.Add(hideTextBlock);

                        border.Child = grid;

                        border.ToolTip = string.Format("Unicode: {0}\nCollection: {1}\nGlyph Name: {2}", ToCodePointString(ivcChar.Code),
                                          ivcChar.IvdCollectionName,
                                          ivcChar.GlyphName);

                        item.Content = border;
                        this.searchListBox.Items.Add(item);
                    }
                }
            }

            if (0 < this.searchListBox.Items.Count)
            {
                this.searchListBox.SelectedIndex = 0;
                this.message.Visibility = Visibility.Hidden;
            }
            else if ((sender as TextBox).Text.Length == 0)
            {
                this.message.Visibility = Visibility.Hidden;
            }
            else
            {
                this.message.Visibility = Visibility.Visible;
            }
        }

        /// <summary>
        /// KeyDown時
        /// </summary>
        /// <param name="e"></param>
        protected override void OnPreviewKeyDown(KeyEventArgs e)
        {
            if (e.Key == Key.Return)
            {
                if (this.editButton.IsEnabled)
                    SetEudc();
            }
            else
            {
                base.OnPreviewKeyDown(e);
            }
        }

        /// <summary>
        /// PreviewKeyDown時
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void InputTextBox_PreviewKeyDown(object sender, KeyEventArgs e)
        {
            // 選択したフォントでFontFamilyを設定する
            if (0 < this.ivsFontList.Items.Count)
                this.inputTextBox.FontFamily = new System.Windows.Media.FontFamily(((FontInfo)this.ivsFontList.SelectedItem).FamilyName);
            // 1文字しか入力できない仕様なのでBackかDeleteが入力されたら基底文字は消える
            if (e.Key == Key.Back && 0 < inputTextBox.CaretIndex || e.Key == Key.Delete && inputTextBox.CaretIndex == 0)
                this.message.Visibility = Visibility.Hidden;

            // 最後にlengthを2にしておかないとサロゲートペアが入力できない
            this.inputTextBox.MaxLength = 2;

            
        }

        /// <summary>
        /// EditButton押下時
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void EditButton_Click(object sender, RoutedEventArgs e)
        {
            SetEudc();
        }

        /// <summary>
        /// 検索結果一覧選択時
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void SearchListBox_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            if (0 < this.searchListBox.SelectedItems.Count)
                this.editButton.IsEnabled = true;
            else
                this.editButton.IsEnabled = false;
        }

        /// <summary>
        /// 一覧ダブルクリック時
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void searchListBox_PreviewMouseDoubleClick(object sender, MouseButtonEventArgs e)
        {
            if (0 < this.searchListBox.Items.Count)
            {
                SetEudc();
            }
        }

        /// <summary>
        /// フォーカス取得時
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void inputTextBox_GotFocus(object sender, RoutedEventArgs e)
        {
            if (this.inputTextBox.IsFocused)
            {
                if (this.ivsFontList.SelectedItem != null)
                {
                    this.inputTextBox.FontFamily = new System.Windows.Media.FontFamily(((FontInfo)ivsFontList.SelectedItem).FamilyName);
                }
            }
        }

        /// <summary>
        /// 入力変更時
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void inputTextBox_TextChanged(object sender, TextChangedEventArgs e)
        {
            if (_imeFlag)
                return;

            // サロゲート文字コピペ対応
            string moji = (sender as TextBox).Text;
            if (1 < (sender as TextBox).Text.Length && !Char.IsSurrogate(moji[0]))
                (sender as TextBox).Text = _prevInput;

            InputTextBox_TextInput(sender, null);

            _prevInput = (sender as TextBox).Text;

            // キャレットの位置が頭に移動することがあるので末尾に設定する
            (sender as TextBox).CaretIndex = (sender as TextBox).Text.Length;
        }

        #region IME状態チェック

        private void OnPreviewTextInput(object sender, TextCompositionEventArgs e)
        {
            _imeFlag = false;
        }

        private void OnPreviewTextInputStart(object sender, TextCompositionEventArgs e)
        {
            _imeFlag = true;
        }

        private void OnPreviewTextInputUpdate(object sender, TextCompositionEventArgs e)
        {
            if (e.TextComposition.CompositionText.Length == 0)
                _imeFlag = false;
        }

        #endregion

        #endregion
    }
}
