﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Runtime.InteropServices;
using System.Security.AccessControl;

using Microsoft.Win32;

using OfficeIVS.Font;
using OfficeIVS.Font.Table;


namespace OfficeIVS
{

	/// <summary>
	/// 外字フォントのマッピング機能を提供します。
	/// </summary>
	internal sealed class CharMap : ICharMap, IDisposable
	{
		#region 定数値

		/// <summary>
		/// 外字割り当てフラグテーブルの開始位置を表します。
		/// </summary>
		/// <remarks>ファイル先頭からのオフセット量（バイト単位）。</remarks>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private const int EudcFlagsTableOffset = 0;

		/// <summary>
		/// 外字割り当てフラグテーブル長を表します。
		/// </summary>
		/// <remarks>テーブル長（バイト単位）。</remarks>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private const int EudcFlagsTableLength = 6400 / 8;

		/// <summary>
		/// EUDC -> IVS 変換テーブルの開始位置を表します。
		/// </summary>
		/// <remarks>ファイル先頭からのオフセット量（バイト単位）。</remarks>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private const int EudcIvsTableOffset = EudcFlagsTableOffset + EudcFlagsTableLength;

		/// <summary>
		/// EUDC -> IVS 変換テーブル長を表します。
		/// </summary>
		/// <remarks>テーブル長（バイト単位）。</remarks>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private const int EudcIvsTableLength = 2 * 6400;

		/// <summary>
		/// IVS 文字情報テーブルまでの開始位置を表します。
		/// </summary>
		/// <remarks>ファイル先頭からのオフセット量（バイト単位）。</remarks>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private const int IvsInfoTableOffset = EudcIvsTableOffset + EudcIvsTableLength;

		/// <summary>
		/// IVS 文字情報テーブルレコードサイズを表します。
		/// </summary>
		/// <value>レコードサイズ（バイト単位）。</value>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private const int IvsInfoRecordLength = 24;

		/// <summary>
		/// 外字割り当てフラグの検索用マスク値の配列です。
		/// </summary>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private static readonly byte[] EudcFlagMasks;

		#endregion

		#region フィールド

		/// <summary>
		/// 外字フォント出力フォルダパス。
		/// </summary>
		private static readonly string EudcFontFolderPath;

		/// <summary>
		/// 元フォントのフォント情報。
		/// </summary>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private readonly FontInfo _baseFontInfo;

		/// <summary>
		/// 元フォントの cmap テーブル format 14 サブテーブル。
		/// </summary>
		/// <remarks>基底文字, 字形選択肢, グリフ番号のマッピングテーブル。</remarks>
		private readonly SortedList<int, SortedList<int, ushort>> _baseFontCmapFormat14;

		/// <summary>
		/// 異体字文字レコードインデックス番号テーブル。
		/// </summary>
		/// <remarks>
		/// 異体字セレクタ付き文字と対応する異体字文字情報レコードのインデックス番号のテーブル。<br/>
		/// 　キー: 異体字セレクタ付き文字の Unicode コードポイント。<br/>
		/// 　　　　（上位 32 bit: 基底文字 下位 32bi: 字形選択肢）<br/>
		/// 　値　: 異体字文字情報レコードのインデックス番号。<br/>
		/// </remarks>
		private readonly Dictionary<long, ushort> _ivsRecordIndexTable;

		/// <summary>
		/// 外字ファイルアクセス同期名称。
		/// </summary>
		private readonly string _syncName;

		/// <summary>
		/// 外字フォントファイル名（フルパス）。
		/// </summary>
		private readonly string _eudcFontFileName;

		/// <summary>
		/// 外字フォントデータファイル名（フルパス）。
		/// </summary>
		private readonly string _eudcFontDataFileName;

		/// <summary>
		/// 外字登録予約中の異体字文字情報レコードのインデックス番号コレクション。
		/// </summary>
		private readonly List<int> _reservedIvsRecordIndexes = new List<int>();

		/// <summary>
		/// 異体字コードから外字コードへの変換キャッシュ。
		/// </summary>
		private readonly Dictionary<long, int> _ivsToEudcCaches = new Dictionary<long, int>();

		/// <summary>
		/// 外字コードから異体字コードへの変換キャッシュ。
		/// </summary>
		private readonly Dictionary<int, long> _eudcToIvsCaches = new Dictionary<int, long>();

		#endregion

		#region コンストラクタ

		/// <summary>
		/// <see cref="CharMap"/> クラスを初期化します。
		/// </summary>
		static CharMap()
		{
			// 外字フォントの出力フォルダパス
			EudcFontFolderPath = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.LocalApplicationData),
											  @"Microsoft\Windows\EUDC");

			if (!Directory.Exists(EudcFontFolderPath))
			{
				Directory.CreateDirectory(EudcFontFolderPath);
			}

			// 外字割り当てフラグの検索用マスク
			EudcFlagMasks = new byte[]
			{
				0xFF, 0x7F, 0x3F, 0x1F, 0x0F, 0x07, 0x03, 0x01,
				0x80, 0xC0, 0xE0, 0xF0, 0xF8, 0xFC, 0xFE, 0xFF,
			};
		}

		/// <summary>
		/// <see cref="CharMap"/> クラスの新しいインスタンスを初期化します。
		/// </summary>
		/// <param name="baseFontInfo">元フォント情報。</param>
		/// <param name="baseFont">元フォント。</param>
		/// <param name="checkCodeRange">コード範囲を検証するかどうか。</param>
		private CharMap(FontInfo baseFontInfo, OpenTypeFont baseFont, bool checkCodeRange)
		{
			// フォント情報作成
			_baseFontInfo = baseFontInfo;

			// 元フォントの cmap format 14 サブテーブルを取得
			_baseFontCmapFormat14 = ((CMap00)baseFont.GetTable(TableTags.cmap)).IvsToGlyphMap;

			// ivs レコードインデックス番号テーブル初期化
			_ivsRecordIndexTable = new Dictionary<long, ushort>();
			{
				ushort recordIndex = 0;

				foreach (int bc in _baseFontCmapFormat14.Keys)
				foreach (int vs in _baseFontCmapFormat14[bc].Keys)
				{
					_ivsRecordIndexTable[CodePoint.ToIvsCode(bc, vs)] = recordIndex++;
				}
			}

			// 外字フォントファイルパス
			_eudcFontFileName = Path.Combine(EudcFontFolderPath,
											 String.Format(@"{0}_EUDC{1}",
														   _baseFontInfo.FontFamilyName,
														   Path.GetExtension(_baseFontInfo.Uri.LocalPath)));

			// 外字フォントデータファイルパス
			_eudcFontDataFileName = Path.ChangeExtension(_eudcFontFileName, @".dat");

			// ファイルアクセス同期用セマフォ
			_syncName = String.Format(@"Office.AddIn.Ivs.{0}",
									  Path.GetFileNameWithoutExtension(_eudcFontFileName));

			// フォントデータ初期化
			InitializeEudcFont(baseFont, checkCodeRange);
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// マッピング情報の対象フォント情報を取得します。
		/// </summary>
		/// <value>マッピング情報の対象フォント情報。</value>
		public FontInfo FontInfo
		{
			[DebuggerStepThrough]
			get { return _baseFontInfo; }
		}

		#endregion

		#region パブリック メソッド

		/// <summary>
		/// システムフォントフォルダ内から全ての IVS 対応フォントを読み込みます。
		/// </summary>
		/// <param name="checkCodeRange">コード範囲を検証するかどうか。</param>
		/// <returns>IVS 対応フォントコレクション。</returns>
		public static IEnumerable<CharMap> LoadAllIvsFonts(bool checkCodeRange)
		{
			List<CharMap> result = new List<CharMap>();

			// 外字 OFF
			EnableEUDC(false);

			// システムフォントフォルダ内のフォントを列挙
			foreach (string fontFilePath in EnumerateSystemFontFiles())
			{
				// 拡張子で OpenType、TrurType フォントを選別
				string extension = Path.GetExtension(fontFilePath);

				if (String.IsNullOrEmpty(extension))
				{
					continue;
				}

				extension = extension.ToUpper();

				if (extension != ".TTF" && extension != ".OTF")
				{
					continue;
				}

				try
				{
					// フォント読み込み
					using (OpenTypeFont font = OpenTypeFont.Load(new Uri(fontFilePath)))
					{
						// 未対応のフォント形式の場合 null が返されるので、その対応
						if (font == null)
						{
							continue;
						}

						// 証明書付きのフォントは改変できないので無視
						if (font.GetTable(TableTags.DSIG) != null)
						{
							continue;
						}

						// IVS をサポートしていなければ無視
						if (!font.IsSupportIVS())
						{
							continue;
						}

						// 'head' テーブル取得
						Head10 head = font.GetTable(TableTags.head) as Head10;

						if (head == null)
						{
							// 未対応の 'head' テーブルバージョンのため残念ながら無視する
							continue;
						}

						// 外字フォントの登録
						result.Add(new CharMap(new FontInfo(font), font, checkCodeRange));
					}	
				}
				catch (FileFormatException)
				{
					// 未対応形式(書式エラーも含む)による例外は無視
				}
				catch (NotSupportedException)
				{
					// 未対応のフォントテーブルバージョン、フォントファミリ名が取得できないなど	
				}
			}

			// 外字 ON
			EnableEUDC(true);

			return result;
		}

		/// <summary>
		/// 予約した外字文字を外字フォントファイルに適用します。
		/// </summary>
		/// <param name="updateLastUsedTime">最終使用日時を更新するかどうか。更新する場合には true（既定値）。</param>
		public void ApplyToFont(bool updateLastUsedTime)
		{
			if (_reservedIvsRecordIndexes.Count == 0)
			{
				return;
			}

			using (new SyncAccess(_syncName))
			{
				bool changeCMap = false;

				// 現在の外字フォントの cmap 読み込み
				CMap00 cmap;

				using (OpenTypeFont eudcFont = OpenTypeFont.Load(new Uri(_eudcFontFileName)))
				{
					cmap = (CMap00)eudcFont.GetTable(TableTags.cmap);
				}
				
				// 外字フォント情報ファイルの更新
				using (FileStream fs = File.Open(_eudcFontDataFileName, FileMode.Open, FileAccess.ReadWrite, FileShare.None))
				using (BinaryReader reader = new BinaryReader(fs))
				using (BinaryWriter writer = new BinaryWriter(fs))
				{
					long lastUsedTime = DateTime.Now.ToBinary();

					foreach (int recordIndex in _reservedIvsRecordIndexes)
					{
						fs.Position = GetIvsInfoRecordOffset(recordIndex);

						// IVS 文字情報レコードの読み込み
						int bc = reader.ReadInt32();
						int vs = reader.ReadInt32();
						ushort eudc = reader.ReadUInt16();
						short refCount = reader.ReadInt16();
						uint count = reader.ReadUInt32();

						// IVS 文字情報レコードの更新
						{
							fs.Position -= 6;

							// 参照カウント
							refCount = unchecked ((short)(refCount + 1));
							if (refCount < 0)
							{
								refCount = 1;
							}
							writer.Write(refCount);

							// 利用回数
							count = unchecked (count + 1);
							if (count == 0)
							{
								count = UInt32.MaxValue;
							}
							writer.Write(count);

							// 最終利用日時
							if (updateLastUsedTime)
							{
								writer.Write(lastUsedTime);
							}
						}

						// cmap 登録
						{
							ushort glyphId = _baseFontCmapFormat14[bc][vs];

							if (cmap.CharacterToGlyphMap.ContainsKey(eudc)
							 && cmap.CharacterToGlyphMap[eudc] == glyphId)
							{
								// 既に登録済みであれば無視
								continue;
							}

							changeCMap = true;
							cmap.CharacterToGlyphMap[eudc] = glyphId;
						}
					}
				}

				// 外字フォント更新
				if (changeCMap)
				{
					EnableEUDC(false);
					OpenTypeFont.ChangeCMapTable(_eudcFontFileName, cmap);
					EnableEUDC(true);
				}
			}

			_reservedIvsRecordIndexes.Clear();
		}

		/// <summary>
		/// 本インスタンスが使用しているリソースを解放します。
		/// </summary>
		public void Dispose()
		{
			// 参照カウント更新
			if (_ivsToEudcCaches.Count > 0)
			{
				using (new SyncAccess(_syncName))
				using (FileStream fs = File.Open(_eudcFontDataFileName, FileMode.Open, FileAccess.ReadWrite, FileShare.None))
				using (BinaryReader reader = new BinaryReader(fs))
				using (BinaryWriter writer = new BinaryWriter(fs))
				{
					foreach (long ivsCode in _ivsToEudcCaches.Keys)
					{
						fs.Position = GetIvsInfoRecordOffset(_ivsRecordIndexTable[ivsCode]) + 10;

						short refCount = reader.ReadInt16();
						fs.Position -= 2;
						writer.Write(refCount <= 0 ? (short)0 : (short)(refCount - 1));
					}
				}

				_ivsToEudcCaches.Clear();
				_eudcToIvsCaches.Clear();
			}

			GC.SuppressFinalize(this);
		}

		/// <summary>
		/// 外字領域の空き領域に指定異体字セレクタ付き文字を予約します。
		/// </summary>
		/// <param name="ivs">異体字セレクタ付き文字。</param>
		/// <returns>
		/// 予約された外字文字コード。
		/// 空きがなく予約できない場合には <c>null</c> を返します。
		/// </returns>
		public string ReserveEUDC(string ivs)
		{
			string eudc = null;

			// 文字列から文字コードに変換
			long ivsCode = CodePoint.ToIvsCodePoint(ivs);

			// キャッシュから検索
			if (_ivsToEudcCaches.ContainsKey(ivsCode))
			{
				eudc = Convert.ToChar(_ivsToEudcCaches[ivsCode]).ToString();
			}
			// キャッシュになければ新規登録
			else if (_ivsRecordIndexTable.ContainsKey(ivsCode))
			{
				int eudcCode;
				bool reserved;

				int ivsRecordIndex = _ivsRecordIndexTable[ivsCode];

				using (new SyncAccess(_syncName))
				using (FileStream fs = File.Open(_eudcFontDataFileName, FileMode.Open, FileAccess.ReadWrite, FileShare.None))
				using (BinaryReader reader = new BinaryReader(fs))
				using (BinaryWriter writer = new BinaryWriter(fs))
				{
					// ivs 文字情報レコードから eudc コード取得
					fs.Position = GetIvsInfoRecordOffset(ivsRecordIndex) + 8;
					eudcCode = reader.ReadUInt16();
					reserved = reader.ReadInt16() < 0;

					// 割り当てされていないければ空きを探す
					if (!CodePoint.IsEudcCode(eudcCode))
					{
						reserved = true;

						// 割り当て可能な外字コード範囲の中から使用していない外字コードを検索
						eudcCode = FindEmptyEudcCode(fs);

						// 空きが存在しない場合には、一番利用されていない外字を検索
						if (eudcCode == -1)
						{
							long replaceIvsCode = FindMostUnusualIvsCode(reader);

							// 置き換え対象となる IVS 文字の外字登録を解除
							if (replaceIvsCode != -1)
							{
								fs.Position = _ivsRecordIndexTable[replaceIvsCode] + 8;
								eudcCode = reader.ReadUInt16();
								fs.Position -= 2;
								writer.Write((ushort)0);
							}
						}

						// 外字コードに割り当て
						if (eudcCode != -1)
						{
							// eudc->ivs テーブルの書き換え
							fs.Position = GetIvsIndexRecordOffset(eudcCode);
							writer.Write((ushort)ivsRecordIndex);

							// ivs 文字情報レコードの書き換え
							fs.Position = GetIvsInfoRecordOffset(ivsRecordIndex) + 8;
							writer.Write((ushort)eudcCode);
							writer.Write((short)-1);
						}
					}
				}

				if (CodePoint.IsEudcCode(eudcCode))
				{
					// 結果
					eudc = Convert.ToChar(eudcCode).ToString();

					// 予約登録
					if (reserved)
					{
						_reservedIvsRecordIndexes.Add(ivsRecordIndex);
					}

					// キャッシュ更新
					_ivsToEudcCaches[ivsCode] = eudcCode;
					_eudcToIvsCaches[eudcCode] = ivsCode;
				}
			}

			return eudc;
		}

		/// <summary>
		/// 指定外字文字に関連付けられている異体字セレクタ付き文字を取得します。
		/// </summary>
		/// <param name="eudc">外字文字。</param>
		/// <returns>
		/// <paramref name="eudc"/> に対応する異体字セレクタ付き文字。
		/// 関連付けされていない場合は <paramref name="eudc"/> を返します。
		/// </returns>
		public string GetIvs(string eudc)
		{
			string ivs = null;

			if (CodePoint.IsEudcCode(eudc))
			{
				if (_eudcToIvsCaches.ContainsKey(eudc[0]))
				{
					ivs = CodePoint.FromIvsCodePoint(_eudcToIvsCaches[eudc[0]]);
				}
				else
				{
					using (new SyncAccess(_syncName))
					using (FileStream fs = File.Open(_eudcFontDataFileName, FileMode.Open, FileAccess.Read, FileShare.None))
					using (BinaryReader reader = new BinaryReader(fs))
					{
						// eudc 割り当てフラグチェック
						int offset = eudc[0] - 0xE000;
						fs.Position = EudcFlagsTableOffset + (offset / 8);
						byte flag = reader.ReadByte();

						if ((flag & (1 << (7 - (offset & 7)))) != 0)
						{
							fs.Position = GetIvsInfoRecordOffset(reader, eudc[0]);
							ivs = CodePoint.FromCodePoint(reader.ReadInt32(), reader.ReadInt32());
						}
					}
				}
			}

			return ivs ?? eudc;
		}

		/// <summary>
		/// 外字領域割り当て済みの異体字セレクタ付き文字の一覧を取得します。
		/// </summary>
		/// <returns>外字領域割り当て済みの異体字セレクタ付き文字の外字コード、異体字セレクタ付き文字ペアのコレクション。</returns>
		public List<Tuple<char, string>> GetRegistedIvs()
		{
			List<Tuple<char, string>> result = new List<Tuple<char, string>>();

			using (new SyncAccess(_syncName))
			using (FileStream fs = File.Open(_eudcFontDataFileName, FileMode.Open, FileAccess.Read, FileShare.None))
			using (BinaryReader reader = new BinaryReader(fs))
			{
				// 外字割り当てフラグ読み込み
				byte[] eudcFlags = new byte[EudcFlagsTableLength];
				fs.Position = EudcFlagsTableOffset;
				fs.Read(eudcFlags, 0, eudcFlags.Length);

				// 外字割り当てされている異体字セレクタ文字を取得
				int eudc = 0xE000;

				foreach (byte eudcFlag in eudcFlags)
				{
					for (byte bitMask=0x80; bitMask>0; bitMask>>=1, eudc++)
					{
						if ((eudcFlag & bitMask) == 0)
						{
							continue;
						}

						fs.Position = GetIvsInfoRecordOffset(reader, eudc);
						string ivs = CodePoint.FromCodePoint(reader.ReadInt32(), reader.ReadInt32());

						if (ivs != null)
						{
							result.Add(new Tuple<char, string>(Convert.ToChar(eudc), ivs));
						}
					}
				}
			}

			return result;
		}

		/// <summary>
		/// フォント内の指定文字を基底文字とする異体字の一覧を取得します。
		/// </summary>
		/// <param name="baseChar">基底文字の Unicode コードポイント。</param>
		/// <returns>
		/// 検索条件に一致した異体字コレクション（字形選択肢、グリフ番号ペアのコレクション）。
		/// 一致する異体字が存在しない場合には空コレクション。
		/// </returns>
		internal IDictionary<int, ushort> FindIvChar(int baseChar)
		{
			SortedList<int, ushort> result;

			if (!_baseFontCmapFormat14.TryGetValue(baseChar, out result))
			{
				result = new SortedList<int, ushort>();
			}

			return result;
		}

		#endregion

		#region プライベート メソッド

		/// <summary>
		/// システムフォントとしてインストールされている有効な OpenType フォントのフォントファイル一覧を列挙します。
		/// </summary>
		/// <returns>有効なフォントファイル（フルパス）の一覧。</returns>
		/// <remarks>有効なフォントファイルとは、フォントフォルダで「プログラムで表示する」設定が有効になっているフォント。</remarks>
		private static IEnumerable<string> EnumerateSystemFontFiles()
		{
			// システムフォントディレクトリの取得
			string systemFontFolder = Environment.GetFolderPath(Environment.SpecialFolder.Fonts);

			// インストール済みフォント内で列挙しないフォント一覧を取得
			HashSet<string> inactiveFonts = new HashSet<string>();
			{
				using (RegistryKey fontManagement =
					Registry.CurrentUser.OpenSubKey(@"Software\Microsoft\Windows NT\CurrentVersion\Font Management",
													RegistryKeyPermissionCheck.ReadSubTree,
													RegistryRights.QueryValues))
				if (fontManagement != null)
				{
					string[] inactiveFontNames = fontManagement.GetValue(@"Inactive Fonts") as string[];

					if (inactiveFontNames != null)
					{
						foreach (string inactiveFontName in inactiveFontNames)
						{
							inactiveFonts.Add(inactiveFontName);
						}
					}
				}
			}

			// インストール済みフォントの取得
			using (RegistryKey fonts =
				Registry.LocalMachine.OpenSubKey(@"SOFTWARE\Microsoft\Windows NT\CurrentVersion\Fonts",
												 RegistryKeyPermissionCheck.ReadSubTree,
												 RegistryRights.QueryValues))
			if (fonts != null)
			{
				const string fontType = @" (TrueType)";

				foreach (string valueName in fonts.GetValueNames())
				{
					// フォント形式が TrueType でないものは無視
					if (!valueName.EndsWith(fontType))
					{
						continue;
					}

					// 非表示設定のフォントは無視
					if (inactiveFonts.Contains(valueName.Substring(0, valueName.Length-fontType.Length)))
					{
						continue;
					}

					string fontFileName = fonts.GetValue(valueName) as string;

					if (fontFileName == null)
					{
						continue;
					}

					// フォントファイルのフルパス取得
					string fontFilePath = Path.IsPathRooted(fontFileName) ? fontFileName
																		  : Path.Combine(systemFontFolder, fontFileName);

					// ファイルが存在しなければ無視
					if (!File.Exists(fontFilePath))
					{
						continue;
					}

					yield return fontFilePath;
				}
			}
		}

		/// <summary>
		/// 外字フォントを初期化します。
		/// </summary>
		/// <param name="baseFont">外字フォントの元となるフォント。</param>
		/// <param name="checkCodeRange">コード範囲を検証するかどうか。</param>
		private void InitializeEudcFont(OpenTypeFont baseFont, bool checkCodeRange)
		{
			using (new SyncAccess(_syncName))
			{
				// 外字フォント情報ファイルの作成
				if (!File.Exists(_eudcFontDataFileName))
				{
					checkCodeRange = false;
					CreateNewEudcFontDataFile();
				}

				// 外字フォントが存在しなければ新規作成
				if (!File.Exists(_eudcFontFileName))
				{
					// フォント作成
					CreateNewEudcFont(baseFont);
				}

				// 外字フォントが存在する場合には、元フォントの更新状態により再作成する
				else
				{
					bool update = false;

					// 元フォントの更新チェック
					using (OpenTypeFont eudcFont = OpenTypeFont.Load(new Uri(_eudcFontFileName)))
					{
						if (eudcFont != null)
						{
							Head10 baseHead = (Head10)baseFont.GetTable(TableTags.head);
							Head10 eudcHead = (Head10)eudcFont.GetTable(TableTags.head);

							update = baseHead.FontRevision != eudcHead.FontRevision
								  || baseHead.Created != eudcHead.Created
								  || baseHead.Modified != eudcHead.Modified;
						}
					}

					// 元フォントが更新されているため外字フォントを強制再生成
					if (update)
					{
						// 外字フォント削除
						File.Delete(_eudcFontFileName);

						// 外字フォント情報データファイルの更新
						UpdateEudcFontDataFile();

						// フォント再生成
						CreateNewEudcFont(baseFont);
					}
				}

				// 外字コード範囲のチェック
				if (checkCodeRange)
				{
					CheckCodeRange();
				}

				// 外字フォント登録
				using (RegistryKey eudcKey = Registry.CurrentUser.OpenSubKey(@"EUDC\932", true))
				{
					if (eudcKey != null)
					{
						eudcKey.SetValue(_baseFontInfo.FontFamilyName, _eudcFontFileName, RegistryValueKind.String);	
					}
				}
			}
		}

		/// <summary>
		/// 指定フォントを元に新規に外字フォントを作成します。
		/// </summary>
		/// <returns>新しく作成された外字フォント。</returns>
		private void CreateNewEudcFont(OpenTypeFont baseFont)
		{
			OpenTypeFont eudcFont = new OpenTypeFont(baseFont.SfntVersion);

			try
			{
				// テーブルコピー
				foreach (ITableRecord tableRecord in baseFont.TableRecords)
				{
					eudcFont.AddTableRecord(tableRecord);
				}

				// head
				Head10 eudcHead = (Head10)eudcFont.GetTable(TableTags.head);

				// cmap
				CMap00 eudcCmap = new CMap00();
				{
					int codePoint = FontManager.GetEudcRange()[0].StartCharCode;
					eudcCmap.CharacterToGlyphMap[codePoint] = 0;
				}
				eudcFont.AddTableRecord(new EditableTableRecord(eudcCmap, 0x40000));

				// OS/2
				OS2_10 eudcOS2 = (OS2_10)baseFont.GetTable(TableTags.OS2);
				{
					// 外字表示の場合、xAvgCharWidth をグリフバウンディングボックスの
					// 半分の値にしておかないと、横方向のサイズが正しく表示されない
					eudcOS2.XAvgCharWidth = (short)(eudcHead.UnitsPerEm / 2);

					// 外字表示の場合、usWinAscent と usWinDescent の合計値が
					// グリフのバウンディングボックスサイズと同じでないと
					// フォント表示縦方向のサイズが正しく表示されない
					eudcOS2.WinAscent = unchecked ((ushort)eudcHead.YMax);
					eudcOS2.WinDescent = unchecked ((ushort)(eudcHead.UnitsPerEm - eudcHead.YMax));
					eudcOS2.TypoAscender = eudcHead.YMax;
					eudcOS2.TypoDescender = unchecked ((short)-eudcOS2.WinDescent);

					// 現在の cmap の情報に修正
					eudcOS2.UnicodeRange1 = 0x00000000;
					eudcOS2.UnicodeRange2 = 0x10000000;
					eudcOS2.UnicodeRange3 = 0x00000000;
					eudcOS2.UnicodeRange4 = 0x00000000;
					eudcOS2.CodePageRange1 = 0x00000000;
					eudcOS2.CodePageRange2 = 0x00000000;
					eudcOS2.FirstCharIndex = eudcCmap.FirstCharIndex;
					eudcOS2.LastCharIndex = eudcCmap.LastCharIndex;
				}
				eudcFont.AddTableRecord(new EditableTableRecord(eudcOS2));

				// ファイル出力
				using (FileStream fs = File.Open(_eudcFontFileName, FileMode.Create))
				{
					eudcFont.WriteTo(fs);
				}
			}
			catch
			{
				eudcFont.Dispose();
			}
		}

		/// <summary>
		/// 新規に外字フォント情報データファイルを作成します。
		/// </summary>
		private void CreateNewEudcFontDataFile()
		{
			// 新規作成
			using (FileStream fs = File.Open(_eudcFontDataFileName, FileMode.Create, FileAccess.Write, FileShare.None))
			using (BinaryWriter writer = new BinaryWriter(fs))
			{
				// 割り当てフラグ領域
				fs.Position += EudcFlagsTableLength;

				// eudc->ivs 変換テーブル領域
				fs.Position += EudcIvsTableLength;

				// ivs 文字情報テーブル
				foreach (int bc in _baseFontCmapFormat14.Keys)
				foreach (int vs in _baseFontCmapFormat14[bc].Keys)
				{
					WriteDefaultIvsInfoRecord(writer, bc, vs);
				}

				// 終端レコード
				WriteTerminationIvsInfoRecord(writer);

				// コードレンジ
				{
					IList<CodeRange> codeRanges = FontManager.GetEudcRange();

					foreach (CodeRange codeRange in codeRanges)
					{
						writer.Write(codeRange.StartCharCode);
						writer.Write(codeRange.EndCharCode);
					}

					writer.Write(codeRanges.Count);
				}
			}
		}

		/// <summary>
		/// 元フォント更新に伴う外字フォント情報データファイルの更新を行います。
		/// </summary>
		private void UpdateEudcFontDataFile()
		{
			// 外字フォント情報データファイルが存在しなければ作成
			if (!File.Exists(_eudcFontDataFileName))
			{
				CreateNewEudcFontDataFile();
				return;
			}

			// 異体字情報の引き継ぎ用データバッファ
			Dictionary<long, byte[]> inheritingData = new Dictionary<long, byte[]>();

			using (FileStream fs = File.Open(_eudcFontDataFileName, FileMode.Open, FileAccess.ReadWrite, FileShare.None))
			using (BinaryReader reader = new BinaryReader(fs))
			using (BinaryWriter writer = new BinaryWriter(fs))
			{
				// 現在の ivs 文字情報テーブルをレコード単位で読み込む
				fs.Position = IvsInfoTableOffset;
				{
					while (true)
					{
						long ivsCode = CodePoint.ToIvsCode(reader.ReadInt32(), reader.ReadInt32());

						if (ivsCode == -1)
						{
							break;
						}

						byte[] recordData = new byte[IvsInfoRecordLength];
						fs.Position -= 8;
						fs.Read(recordData, 0, recordData.Length);

						inheritingData[ivsCode] = recordData;
					}
				}

				// 新しい ivs 文字情報テーブルを出力
				fs.Position = IvsInfoTableOffset;
				{
					// ivs 文字情報テーブル
					foreach (int bc in _baseFontCmapFormat14.Keys)
					foreach (int vs in _baseFontCmapFormat14[bc].Keys)
					{
						byte[] recordData;

						if (!inheritingData.TryGetValue(CodePoint.ToIvsCode(bc, vs), out recordData))
						{
							WriteDefaultIvsInfoRecord(writer, bc, vs);
						}
						else
						{
							fs.Write(recordData, 0, recordData.Length);
						}
					}

					// 終端レコード
					WriteTerminationIvsInfoRecord(writer);
				}

				// コードレンジ
				{
					IList<CodeRange> codeRanges = FontManager.GetEudcRange();

					foreach (CodeRange codeRange in codeRanges)
					{
						writer.Write(codeRange.StartCharCode);
						writer.Write(codeRange.EndCharCode);
					}

					writer.Write(codeRanges.Count);
				}

				fs.SetLength(fs.Position);
			}
		}

		/// <summary>
		/// 割り当て可能な外字コード範囲を検証します。
		/// </summary>
		private void CheckCodeRange()
		{
			using (FileStream fs = File.Open(_eudcFontDataFileName, FileMode.Open, FileAccess.ReadWrite, FileShare.None))
			using (BinaryReader reader = new BinaryReader(fs))
			using (BinaryWriter writer = new BinaryWriter(fs))
			{
				// 外字フォント設定ファイルに設定されている外字コード範囲を取得
				CodeRange[] codeRanges;
				{
					fs.Seek(-4, SeekOrigin.End);

					codeRanges = new CodeRange[reader.ReadInt32()];

					fs.Seek(-(4 + 8 * codeRanges.Length), SeekOrigin.End);

					for (int i=0; i<codeRanges.Length; i++)
					{
						codeRanges[i] = new CodeRange(reader.ReadInt32(), reader.ReadInt32());
					}
				}

				// 現在の外字コード範囲と比較
				IList<CodeRange> currentCodeRanges = FontManager.GetEudcRange();

				bool isChanged = codeRanges.Length != currentCodeRanges.Count;

				if (!isChanged)
				{
					for (int i=0; i<codeRanges.Length; i++)
					{
						if (codeRanges[i].StartCharCode != currentCodeRanges[i].StartCharCode
						 || codeRanges[i].EndCharCode != currentCodeRanges[i].EndCharCode)
						{
							isChanged = true;
							break;
						}
					}
				}

				// コード範囲が異なる場合にはリセット
				if (isChanged)
				{
					// 外字フォント情報ファイル更新
					{
						// 割り当てフラグ領域のクリア
						fs.Position = EudcFlagsTableOffset;
						for (int i=0; i<EudcFlagsTableLength/8; i++)
						{
							writer.Write((ulong)0);
						}

						// eudc->ivs テーブルクリア
						fs.Position = EudcIvsTableOffset;
						for (int i=0; i<EudcIvsTableLength/8; i++)
						{
							writer.Write((ulong)0);
						}

						// ivs 文字情報レコード
						fs.Position = IvsInfoTableOffset;
						while (reader.ReadInt32() != -1)
						{
							// eudc, refCount クリア
							fs.Seek(4, SeekOrigin.Current);
							writer.Write(0);
							fs.Seek(IvsInfoRecordLength-12, SeekOrigin.Current);
						}

						// 終端レコード
						fs.Seek(IvsInfoRecordLength - 4, SeekOrigin.Current);

						// コードレンジ
						{
							foreach (CodeRange codeRange in currentCodeRanges)
							{
								writer.Write(codeRange.StartCharCode);
								writer.Write(codeRange.EndCharCode);
							}
							writer.Write(currentCodeRanges.Count);
						}
						fs.SetLength(fs.Position);
					}

					// 外字フォント更新
					{
						// cmap クリア
						CMap cmap = new CMap00();
						cmap.CharacterToGlyphMap[currentCodeRanges[0].StartCharCode] = 0;
						OpenTypeFont.ChangeCMapTable(_eudcFontFileName, cmap);
					}
				}
			}
		}

		/// <summary>
		/// 指定 EUDC の IVS 変換テーブルレコードの位置を取得します。
		/// </summary>
		/// <param name="eudcCode">外字コード。</param>
		/// <returns>ファイル先頭からの位置。</returns>
		private static long GetIvsIndexRecordOffset(int eudcCode)
		{
			return EudcIvsTableOffset + 2 * (eudcCode - 0xE000);
		}

		/// <summary>
		/// IVS 文字情報レコードの位置を取得します。
		/// </summary>
		/// <param name="recordIndex">レコード番号。</param>
		/// <returns>ファイル先頭からの位置。</returns>
		private static long GetIvsInfoRecordOffset(int recordIndex)
		{
			return IvsInfoTableOffset + IvsInfoRecordLength * recordIndex;
		}

		/// <summary>
		/// 指定の外字コードに関連付けられている IVS 文字情報レコードの位置を取得します。
		/// </summary>
		/// <param name="reader">外字フォント情報ファイルの読み込み用ストリーム。</param>
		/// <param name="eudcCode">外字コード。</param>
		/// <returns>ファイル先頭からの位置。</returns>
		private static long GetIvsInfoRecordOffset(BinaryReader reader, int eudcCode)
		{
			reader.BaseStream.Position = GetIvsIndexRecordOffset(eudcCode);
			return GetIvsInfoRecordOffset(reader.ReadUInt16());
		}

		/// <summary>
		/// 割り当て可能な外字コードの中から割り当てされていない外字コードを検索します。
		/// </summary>
		/// <param name="eudcFontDataStream">外字フォント情報ファイルストリーム。</param>
		/// <returns>割り当て可能な外字コード。存在しない場合には -1。</returns>
		private static int FindEmptyEudcCode(Stream eudcFontDataStream)
		{
			int eudcCode = -1;
			byte flag = 0;
			int segment = -1;

			// 外字割り当てフラグ
			byte[] eudcFlags = new byte[EudcFlagsTableLength];
			eudcFontDataStream.Position = EudcFlagsTableOffset;
			eudcFontDataStream.Read(eudcFlags, 0, eudcFlags.Length);

			// 空き外字コードの検索
			foreach (CodeRange range in FontManager.GetEudcRange())
			{
				int offsetS = (range.StartCharCode - 0xE000);
				int segmentS = offsetS / 8;
				byte maskS = EudcFlagMasks[offsetS & 7];

				int offsetE = (range.EndCharCode - 0xE000);
				int segmentE = offsetE / 8;
				byte maskE = EudcFlagMasks[8 + (offsetE & 7)];

				if (segmentS == segmentE)
				{
					// コード範囲が同じセグメント内の場合の特別処理
					int mask = maskS & maskE;
					flag = (byte)(eudcFlags[segmentS] & mask);

					if (flag != mask)
					{
						flag = (byte)(~flag & mask);
						segment = segmentS;
						break;
					}
				}
				else
				{
					// 先頭
					{
						flag = (byte)(eudcFlags[segmentS] & maskS);

						if (flag != maskS)
						{
							flag = (byte)(~flag & maskS);
							segment = segmentS;
							break;
						}
					}

					// 中間
					{
						for (int i=segmentS+1; i<segmentE; i++)
						{
							if (eudcFlags[i] != 0xFF)
							{
								flag = (byte)(~eudcFlags[i]);
								segment = i;
								break;
							}
						}
						if (segment != -1) break;
					}

					// 末尾
					{
						flag = (byte)(eudcFlags[segmentE] & maskE);

						if (flag != maskE)
						{
							flag = (byte)(~flag & maskE);
							segment = segmentE;
							break;
						}
					}
				}				
			}

			if (segment != -1)
			{
				for (int i=0; i<8; i++)
				{
					if ((flag & EudcFlagMasks[8+i]) != 0)
					{
						eudcCode = 0xE000 + (8 * segment) + i;

						// 外字割り当てフラグの更新
						eudcFlags[segment] |= (byte)(1 << (7 - i));
						eudcFontDataStream.Position = EudcFlagsTableOffset + segment;
						eudcFontDataStream.Write(eudcFlags, segment, 1);

						break;
					}
				}
			}

			return eudcCode;
		}

		/// <summary>
		/// 外字割り当てされている IVS 文字の中で再割り当て可能な一番利用されていない文字を取得します。
		/// </summary>
		/// <param name="reader">外字フォント情報ファイル。</param>
		/// <returns>一番利用されていない IVS 文字。再割り当て可能な文字が存在しない場合には -1。</returns>
		private static long FindMostUnusualIvsCode(BinaryReader reader)
		{
			long ivsCode = -1;
			uint minCount = UInt32.MaxValue;
			long minLastUsedTime = DateTime.MaxValue.ToBinary();

			foreach (CodeRange range in FontManager.GetEudcRange())
			{
				for (int code=range.StartCharCode; code<=range.EndCharCode; code++)
				{
					// eudc->ivs 変換テーブル
					reader.BaseStream.Position = GetIvsIndexRecordOffset(code);
					int index = reader.ReadUInt16();

					// ivs 文字情報テーブル
					long ivsRecordPos = GetIvsInfoRecordOffset(index);
					reader.BaseStream.Position = ivsRecordPos + 10;

					// 現在、参照中のものは置き換えできない
					short refCount = reader.ReadInt16();

					if (refCount != 0)
					{
						continue;
					}

					// 使用回数
					uint count = reader.ReadUInt32();

					if (count > minCount)
					{
						continue;
					}

					// 最終使用日時
					long lastUsedTime = reader.ReadInt64();

					if (lastUsedTime > minLastUsedTime)
					{
						continue;
					}

					// 現時点での最適な候補
					minCount = count;
					minLastUsedTime = lastUsedTime;
					reader.BaseStream.Position = ivsRecordPos;
					ivsCode = CodePoint.ToIvsCode(reader.ReadInt32(), reader.ReadInt32());
				}
			}

			return ivsCode;
		}

		/// <summary>
		/// 既定の IVS 文字情報レコードを出力します。
		/// </summary>
		/// <param name="writer">出力先ストリーム。</param>
		/// <param name="baseChar">基底文字 Unicode コードポイント。</param>
		/// <param name="varSelector">字形選択肢 Unicode コードポイント。</param>
		private static void WriteDefaultIvsInfoRecord(BinaryWriter writer, int baseChar, int varSelector)
		{
			// ivs 文字コード
			writer.Write(baseChar);
			writer.Write(varSelector);

			// eudc コード、参照カウント、利用回数、最終利用日時
			writer.Write((ushort)0);
			writer.Write((short)0);
			writer.Write((uint)0);
			writer.Write((long)0);
		}

		/// <summary>
		/// 終端 IVS 文字情報レコードを出力します。
		/// </summary>
		/// <param name="writer">出力先ストリーム。</param>
		private static void WriteTerminationIvsInfoRecord(BinaryWriter writer)
		{
			// ivs 文字コード
			writer.Write(-1);
			writer.Write(-1);

			// eudc コード、参照カウント、利用回数、最終利用日時
			writer.Write((ushort)0);
			writer.Write((short)0);
			writer.Write((uint)0);
			writer.Write((long)0);			
		}

		/// <summary>
		/// 外字（EUDC）の有効・無効を切り替えます。
		/// </summary>
		/// <param name="enable">外字を有効にする場合には true、それ以外の場合には false。</param>
		/// <returns>処理成功時には true、それ以外の場合には false。</returns>
		[DllImport("GDI32")]
		[return:MarshalAs(UnmanagedType.Bool)]
		private static extern bool EnableEUDC(bool enable);

		#endregion
	}
}
