﻿using System;


namespace OfficeIVS
{

	/// <summary>
	/// Unicode コードポイント、UTF-16 文字列の相互変換機能を提供します。
	/// </summary>
	internal static class CodePoint
	{
		#region メソッド

		/// <summary>
		/// 指定された文字が外字文字（Unicode PUA 領域文字）かどうかを検証します。
		/// </summary>
		/// <param name="ch">検証対象文字。</param>
		/// <returns>外字文字の場合には true、それ以外の場合には false。</returns>
		public static bool IsEudcCode(int ch)
		{
			return 0xE000 <= ch && ch <= 0xF8FF;
		}

		/// <summary>
		/// 指定された文字列の先頭文字が外字文字（Unicode PUA 領域文字）かどうかを検証します。
		/// </summary>
		/// <param name="s">検証対象文字列。</param>
		/// <returns>外字文字の場合には true、それ以外の場合には false。</returns>
		public static bool IsEudcCode(string s)
		{
			if (String.IsNullOrEmpty(s))
			{
				return false;
			}
			return (s[0] - 0xE000) <= 0x18FF;
		}

		/// <summary>
		/// コードポイントを Unicode 文字列に変換します。
		/// </summary>
		/// <param name="bc">基底文字。</param>
		/// <param name="vs">異体字セレクタ。</param>
		/// <returns>文字列。</returns>
		public static string FromCodePoint(int bc, int vs)
		{
			int index = 0;
			
			char[] code = new char[4];

			if (bc <= 0xFFFF)
			{
				code[index++] = Convert.ToChar(bc);
			}
			else
			{
				bc -= 0x10000;
				code[index++] = Convert.ToChar(0xD800 + (bc >> 10));
				code[index++] = Convert.ToChar(0xDC00 + (bc & 0x03FF));
			}

			if (0xE0100 <= vs && vs <= 0xE01EF)
			{
				vs -= 0x10000;
				code[index++] = Convert.ToChar(0xD800 + (vs >> 10));
				code[index++] = Convert.ToChar(0xDC00 + (vs & 0x03FF));
			}

			return new string(code, 0, index);	
		}

		/// <summary>
		/// 異体字セレクタ付き文字コードを Unicode 文字列に変換します。
		/// </summary>
		/// <param name="codePoint">異体字セレクタ付き文字コード。</param>
		/// <returns>文字列。</returns>
		public static string FromIvsCodePoint(long codePoint)
		{
			if (codePoint < 0x100000000)
			{
				return FromCodePoint(unchecked ((int)codePoint), 0);
			}
			return FromCodePoint(unchecked ((int)(codePoint >> 32)),
									unchecked ((int)(codePoint & 0xFFFFFFFF)));
		}

		/// <summary>
		/// 指定文字を Unicode コードポイントに変換します。
		/// </summary>
		/// <param name="s">文字列。</param>
		/// <param name="index">変換対象文字の位置。</param>
		/// <returns>指定文字の Unicode コードポイント。変換できない場合には -1。</returns>
		public static int ToCodePoint(string s, int index=0)
		{
			int codePoint = -1;

			if (!string.IsNullOrEmpty(s) && index < s.Length)
			{
				if (!Char.IsSurrogate(s[index]))
				{
					codePoint = s[index];
				}
				else if (Char.IsSurrogatePair(s, index))
				{
					codePoint = 0x10000 + ((s[index] - 0xD800) << 10) + (s[index+1] - 0xDC00);
				}
			}

			return codePoint;
		}

		/// <summary>
		/// 指定文字を Unicode コードポイント（異体字セレクタ対応）に変換します。
		/// </summary>
		/// <param name="s">文字列。</param>
		/// <param name="index">変換対象文字の位置。</param>
		/// <returns>指定文字の Unicode コードポイント。変換できない場合には -1。</returns>
		public static long ToIvsCodePoint(string s, int index=0)
		{
			long codePoint = ToCodePoint(s, index);

			if (codePoint > 0)
			{
				int vs = ToCodePoint(s, index + codePoint <= 0xFFFF ? 1 : 2);

				if (vs > 0 && (vs - 0xE0100) <= 0x000EF)
				{
					codePoint <<= 32;
					codePoint += vs;
				}
			}

			return codePoint;
		}

		/// <summary>
		/// 異体字セレクタ付き文字コードを取得します。
		/// </summary>
		/// <param name="baseChar">基底文字コード。</param>
		/// <param name="varSelector">字形選択肢コード。</param>
		/// <returns>異体字セレクタ付き文字コード。</returns>
		public static long ToIvsCode(int baseChar, int varSelector)
		{
			return (((long)baseChar) << 32) | unchecked ((uint)varSelector);
		}

		#endregion
	}
}
