﻿using System;
using System.Diagnostics;
using System.IO;


namespace OfficeIVS.Font
{

	/// <summary>
	/// バイナリストリーム（ビッグエンディアン）からプリミティブ データ型を読み取ります。
	/// </summary>
	internal class BigEndianBinaryReader : IDisposable
	{
		#region フィールド

		/// <summary>
		/// 読み取り元ストリーム。
		/// </summary>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private readonly Stream _input;

		/// <summary>
		/// 読み取り元ストリームを閉じるかどうか。
		/// </summary>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private readonly bool _closeInput;

		/// <summary>
		/// 読み取り用バッファ。
		/// </summary>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private readonly byte[] _buffer = new byte[16];

		#endregion

		#region コンストラクタ

		/// <summary>
		/// 指定したストリームを基にし <c>BinaryReader</c> クラスの新しいインスタンスを初期化します。
		/// </summary>
		/// <param name="input">読み込み元のストリーム。</param>
		/// <param name="closeInput">終了時に読み取り元ストリームを閉じるかどうか。</param>
		/// <exception cref="ArgumentException"><paramref name="input"/> が読み取りをサポートしていない、もしくは既に閉じられている。</exception>
		/// <exception cref="ArgumentNullException"><paramref name="input"/> が <c>null</c>。</exception>
		public BigEndianBinaryReader(Stream input, bool closeInput=false)
		{
			if (input == null)
			{
				throw new ArgumentNullException("input");
			}

			if (!input.CanRead)
			{
				throw new ArgumentException("指定ストリームは読み取りをサポートしていません。", "input");
			}

			_input = input;
			_closeInput = closeInput;
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// <see cref="BigEndianBinaryReader"/> に関連付けられた読み込み元のストリームを取得します。
		/// </summary>
		/// <value><see cref="BigEndianBinaryReader"/> に関連付けられたストリーム。</value>
		public Stream BaseStream
		{
			[DebuggerStepThrough]
			get { return _input; }
		}

		/// <summary>
		/// 現在のストリームの読み取り位置を取得または設定します。
		/// </summary>
		/// <value>現在のストリームの読み取り位置。</value>
		public long Position
		{
			[DebuggerStepThrough]
			get { return _input.Position; }

			[DebuggerStepThrough]
			set { _input.Position = value; }
		}

		#endregion

		#region メソッド

		/// <summary>
		/// 現在のリーダーと基になるストリームをクローズします。
		/// </summary>
		public void Close()
		{
			if (_closeInput) _input.Close();
		}

		/// <summary>
		/// <see cref="BigEndianBinaryReader"/> クラスの現在のインスタンスによって使用されているすべてのリソースを解放します。
		/// </summary>
		public void Dispose()
		{
			if (_closeInput) _input.Dispose();
		}

		/// <summary>
		/// ストリームから指定されたバイト数を読み取り、バイト配列内の指定された位置に書き込みます。 
		/// </summary>
		/// <param name="buffer">データを受け取るバッファー。</param>
		/// <param name="index">バッファー内の書き込み開始位置。 </param>
		/// <param name="count">読み取るバイト数。</param>
		/// <returns>
		/// <paramref name="buffer"/> に書き込んだバイト数。
		/// 要求したバイト数分のバイトを読み取れなかった場合、この値は要求したバイト数より小さくなります。
		/// ストリームの末尾に到達した場合は 0 になることがあります。 
		/// </returns>
		/// <exception cref="ArgumentException">バッファー長から <paramref name="index"/> を差し引いた値が <paramref name="count"/> より小さい値です。 </exception>
		/// <exception cref="ArgumentNullException"><paramref name="buffer"/> が null。</exception>
		/// <exception cref="ObjectDisposedException">ストリームが閉じられている。</exception>
		/// <exception cref="IOException">I/O エラーの発生。</exception>
		public int Read(byte[] buffer, int index, int count)
		{
			return _input.Read(buffer, index, count);
		}

		/// <summary>
		/// 現在のストリーム位置から符号付き 8bit 整数値を読み取り、ストリーム位置を 1 バイトだけ進めます。
		/// </summary>
		/// <returns>符号付き 8bit 整数値。</returns>
		/// <exception cref="EndOfStreamException">ストリームの末尾に到達。</exception>
		/// <exception cref="ObjectDisposedException">ストリームが閉じられている。</exception>
		/// <exception cref="IOException">I/O エラーの発生。</exception>
		public byte ReadByte()
		{
			int data = _input.ReadByte();

			if (data == -1)
			{
				throw new EndOfStreamException();
			}

			return Convert.ToByte(data);
		}

		/// <summary>
		/// 現在のストリーム位置から符号付き 8bit 整数値を読み取り、ストリーム位置を 1 バイトだけ進めます。
		/// </summary>
		/// <returns>符号付き 8bit 整数値。</returns>
		/// <exception cref="EndOfStreamException">ストリームの末尾に到達。</exception>
		/// <exception cref="ObjectDisposedException">ストリームが閉じられている。</exception>
		/// <exception cref="IOException">I/O エラーの発生。</exception>
		public sbyte ReadInt8()
		{
			return unchecked((sbyte)ReadByte());
		}

		/// <summary>
		/// 現在のストリーム位置から符号付き 16bit 整数値を読み取り、ストリーム位置を 2 バイトだけ進めます。
		/// </summary>
		/// <returns>符号付き 16bit 整数値。</returns>
		/// <exception cref="EndOfStreamException">ストリームの末尾に到達。</exception>
		/// <exception cref="ObjectDisposedException">ストリームが閉じられている。</exception>
		/// <exception cref="IOException">I/O エラーの発生。</exception>
		public short ReadInt16()
		{
			return BitConverter.ToInt16(FillBuffer(2), 0);
		}

		/// <summary>
		/// 現在のストリーム位置から符号付き 24bit 整数値を読み取り、ストリーム位置を 3 バイトだけ進めます。
		/// </summary>
		/// <returns>符号付き 24bit 整数値。</returns>
		/// <exception cref="EndOfStreamException">ストリームの末尾に到達。</exception>
		/// <exception cref="ObjectDisposedException">ストリームが閉じられている。</exception>
		/// <exception cref="IOException">I/O エラーの発生。</exception>
		public int ReadInt24()
		{	
			return BitConverter.ToInt32(FillBuffer(3, 1), 0);
		}

		/// <summary>
		/// 現在のストリーム位置から符号付き 32bit 整数値を読み取り、ストリーム位置を 4 バイトだけ進めます。
		/// </summary>
		/// <returns>符号付き 32bit 整数値。</returns>
		/// <exception cref="EndOfStreamException">ストリームの末尾に到達。</exception>
		/// <exception cref="ObjectDisposedException">ストリームが閉じられている。</exception>
		/// <exception cref="IOException">I/O エラーの発生。</exception>
		public int ReadInt32()
		{	
			return BitConverter.ToInt32(FillBuffer(4), 0);
		}

		/// <summary>
		/// 現在のストリーム位置から符号付き 64bit 整数値を読み取り、ストリーム位置を 8 バイトだけ進めます。
		/// </summary>
		/// <returns>符号付き 64bit 整数値。</returns>
		/// <exception cref="EndOfStreamException">ストリームの末尾に到達。</exception>
		/// <exception cref="ObjectDisposedException">ストリームが閉じられている。</exception>
		/// <exception cref="IOException">I/O エラーの発生。</exception>
		public long ReadInt64()
		{
			return BitConverter.ToInt64(FillBuffer(8), 0);
		}

		/// <summary>
		/// 現在のストリーム位置から符号付き 8bit 整数値を読み取り、ストリーム位置を 1 バイトだけ進めます。
		/// </summary>
		/// <returns>符号付き 8bit 整数値。</returns>
		/// <exception cref="EndOfStreamException">ストリームの末尾に到達。</exception>
		/// <exception cref="ObjectDisposedException">ストリームが閉じられている。</exception>
		/// <exception cref="IOException">I/O エラーの発生。</exception>
		public byte ReadUInt8()
		{
			return ReadByte();
		}

		/// <summary>
		/// 現在のストリーム位置から符号無し 16bit 整数値を読み取り、ストリーム位置を 2 バイトだけ進めます。
		/// </summary>
		/// <returns>符号無し 16bit 整数値。</returns>
		/// <exception cref="EndOfStreamException">ストリームの末尾に到達。</exception>
		/// <exception cref="ObjectDisposedException">ストリームが閉じられている。</exception>
		/// <exception cref="IOException">I/O エラーの発生。</exception>
		public ushort ReadUInt16()
		{
			return BitConverter.ToUInt16(FillBuffer(2), 0);
		}

		/// <summary>
		/// 現在のストリーム位置から符号無し 24bit 整数値を読み取り、ストリーム位置を 3 バイトだけ進めます。
		/// </summary>
		/// <returns>符号無し 24bit 整数値。</returns>
		/// <exception cref="EndOfStreamException">ストリームの末尾に到達。</exception>
		/// <exception cref="ObjectDisposedException">ストリームが閉じられている。</exception>
		/// <exception cref="IOException">I/O エラーの発生。</exception>
		public uint ReadUInt24()
		{	
			return BitConverter.ToUInt32(FillBuffer(3, 1), 0);
		}

		/// <summary>
		/// 現在のストリーム位置から符号無し 32bit 整数値を読み取り、ストリーム位置を 4 バイトだけ進めます。
		/// </summary>
		/// <returns>符号無し 32bit 整数値。</returns>
		/// <exception cref="EndOfStreamException">ストリームの末尾に到達。</exception>
		/// <exception cref="ObjectDisposedException">ストリームが閉じられている。</exception>
		/// <exception cref="IOException">I/O エラーの発生。</exception>
		public uint ReadUInt32()
		{	
			return BitConverter.ToUInt32(FillBuffer(4), 0);
		}

		/// <summary>
		/// 現在のストリーム位置から符号無し 64bit 整数値を読み取り、ストリーム位置を 8 バイトだけ進めます。
		/// </summary>
		/// <returns>符号無し 64bit 整数値。</returns>
		/// <exception cref="EndOfStreamException">ストリームの末尾に到達。</exception>
		/// <exception cref="ObjectDisposedException">ストリームが閉じられている。</exception>
		/// <exception cref="IOException">I/O エラーの発生。</exception>
		public ulong ReadUInt64()
		{
			return BitConverter.ToUInt64(FillBuffer(8), 0);
		}

		/// <summary>
		/// ストリームから指定したバイト数分だけ読み取り内部バッファーに格納します。
		/// </summary>
		/// <param name="numBytes">読み取るバイト数。</param>
		/// <param name="offset">内部バッファーの書き込み開始位置。</param>
		/// <returns>内部バッファ。</returns>
		/// <exception cref="ArgumentOutOfRangeException"><paramref name="numBytes"/> が、内部バッファー サイズより大きい。</exception>
		/// <exception cref="EndOfStreamException"><paramref name="numBytes"/> 読み取る前にストリーム末尾に到達。 </exception>
		/// <exception cref="IOException">I/O エラーの発生。</exception>
		private byte[] FillBuffer(int numBytes, int offset=0)
		{
			// 引数検証
			if (_buffer.Length < offset + numBytes)
			{
				throw new ArgumentOutOfRangeException();
			}

			// オフセット分クリア
			for (int i=0; i<offset; i++)
			{
				_buffer[i] = 0;
			}

			// データ取得
			int readBytes = _input.Read(_buffer, offset, numBytes);

			if (readBytes != numBytes)
			{
				throw new EndOfStreamException();
			}

			// エンディアン変換
			Array.Reverse(_buffer, 0, offset+numBytes);

			return _buffer;
		}

		#endregion
	}
}
