﻿using System;
using System.Diagnostics;
using System.IO;


namespace OfficeIVS.Font
{

	/// <summary>
	/// プリミティブ データ型をバイナリストリーム（ビッグエンディアン）に書き込みます。
	/// </summary>
	internal class BigEndianBinaryWriter : IDisposable
	{
		#region フィールド

		/// <summary>
		/// 書き込み用バイナリライター。
		/// </summary>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private readonly Stream _output;

		/// <summary>
		/// 終了時に書き込み先のストリームを閉じるかどうか。
		/// </summary>
		private readonly bool _closeOutput;

		#endregion

		#region コンストラクタ

		/// <summary>
		/// 指定したストリームを基にし <c>BigEndianBinaryWriter</c> クラスの新しいインスタンスを初期化します。
		/// </summary>
		/// <param name="output">書き込み先のストリーム。</param>
		/// <param name="closeOutput">終了時に書き込み先のストリームを閉じるかどうか。</param>
		/// <exception cref="ArgumentException"><paramref name="output"/> が書き込みをサポートしていないか、既に閉じられている。</exception>
		/// <exception cref="ArgumentNullException"><paramref name="output"/> が null。</exception>
		public BigEndianBinaryWriter(Stream output, bool closeOutput=false)
		{
			if (output == null)
			{
				throw new ArgumentNullException("output");
			}

			if (!output.CanWrite)
			{
				throw new ArgumentException("ストリームが書き込みをサポートしていません。", "output");
			}

			_output = output;
			_closeOutput = closeOutput;
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// <see cref="BigEndianBinaryWriter"/> に関連付けられた読み込み元のストリームを取得します。
		/// </summary>
		/// <value><see cref="BigEndianBinaryWriter"/> に関連付けられたストリーム。</value>
		public Stream BaseStream
		{
			[DebuggerStepThrough]
			get { return _output; }
		}

		/// <summary>
		/// 現在のストリームの読み取り位置を取得または設定します。
		/// </summary>
		/// <value>現在のストリームの読み取り位置。</value>
		public long Position
		{
			[DebuggerStepThrough]
			get { return _output.Position; }

			[DebuggerStepThrough]
			set { _output.Position = value; }
		}

		#endregion

		#region メソッド

		/// <summary>
		/// 現在のリーダーと基になるストリームをクローズします。
		/// </summary>
		public void Close()
		{
			if (_closeOutput) _output.Close();
		}

		/// <summary>
		/// <see cref="BigEndianBinaryWriter"/> クラスの現在のインスタンスによって使用されているすべてのリソースを解放します。
		/// </summary>
		public void Dispose()
		{
			if (_closeOutput) _output.Dispose();
		}

		/// <summary>
		/// 現在のストリーム位置に符号付き 8bit 整数値を書き込み、ストリーム位置を 1 バイトだけ進めます。
		/// </summary>
		/// <param name="value">符号付き 8bit 整数値。</param>
		/// <exception cref="ObjectDisposedException">ストリームが閉じられている。</exception>
		/// <exception cref="IOException">I/O エラーの発生。</exception>
		public void Write(sbyte value)
		{
			_output.WriteByte(unchecked((byte)value));
		}

		/// <summary>
		/// 現在のストリーム位置に符号付き 16bit 整数値を書き込み、ストリーム位置を 2 バイトだけ進めます。
		/// </summary>
		/// <param name="value">符号付き 16bit 整数値。</param>
		/// <exception cref="ObjectDisposedException">ストリームが閉じられている。</exception>
		/// <exception cref="IOException">I/O エラーの発生。</exception>
		public void Write(short value)
		{
			InternalWrite(BitConverter.GetBytes(value), 2);
		}

		/// <summary>
		/// 現在のストリーム位置に符号付き 32bit 整数値を書き込み、ストリーム位置を 4 バイトだけ進めます。
		/// </summary>
		/// <param name="value">符号付き 32bit 整数値。</param>
		/// <exception cref="ObjectDisposedException">ストリームが閉じられている。</exception>
		/// <exception cref="IOException">I/O エラーの発生。</exception>
		public void Write(int value)
		{
			InternalWrite(BitConverter.GetBytes(value), 4);
		}

		/// <summary>
		/// 現在のストリーム位置に符号付き 64bit 整数値を書き込み、ストリーム位置を 8 バイトだけ進めます。
		/// </summary>
		/// <param name="value">符号付き 64bit 整数値。</param>
		/// <exception cref="ObjectDisposedException">ストリームが閉じられている。</exception>
		/// <exception cref="IOException">I/O エラーの発生。</exception>
		public void Write(long value)
		{
			InternalWrite(BitConverter.GetBytes(value), 8);
		}

		/// <summary>
		/// 現在のストリーム位置に符号無し 8bit 整数値を書き込み、ストリーム位置を 1 バイトだけ進めます。
		/// </summary>
		/// <param name="value">符号無し 8bit 整数値。</param>
		/// <exception cref="ObjectDisposedException">ストリームが閉じられている。</exception>
		/// <exception cref="IOException">I/O エラーの発生。</exception>
		public void Write(byte value)
		{
			_output.WriteByte(value);
		}

		/// <summary>
		/// 現在のストリーム位置に符号無し 16bit 整数値を書き込み、ストリーム位置を 2 バイトだけ進めます。
		/// </summary>
		/// <param name="value">符号無し 16bit 整数値。</param>
		/// <exception cref="ObjectDisposedException">ストリームが閉じられている。</exception>
		/// <exception cref="IOException">I/O エラーの発生。</exception>
		public void Write(ushort value)
		{
			InternalWrite(BitConverter.GetBytes(value), 2);
		}

		/// <summary>
		/// 現在のストリーム位置に符号無し 32bit 整数値を書き込み、ストリーム位置を 4 バイトだけ進めます。
		/// </summary>
		/// <param name="value">符号無し 32bit 整数値。</param>
		/// <exception cref="ObjectDisposedException">ストリームが閉じられている。</exception>
		/// <exception cref="IOException">I/O エラーの発生。</exception>
		public void Write(uint value)
		{
			InternalWrite(BitConverter.GetBytes(value), 4);
		}

		/// <summary>
		/// 現在のストリーム位置に符号無し 64bit 整数値を書き込み、ストリーム位置を 8 バイトだけ進めます。
		/// </summary>
		/// <param name="value">符号無し 64bit 整数値。</param>
		/// <exception cref="ObjectDisposedException">ストリームが閉じられている。</exception>
		/// <exception cref="IOException">I/O エラーの発生。</exception>
		public void Write(ulong value)
		{
			InternalWrite(BitConverter.GetBytes(value), 8);
		}

		/// <summary>
		/// 現在のストリーム位置に指定バイト書き込み、ストリーム位置をバイト数分だけ進めます。
		/// </summary>
		/// <remarks>
		/// 本メソッドは、引数 <paramref name="values"/> の値をそのまま書き込みます。
		/// エンディアン変換は行いません。
		/// </remarks>
		/// <param name="values">書き込み元のデータバッファ。</param>
		/// <exception cref="ObjectDisposedException">ストリームが閉じられている。</exception>
		/// <exception cref="IOException">I/O エラーの発生。</exception>
		public void Write(byte[] values)
		{
			_output.Write(values, 0, values.Length);
		}

		/// <summary>
		/// 現在のストリーム位置に指定バイト書き込み、ストリーム位置をバイト数分だけ進めます。
		/// </summary>
		/// <remarks>
		/// 本メソッドは、引数 <paramref name="values"/> の値をそのまま書き込みます。
		/// エンディアン変換は行いません。
		/// </remarks>
		/// <param name="values">書き込み元のデータバッファ。</param>
		/// <param name="index">データバッファ内における書き込み開始位置。</param>
		/// <param name="count">書き込みバイト数。</param>
		/// <exception cref="ObjectDisposedException">ストリームが閉じられている。</exception>
		/// <exception cref="IOException">I/O エラーの発生。</exception>
		public void Write(byte[] values, int index, int count)
		{
			_output.Write(values, index, count);
		}

		/// <summary>
		/// 現在のストリーム位置に符号無し 24bit 整数値を書き込み、ストリーム位置を 3 バイトだけ進めます。
		/// </summary>
		/// <param name="value">符号無し 24bit 整数値。</param>
		/// <exception cref="ObjectDisposedException">ストリームが閉じられている。</exception>
		/// <exception cref="IOException">I/O エラーの発生。</exception>
		public void WriteUInt24(uint value)
		{
			InternalWrite(BitConverter.GetBytes(value & 0x00FFFFFF), 3);
		}

		/// <summary>
		/// 現在のストリーム位置に指定バイト書き込み、ストリーム位置をバイト数分だけ進めます。
		/// </summary>
		/// <remarks>
		/// 本メソッドは、引数 <paramref name="values"/> の値をエンディアン変換したうえで書き込みます。
		/// </remarks>
		/// <param name="values">書き込み元のデータバッファ。</param>
		/// <param name="length">書き込みバイト数。</param>
		/// <exception cref="ObjectDisposedException">ストリームが閉じられている。</exception>
		/// <exception cref="IOException">I/O エラーの発生。</exception>
		private void InternalWrite(byte[] values, int length)
		{
			// エンディアン変換
			Array.Reverse(values, 0, length);

			// 書き込み
			_output.Write(values, 0, length);
		}

		#endregion
	}
}
