﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.IO.MemoryMappedFiles;
using System.Threading;
using System.Windows;
using System.Windows.Media;

using OfficeIVS.Font.Table;


namespace OfficeIVS.Font
{

	/// <summary>
	/// OpenType フォントのファイル形式を表します。
	/// </summary>
	internal static class SfntVersion
	{
		#region 定数フィールド

		/// <summary>
		/// フォントファイル形式は Compact Font Format です。
		/// </summary>
		public const uint CFF = 0x4F54544F;

		/// <summary>
		/// フォントファイル形式は TrueType Collection です。
		/// </summary>
		public const uint TTC = 0x74746366;

		/// <summary>
		/// フォントファイル形式は TrueType です。
		/// </summary>
		public const uint TTF = 0x00010000;

		#endregion
	}



	/// <summary>
	/// OpenType フォント情報を表します。
	/// </summary>
	internal class OpenTypeFont : IDisposable
	{
		#region フィールド

		/// <summary>
		/// オブジェクト解放済みかどうかのフラグ
		/// </summary>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private bool _disposed;

		/// <summary>
		/// 操作対象のフォントファイル。
		/// </summary>
		private MemoryMappedFile _fontFile;

		/// <summary>
		/// OpenType フォント テーブルレコード コレクション。
		/// </summary>
		private readonly Dictionary<uint, ITableRecord> _tableRecords;

		/// <summary>
		/// OpenType フォント テーブル コレクション。
		/// </summary>
		/// <remarks>キャッシュ用。</remarks>
		private readonly Dictionary<uint, OpenTypeFontTable> _tables;

		#endregion

		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="OpenTypeFont"/> クラスのインスタンスを初期化します。
		/// </summary>
		private OpenTypeFont()
		{
			_fontFile = null;
			_tableRecords = new Dictionary<uint, ITableRecord>();
			_tables = new Dictionary<uint, OpenTypeFontTable>();
		}

		/// <summary>
		/// 新しい <see cref="OpenTypeFont"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="version">sfnt バージョン番号。</param>
		/// <exception cref="NotSupportedException">指定された snft バージョン形式をサポートしていない。</exception>
		public OpenTypeFont(uint version) : this()
		{
			if (version != Font.SfntVersion.CFF && version != Font.SfntVersion.TTF)
			{
				throw new NotSupportedException();
			}
			SfntVersion = version;
		}

		/// <summary>
		/// 新しい <see cref="OpenTypeFont"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="fontUri">読み込むフォントの URI（ローカルファイルのみ対応）。</param>
		/// <exception cref="ArgumentNullException"><paramref name="fontUri"/> が null。</exception>
		/// <exception cref="NotSupportedException">対象フォントがファイル形式ではない。</exception>
		/// <exception cref="FileFormatException">対象フォントが未対応のフォント形式。</exception>
		private OpenTypeFont(Uri fontUri) : this()
		{
			if (fontUri == null)
			{
				throw new ArgumentNullException("fontUri");
			}

			// フォントがファイルでない場合にはサポートしない
			if (!fontUri.IsFile)
			{
				throw new NotSupportedException();
			}

			FontUri = fontUri;

			// メモリマップドファイル作成
			_fontFile = MemoryMappedFile.CreateFromFile(
				new FileStream(fontUri.LocalPath, FileMode.Open, FileAccess.Read, FileShare.Read),
				fontUri.PathAndQuery,
				0,
				MemoryMappedFileAccess.Read,
				null,
				HandleInheritability.None,
				false);

			// 対象フォントの OffsetTable 位置の取得
			long offsetTableOffset;

			using (Stream fontStream = _fontFile.CreateViewStream(0, 1024, MemoryMappedFileAccess.Read))
			using (BigEndianBinaryReader reader = new BigEndianBinaryReader(fontStream))
			{
				// 対象フォントの TrueType Collection 内での位置取得
				//  : TrueType Collection 内でのインデックス番号は
				//    GlyphTypeface の FontUri の Fragment に 「#(数値)」形式で格納されている
				int fontIndex = 0;

				if (!string.IsNullOrEmpty(fontUri.Fragment))
				{
					Int32.TryParse(fontUri.Fragment.Remove(0, 1), out fontIndex);
				}

				offsetTableOffset = GetOffsetTableOffset(reader, fontIndex);

				if (offsetTableOffset == -1)
				{
					throw new FileFormatException(fontUri, "未対応のフォント形式です。");
				}
			}

			// OffsetTable の読み込み
			using (Stream fontStream = _fontFile.CreateViewStream(offsetTableOffset, 1024, MemoryMappedFileAccess.Read))
			using (BigEndianBinaryReader reader = new BigEndianBinaryReader(fontStream))
			{
				// sfnt バージョン番号
				SfntVersion = reader.ReadUInt32();

				// テーブル数
				ushort numTables = reader.ReadUInt16();

				// searchRange, entrySelector, rangeShift は使用しないので読み飛ばし
				reader.Position += 6;

				// TableRecord 読み込み
				for (int i=0; i<numTables; i++)
				{
					uint tagValue = reader.ReadUInt32();

					_tableRecords[tagValue] = new ReadOnlyTableRecord(_fontFile,
																	  tagValue,
																	  reader.ReadUInt32(),
																	  reader.ReadUInt32(),
																	  reader.ReadUInt32());
				}
			}
		}

		/// <summary>
		/// ファイナライザ
		/// </summary>
		~OpenTypeFont()
		{
			Dispose(false);
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// オブジェクト解放済みかどうか
		/// </summary>
		public bool IsDisposed
		{
			[DebuggerStepThrough]
			get { return _disposed; }
		}

		/// <summary>
		/// フォントの読み込み元 URI を取得します。
		/// </summary>
		/// <value>フォントの読み込み元 URI。</value>
		public Uri FontUri { get; private set; }
		
		/// <summary>
		/// sfnt バージョン番号を取得または設定します。
		/// </summary>
		/// <value>sfnt バージョン番号。</value>
		public uint SfntVersion { get; private set; }

		/// <summary>
		/// OpenType フォントの各テーブルレコードを取得または設定します。
		/// </summary>
		/// <param name="tag">テーブル識別子。</param>
		/// <returns>テーブルレコード。存在しない場合には <c>null</c>。</returns>
		public ITableRecord this[Tag tag]
		{
			get
			{
				ITableRecord tableRecord;

				_tableRecords.TryGetValue(tag.Value, out tableRecord);

				return tableRecord;
			}

			set
			{
				if (value == null)
				{
					_tableRecords.Remove(tag.Value);
				}
				else
				{
					_tableRecords[tag.Value] = value;	
				}
				_tables.Remove(tag.Value);
			}
		}

		/// <summary>
		/// テーブルレコードの一覧を取得します。
		/// </summary>
		/// <value>テーブルレコード一覧。</value>
		public IEnumerable<ITableRecord> TableRecords
		{
			[DebuggerStepThrough]
			get { return _tableRecords.Values; }
		}

		#endregion

		#region パブリック メソッド

		/// <summary>
		/// 本オブジェクトが使用しているリソースを解放します。
		/// </summary>
		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.OpenTypeFont"/> クラスのインスタンスを作成します。
		/// </summary>
		/// <param name="fontUri">読み込むフォントの URI（ローカルファイルのみ対応）。</param>
		/// <exception cref="ArgumentNullException"><paramref name="fontUri"/> が null。</exception>
		/// <exception cref="NotSupportedException">対象フォントがファイル形式ではない。</exception>
		/// <exception cref="FileFormatException">対象フォントが未対応のフォント形式。</exception>
		public static OpenTypeFont Load(Uri fontUri)
		{
			return new OpenTypeFont(fontUri);
		}
	
		/// <summary>
		/// タイプフェースから新しい <see cref="T:OfficeIVS.Font.OpenTypeFont"/> クラスのインスタンスを作成します。
		/// </summary>
		/// <param name="typeface">読み込むフォントのタイプフェース。</param>
		/// <returns>
		/// 新しい <see cref="T:OfficeIVS.Font.OpenTypeFont"/> クラスのインスタンス。
		/// なお、タイプフェースに一致するフォントが存在しない場合には <c>null</c> を返します。
		/// </returns>
		/// <exception cref="ArgumentNullException"><paramref name="typeface"/> が <c>null</c>。</exception>
		/// <exception cref="NotSupportedException">対象フォントがファイル形式ではない。</exception>
		/// <exception cref="FileFormatException">対象フォントが未対応のフォント形式。</exception>
		public static OpenTypeFont Load(Typeface typeface)
		{
			if (typeface == null)
			{
				throw new ArgumentNullException("typeface");
			}

			GlyphTypeface glyphTypeface;

			if (!typeface.TryGetGlyphTypeface(out glyphTypeface))
			{
				return null;
			}

			return new OpenTypeFont(glyphTypeface.FontUri);
		}

		/// <summary>
		/// 指定フォントファミリを使用して
		/// 新しい <see cref="T:OfficeIVS.Font.OpenTypeFont"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="fontFamily">読み込むフォントのフォントファミリ。</param>
		/// <returns>
		/// 新しい <see cref="T:OfficeIVS.Font.OpenTypeFont"/> クラスのインスタンス。
		/// なお、フォントファミリに一致するフォントが存在しない場合には <c>null</c> を返します。
		/// </returns>
		/// <exception cref="ArgumentNullException"><paramref name="fontFamily"/> が <c>null</c>。</exception>
		/// <exception cref="NotSupportedException">対象フォントがファイル形式ではない。</exception>
		/// <exception cref="FileFormatException">対象フォントが未対応のフォント形式。</exception>
		public static OpenTypeFont Load(FontFamily fontFamily)
		{
			return Load(fontFamily, FontStyles.Normal, FontWeights.Normal, FontStretches.Normal);
		}

		/// <summary>
		/// 指定フォントファミリ、<see cref="T:System.Windows.FontStyle"/>、<see cref="T:System.Windows.FontWeight"/>、
		/// および <see cref="T:System.Windows.FontStretch"/> を使用して、
		/// 新しい <see cref="T:OfficeIVS.Font.OpenTypeFont"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="fontFamily">フォントのフォントファミリ。</param>
		/// <param name="style">フォントのスタイル。</param>
		/// <param name="weight">フォントの太さ。</param>
		/// <param name="stretch">フォントの比率。</param>
		/// <returns>
		/// 新しい <see cref="T:OfficeIVS.Font.OpenTypeFont"/> クラスのインスタンス。
		/// なお、指定条件に一致するフォントが存在しない場合には <c>null</c> を返します。
		/// </returns>
		/// <exception cref="ArgumentNullException"><paramref name="fontFamily"/> が <c>null</c>。</exception>
		/// <exception cref="NotSupportedException">対象フォントがファイル形式ではない。</exception>
		/// <exception cref="FileFormatException">対象フォントが未対応のフォント形式。</exception>
		public static OpenTypeFont Load(FontFamily fontFamily, FontStyle style, FontWeight weight, FontStretch stretch)
		{
			if (fontFamily == null)
			{
				throw new ArgumentNullException("fontFamily");
			}
			return Load(new Typeface(fontFamily, style, weight, stretch));
		}

		/// <summary>
		/// フォントが異体字セレクタに対応しているかを検証します。
		/// </summary>
		/// <returns>異体字セレクタに対応している場合には true、それ以外の場合には false。</returns>
		public bool IsSupportIVS()
		{
			ITableRecord tableRecord = this[TableTags.cmap];

			if (tableRecord == null)
			{
				return false;
			}

			// 異体字セレクタに対応しているかどうかは cmap format14 サブテーブルに
			// 日本語用の異体字セレクタ（U+E0100 - U+E01EF）が存在するかどうかで判断を行う

			using (Stream stream = tableRecord.GetTableStream())
			using (BigEndianBinaryReader reader = new BigEndianBinaryReader(stream))
			{
				ushort version = reader.ReadUInt16();

				if (version != 0)
				{
					return false;
				}

				ushort numTables = reader.ReadUInt16();

				for (int i=0; i<numTables; i++)
				{
					ushort platformId = reader.ReadUInt16();
					ushort encodingId = reader.ReadUInt16();

					// format14 サブテーブルの検証
					if (platformId == 0 && encodingId == 5)
					{
						reader.Position = reader.ReadUInt32() + 6;

						uint numVarSelectorRecords = reader.ReadUInt32();

						for (uint index=0; index<numVarSelectorRecords; index++)
						{
							int varSelector = reader.ReadInt24();

							if (0xE0100 <= varSelector && varSelector <= 0xE01EF)
							{
								return true;
							}

							reader.Position += 8;
						}

						return false;
					}

					reader.Position += 4;
				}
			}

			return false;
		}

		/// <summary>
		/// フォントに指定テーブルレコードを追加します。
		/// </summary>
		/// <param name="tableRecord">テーブルレコード。</param>
		public void AddTableRecord(ITableRecord tableRecord)
		{
			if (tableRecord != null)
			{
				_tableRecords[tableRecord.Tag.Value] = tableRecord;
			}
		}

		/// <summary>
		/// フォントファミリー名を取得します。
		/// </summary>
		/// <returns>フォントファミリー名。未定義の場合には null。</returns>
		/// <remarks>
		/// 複数の言語でフォントファミリが設定されている場合、以下の言語の優先度でフォントファミリ名を返します。<br/>
		///  1. UI スレッドの言語<br/>
		///  2. 英語<br/>
		///  3. 最初に見つかったもの<br/>
		/// </remarks>
		public string GetFontFamilyName()
		{
			string fontFamilyName = null;

			Name00 name = GetTable(TableTags.name) as Name00;

			if (name != null)
			{
				// UI スレッドのコードページで取得
				fontFamilyName = name.GetNameValue(Thread.CurrentThread.CurrentUICulture.Name, 1);

				if (fontFamilyName == null)
				{
					// 英語
					fontFamilyName = name.GetNameValue("en-US", 1);

					if (fontFamilyName == null)
					{
						// 最初に見つかったもの
						fontFamilyName = name.GetNameValue(null, 1);
					}
				}
			}

			return fontFamilyName;
		}

		/// <summary>
		/// 指定の言語のフォントファミリー名を取得します。
		/// </summary>
		/// <param name="languageTag">IETF 言語タグ（RFC 3066）。ja-JPなど。</param>
		/// <returns>指定言語のフォントファミリー名。指定言語のフォントファミリー名が設定されていない場合には null。</returns>
		public string GetFontFamilyName(string languageTag)
		{
			string fontFamilyName = null;

			Name00 name = GetTable(TableTags.name) as Name00;

			if (name != null)
			{
				fontFamilyName = name.GetNameValue(languageTag, 1);
			}

			return fontFamilyName;
		}

		/// <summary>
		/// フォント内の指定テーブルを取得します。
		/// </summary>
		/// <param name="tableTag">取得するテーブル識別子。</param>
		/// <returns>テーブルデータ。存在しない場合には <c>null</c>。</returns>
		public OpenTypeFontTable GetTable(Tag tableTag)
		{
			OpenTypeFontTable table;

			if (!_tables.TryGetValue(tableTag.Value, out table))
			{
				table = OpenTypeFontTable.Load(this, tableTag);

				if (table != null)
				{
					_tables[tableTag.Value] = table;
				}
			}

			return table;
		}

		/// <summary>
		/// 指定ストリームにフォントデータを出力します。
		/// </summary>
		/// <param name="output">出力先ストリーム。</param>
		public void WriteTo(Stream output)
		{
			// テーブルレコードの書き込み順
			List<uint> tableTags = new List<uint>();
			{
				foreach (ITableRecord tableRecord in _tableRecords.Values)
				{
					tableTags.Add(tableRecord.Tag.Value);
				}
				tableTags.Sort();
			}

			// 出力順にテーブル並び替え
			List<ITableRecord> tableRecords = new List<ITableRecord>(_tableRecords.Values);
			tableRecords.Sort();

			// ファイル全体のチェックサム
			uint checkSum = 0;

			// OffsetTable 用バッファ
			int offsetTableSize = 12 + 16 * tableRecords.Count;
			byte[] offsetTable = new byte[offsetTableSize];

			// OpenType フォント出力
			using (BigEndianBinaryWriter header =
				new BigEndianBinaryWriter(new MemoryStream(offsetTable, true), true))
			using (BigEndianBinaryWriter writer = new BigEndianBinaryWriter(output))
			{
				// sfnt version
				header.Write(SfntVersion);

				// numTables, searchRange, entrySelector, rangeShift
				ushort numTables = (ushort)tableRecords.Count;
				ushort searchRange = (ushort)Math.Pow(2, Math.Floor(Math.Log(numTables, 2)));
				header.Write(numTables);
				header.Write((ushort)(16 * searchRange));
				header.Write((ushort)Math.Log(searchRange, 2));
				header.Write((ushort)(16 * (numTables - searchRange)));

				// table
				writer.Position = offsetTableSize;

				foreach (ITableRecord tableRecord in tableRecords)
				{
					// テーブルは常に 4 バイト境界に配置
					writer.Position = (uint)((writer.Position + 3) & ~3);

					// table records
					header.Position = 12 + 16 * tableTags.IndexOf(tableRecord.Tag.Value);
					header.Write(tableRecord.Tag.Value);
					header.Write(tableRecord.CheckSum);
					header.Write((uint)writer.Position);
					header.Write(tableRecord.Length);

					// table
					using (Stream tableStream = tableRecord.GetTableStream())
					{
						tableStream.CopyTo(writer.BaseStream);
					}

					// ファイル全体のチェックサム更新
					checkSum = unchecked (checkSum + tableRecord.CheckSum);
				}

				// OffsetTableのチェックサム計算
				using (BigEndianBinaryReader reader =
					new BigEndianBinaryReader(new MemoryStream(offsetTable, false), true))
				{
					for (int i=0; i<offsetTableSize/4; i++)
					{
						checkSum = unchecked(checkSum + reader.ReadUInt32());
					}
				}

				// OffsetTable 
				writer.Position = 0;
				writer.Write(offsetTable);

				// 'head' テーブルの CheckSumAdjustment を更新
				//  : head テーブルは必ず offset table の次にある
				writer.Position += 8;
				writer.Write(unchecked(0xB1B0AFBA - checkSum));
			}

			output.Seek(0, SeekOrigin.End);
		}

		/// <summary>
		/// 指定されたフォントファイルの 'cmap' テーブルのみを変更します。
		/// </summary>
		/// <param name="fontFileName">変更対象のフォントファイル。</param>
		/// <param name="cmap">新しい cmap テーブル。</param>
		public static void ChangeCMapTable(string fontFileName, CMap cmap)
		{
			using (FileStream fontStream = File.Open(fontFileName, FileMode.Open, FileAccess.ReadWrite, FileShare.None))
			using (BigEndianBinaryReader reader = new BigEndianBinaryReader(fontStream))
			using (BigEndianBinaryWriter writer = new BigEndianBinaryWriter(fontStream))
			{
				// OffsetTable 位置の取得
				long offsetTableOffset = GetOffsetTableOffset(reader);

				if (offsetTableOffset == -1)
				{
					return;
				}

				// OffsetTable 読み取り
				fontStream.Position = offsetTableOffset;
				{
					long headTableOffset = 0;
					uint checkSumAdjust = 0;

					// sfnt version
					reader.Position += 4;

					// テーブル数
					ushort numTables = reader.ReadUInt16();

					// searchRange, entrySelector, rangeShift
					reader.Position += 6;

					// TableRecord 読み込み
					for (int i=0; i<numTables; i++)
					{
						uint tagValue = reader.ReadUInt32();
						uint checkSum = reader.ReadUInt32();
						uint offset = reader.ReadUInt32();
						uint length = reader.ReadUInt32();

						if (tagValue == TableTags.cmap.Value)
						{
							// cmap 更新
							TableRecord tableRecord = new EditableTableRecord(cmap, length);
						
							using (new StreamPositionStack(fontStream, offset))
							using (Stream tableStream = tableRecord.GetTableStream())
							{
								tableStream.CopyTo(writer.BaseStream);
							}

							//  CheckSum 書き換え
							writer.Position -= 12;
							writer.Write((checkSum = tableRecord.CheckSum));
							writer.Position += 8;
						}
						else if (tagValue == TableTags.head.Value)
						{
							headTableOffset = offset;
						}
						else if (tagValue == TableTags.OS2.Value)
						{
							// usFirstCharIndex, usLastCharIndex の更新
							using (new StreamPositionStack(fontStream, offset+64))
							{
								checkSum = unchecked(checkSum - reader.ReadUInt32());

								fontStream.Position -= 4;
								writer.Write(cmap.FirstCharIndex);
								writer.Write(cmap.LastCharIndex);

								fontStream.Position -= 4;
								checkSum = unchecked(checkSum + reader.ReadUInt32());
							}

							//  CheckSum 書き換え
							writer.Position -= 12;
							writer.Write(checkSum);
							writer.Position += 8;
						}

						checkSumAdjust = unchecked (checkSumAdjust + checkSum);							
					}

					// TableRecord のチェックサム再計算
					fontStream.Position = offsetTableOffset;

					for (int i=0; i<(12 + 16*numTables)/4; i++)
					{
						checkSumAdjust = unchecked (checkSumAdjust + reader.ReadUInt32());
					}

					// head テーブルのファイルチェックサム値を更新
					if (headTableOffset != 0)
					{
						fontStream.Position = headTableOffset + 8;
						writer.Write(unchecked(0xB1B0AFBA - checkSumAdjust));
					}
				}
			}			
		}

		#endregion

		#region プロテクト メソッド

		/// <summary>
		/// 本オブジェクトが使用しているリソースを解放します。
		/// </summary>
		/// <param name="disposing">マネージドリソースも解放する場合には true。</param>
		protected virtual void Dispose(bool disposing)
		{
			if (_disposed) return;

			// マネージドリソースの解放
			if (disposing)
			{
				if (_fontFile != null)
				{
					_fontFile.Dispose();
					_fontFile = null;
				}
			}

			// アンマネージドリソースの解放
			{
			}

			_disposed = true;
		}
	
		#endregion

		#region プライベート メソッド

		/// <summary>
		/// ファイルの先頭から OffsetTable までのオフセット量（バイト単位）を取得します。
		/// </summary>
		/// <param name="reader">ストリームリーダー。</param>
		/// <param name="fontIndex">フォント内のフォント番号（TrueType Collection の場合のみ有効）。</param>
		/// <returns>ファイルの先頭から OffsetTable までのオフセット量。未対応の形式の場合には -1。</returns>
		/// <exception cref="ArgumentOutOfRangeException"><paramref name="fontIndex"/> が範囲外。</exception>
		private static long GetOffsetTableOffset(BigEndianBinaryReader reader, int fontIndex=0)
		{
			long offsetTableOffset;

			// OffsetTable 解析
			reader.BaseStream.Seek(0, SeekOrigin.Begin);
			uint sfntVersion = reader.ReadUInt32();

			// TrueType もしくは CFF 形式の場合 OffsetTable 位置はファイルの先頭
			if (sfntVersion == Font.SfntVersion.TTF || sfntVersion == Font.SfntVersion.CFF)
			{
				offsetTableOffset = 0;
			}

			// TrueType Collection 形式の場合 TTC Header を解析し対象フォントの OffsetTable 位置を取得
			else if (sfntVersion == Font.SfntVersion.TTC)
			{
				// バージョン番号はバージョン依存の処理を行わないため無視
				reader.BaseStream.Seek(4, SeekOrigin.Current);

				// TrueType Collection 内のフォント数
				uint numFonts = reader.ReadUInt32();

				if (numFonts <= fontIndex)
				{
					throw new ArgumentOutOfRangeException("fontIndex");
				}

				// OffsetTable から対象フォントの OffsetTable 位置を取得
				reader.BaseStream.Seek(4 * fontIndex, SeekOrigin.Current);
				offsetTableOffset = reader.ReadUInt32();
			}

			// 未対応の形式
			//  : Apple の旧 TrueType 形式など…
			else
			{
				offsetTableOffset = -1;
			}

			return offsetTableOffset;
		}

		#endregion
	}
}
