﻿using System;
using System.Diagnostics;
using System.IO;


namespace OfficeIVS.Font
{


	/// <summary>
	/// ストリーム位置のスタック機能を提供します。
	/// </summary>
	internal struct StreamPositionStack : IDisposable
	{
		#region フィールド

		/// <summary>
		/// 操作対象ストリーム。
		/// </summary>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private readonly Stream _stream;

		/// <summary>
		/// 変更前（終了時に戻す）のストリーム位置。
		/// </summary>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private readonly long _position;

		#endregion

		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.StreamPositionStack"/> のインスタンスを初期化します。
		/// </summary>
		/// <param name="stream">対象ストリーム。</param>
		/// <exception cref="ArgumentException"><paramref name="stream"/> がシークをサポートしていない。</exception>
		/// <exception cref="ArgumentNullException"><paramref name="stream"/> が null。</exception>
		public StreamPositionStack(Stream stream)
		{
			if (stream == null)
			{
				throw new ArgumentNullException("stream");
			}

			if (!stream.CanSeek)
			{
				throw new ArgumentException("stream");
			}

			_stream = stream;
			_position = _stream.Position;
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.StreamPositionStack"/> のインスタンスを初期化します。
		/// </summary>
		/// <param name="stream">対象ストリーム。</param>
		/// <param name="newPosition">新しいストリーム内での位置。</param>
		/// <exception cref="ArgumentException"><paramref name="stream"/> がシークをサポートしていない。</exception>
		/// <exception cref="ArgumentNullException"><paramref name="stream"/> が null。</exception>
		public StreamPositionStack(Stream stream, long newPosition)
			: this(stream)
		{
			_stream.Position = newPosition;
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.StreamPositionStack"/> のインスタンスを初期化します。
		/// </summary>
		/// <param name="stream">対象ストリーム。</param>
		/// <exception cref="ArgumentException"><paramref name="stream"/> がシークをサポートしていない。</exception>
		/// <exception cref="ArgumentNullException"><paramref name="stream"/> が null。</exception>
		public StreamPositionStack(BigEndianBinaryReader stream)
			: this(stream.BaseStream)
		{
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.StreamPositionStack"/> のインスタンスを初期化します。
		/// </summary>
		/// <param name="stream">対象ストリーム。</param>
		/// <param name="newPosition">新しいストリーム内での位置。</param>
		/// <exception cref="ArgumentException"><paramref name="stream"/> がシークをサポートしていない。</exception>
		/// <exception cref="ArgumentNullException"><paramref name="stream"/> が null。</exception>
		public StreamPositionStack(BigEndianBinaryReader stream, long newPosition)
			: this(stream.BaseStream, newPosition)
		{
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.StreamPositionStack"/> のインスタンスを初期化します。
		/// </summary>
		/// <param name="stream">対象ストリーム。</param>
		/// <exception cref="ArgumentException"><paramref name="stream"/> がシークをサポートしていない。</exception>
		/// <exception cref="ArgumentNullException"><paramref name="stream"/> が null。</exception>
		public StreamPositionStack(BigEndianBinaryWriter stream)
			: this(stream.BaseStream)
		{
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.StreamPositionStack"/> のインスタンスを初期化します。
		/// </summary>
		/// <param name="stream">対象ストリーム。</param>
		/// <param name="newPosition">新しいストリーム内での位置。</param>
		/// <exception cref="ArgumentException"><paramref name="stream"/> がシークをサポートしていない。</exception>
		/// <exception cref="ArgumentNullException"><paramref name="stream"/> が null。</exception>
		public StreamPositionStack(BigEndianBinaryWriter stream, long newPosition)
			: this(stream.BaseStream, newPosition)
		{
		}

		#endregion

		#region メソッド

		/// <summary>
		/// リソースを破棄します。
		/// </summary>
		public void Dispose()
		{
			// ストリーム位置を開始前の位置に戻す
			_stream.Position = _position;

			GC.SuppressFinalize(this);
		}

		#endregion
	}
}