using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;


namespace OfficeIVS.Font.Table
{

	/// <summary>
	/// OpenType フォントの cmap (Character To Glyph Index Mapping) テーブルを表します。
	/// </summary>
	[OpenTypeFontTable("cmap")]
	internal class CMap : OpenTypeFontTable
	{
		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="CMap"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="version">テーブルバージョン番号。</param>
		public CMap(ushort version) : base(TableTags.cmap)
		{
			Version = version;
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// テーブルバージョン番号を取得または設定します。
		/// </summary>
		/// <value>テーブルバージョン番号。</value>
		/// <remarks>1.0 であれば 0x00010000。</remarks>
		public ushort Version { get; protected set; }

		/// <summary>
		/// Unicode コードポイントとグリフ番号のマッピングテーブルを取得します。
		/// </summary>
		/// <value> Unicode コードポイントとグリフ番号のマッピングテーブル。</value>
		public virtual SortedList<int, ushort> CharacterToGlyphMap
		{
			[DebuggerStepThrough]
			get { throw new NotImplementedException(); }
		}

		/// <summary>
		/// 異体字セレクタ付き文字とグリフ番号のマッピングテーブルを取得します。
		/// </summary>
		/// <value>異体字セレクタ付き文字とグリフ番号のマッピングテーブル。</value>
		/// <remarks>キー1:基底文字、キー2:異体字セレクタ、値:グリフ番号。</remarks>
		public virtual SortedList<int, SortedList<int, ushort>> IvsToGlyphMap
		{
			[DebuggerStepThrough]
			get { throw new NotImplementedException(); }
		}

		/// <summary>
		/// cmap に登録されている最小の文字 Unicode コードポイントを取得します。
		/// </summary>
		/// <value>
		/// cmap に登録されている最小の Unicode コードポイント。
		/// なお、最小の Unicode コードポイント が基本多言語面外である場合には 0xFFFF を返します。
		/// </value>
		/// <remarks>'OS/2' テーブルでの利用を想定。</remarks>
		public virtual ushort FirstCharIndex
		{
			[DebuggerStepThrough]
			get { throw new NotImplementedException(); }
		}

		/// <summary>
		/// cmap に登録されている最大の Unicode コードポイントを取得します。
		/// </summary>
		/// <value>
		/// cmap に登録されている最大の Unicode コードポイント。
		/// なお、最小の Unicode コードポイント が基本多言語面外である場合には 0xFFFF を返します。
		/// </value>
		/// <remarks>'OS/2' テーブルでの利用を想定。</remarks>
		public virtual ushort LastCharIndex
		{
			[DebuggerStepThrough]
			get { throw new NotImplementedException(); }
		}

		#endregion

		#region パブリック メソッド

		/// <summary>
		/// フォントから 'cmap' テーブルを読み込みます。
		/// </summary>
		/// <param name="font">読み込み元となるフォント。</param>
		/// <returns><see cref="T:OfficeIVS.Font.Table.CMap"/> 型のインスタンス。テーブルが存在しない場合には <c>null</c>。</returns>
		[OpenTypeFontTableFactoryMethod]
		public static OpenTypeFontTable Load(OpenTypeFont font)
		{
			CMap table = null;

			ITableRecord tableRecord = font[TableTags.cmap];

			if (tableRecord != null)
			{
				using (Stream stream = tableRecord.GetTableStream())
				using (BigEndianBinaryReader reader = new BigEndianBinaryReader(stream))
				{
					ushort version = reader.ReadUInt16();

					switch (version)
					{
					case 0:
						table = new CMap00();
						break;

					default:
						table = new CMap(version);
						break;
					}

					stream.Seek(-2, SeekOrigin.Current);
					table.LoadEntries(reader, tableRecord.Length);

					table.TableRecord = tableRecord;
				}
			}

			return table;
		}

		#endregion

		#region プロテクト メソッド

		/// <summary>
		/// テーブル依存のエントリデータを読み込みます。
		/// </summary>
		/// <param name="input">テーブルデータストリーム。</param>
		/// <param name="length">テーブル長（バイト単位）。</param>
		protected override void LoadEntries(BigEndianBinaryReader input, long length)
		{
			Version = input.ReadUInt16();

			LoadExtendedEntries(input, length - 2);
		}

		/// <summary>
		/// テーブル依存のエントリデータを書き込みます。
		/// </summary>
		/// <param name="output">書き込み先のストリーム。</param>
		protected override void WriteEntries(BigEndianBinaryWriter output)
		{
			output.Write(Version);

			WriteExtendedEntries(output);
		}

		#endregion
	}


	/// <summary>
	/// OpenType フォントの cmap (Character To Glyph Index Mapping) テーブル バージョン 0 を表します。
	/// </summary>
	internal class CMap00 : CMap
	{
		#region フィールド

		/// <summary>
		/// Unicode 文字とグリフ番号のマッピングテーブル。
		/// </summary>
		/// <remarks>キー:コードポイント、値:グリフ番号。</remarks>
		private readonly SortedList<int, ushort> _charToGlyphIndex;

		/// <summary>
		/// 異体字セレクタ付き文字とグリフ番号のマッピングテーブル。
		/// </summary>
		/// <remarks>キー1:基底文字、キー2:異体字セレクタ、値:グリフ番号。</remarks>
		private readonly SortedList<int, SortedList<int,ushort>> _ivsToGlyphIndex;

		#endregion

		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="CMap00"/> クラスのインスタンスを初期化します。
		/// </summary>
		public CMap00() : base(0x0000)
		{
			_charToGlyphIndex = new SortedList<int, ushort>();
			_ivsToGlyphIndex = new SortedList<int, SortedList<int, ushort>>();
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// Unicode コードポイントとグリフ番号のマッピングテーブルを取得します。
		/// </summary>
		/// <value> Unicode コードポイントとグリフ番号のマッピングテーブル。</value>
		public override SortedList<int, ushort> CharacterToGlyphMap
		{
			[DebuggerStepThrough]
			get { return _charToGlyphIndex; }
		}

		/// <summary>
		/// 異体字セレクタ付き文字とグリフ番号のマッピングテーブルを取得します。
		/// </summary>
		/// <value>異体字セレクタ付き文字とグリフ番号のマッピングテーブル。</value>
		/// <remarks>キー1:基底文字、キー2:異体字セレクタ、値:グリフ番号。</remarks>
		public override SortedList<int, SortedList<int,ushort>> IvsToGlyphMap
		{
			[DebuggerStepThrough]
			get { return _ivsToGlyphIndex; }
		}

		/// <summary>
		/// cmap に登録されている最小の文字 Unicode コードポイントを取得します。
		/// </summary>
		/// <value>
		/// cmap に登録されている最小の Unicode コードポイント。
		/// なお、最小の Unicode コードポイント が基本多言語面外である場合には 0xFFFF を返します。
		/// </value>
		/// <remarks>'OS/2' テーブルでの利用を想定。</remarks>
		public override ushort FirstCharIndex
		{
			[DebuggerStepThrough]
			get
			{
				return (ushort)Math.Max(0, Math.Min(0xFFFF, _charToGlyphIndex.Keys[0]));
			}
		}

		/// <summary>
		/// cmap に登録されている最大の Unicode コードポイントを取得します。
		/// </summary>
		/// <value>
		/// cmap に登録されている最大の Unicode コードポイント。
		/// なお、最小の Unicode コードポイント が基本多言語面外である場合には 0xFFFF を返します。
		/// </value>
		/// <remarks>'OS/2' テーブルでの利用を想定。</remarks>
		public override ushort LastCharIndex
		{
			[DebuggerStepThrough]
			get
			{
				return (ushort)Math.Max(0, Math.Min(0xFFFF, _charToGlyphIndex.Keys[_charToGlyphIndex.Keys.Count-1]));
			}
		}

		#endregion

		#region パブリック メソッド

		/// <summary>
		/// 指定文字に対応するグリフ番号を取得します。
		/// </summary>
		/// <param name="baseChar">基底文字コード。</param>
		/// <param name="varSelector">異体字セレクタコード。</param>
		/// <returns>グリフ番号。存在しない場合には 0。</returns>
		public ushort GetGlyphId(int baseChar, int varSelector=0)
		{
			ushort glyphId = 0;

			if (varSelector == 0)
			{
				_charToGlyphIndex.TryGetValue(baseChar, out glyphId);
			}
			else
			{
				SortedList<int, ushort> mapTable;

				if (_ivsToGlyphIndex.TryGetValue(baseChar, out mapTable))
				{
					mapTable.TryGetValue(varSelector, out glyphId);	
				}
			}

			return glyphId;
		}

		/// <summary>
		/// 指定文字コードに対応するグリフのグリフ番号を設定します。
		/// </summary>
		/// <param name="unicodeValue">文字コード。</param>
		/// <param name="glyphIndex">グリフ番号。</param>
		/// <remarks><paramref name="glyphIndex"/> に 0 を設定した場合、登録解除。</remarks>
		public void SetGlyphId(int unicodeValue, ushort glyphIndex)
		{
			if (glyphIndex != 0)
			{
				_charToGlyphIndex[unicodeValue] = glyphIndex;
			}
			else if (_charToGlyphIndex.ContainsKey(unicodeValue))
			{
				_charToGlyphIndex.Remove(unicodeValue);
			}
		}

		/// <summary>
		/// 指定された基底文字と異体字セレクタの組み合わせに対応するグリフのグリフ番号を設定します。
		/// </summary>
		/// <param name="baseChar">基底文字コード。</param>
		/// <param name="varSelector">異体字セレクタコード。</param>
		/// <param name="glyphIndex">グリフ番号。</param>
		/// <remarks><paramref name="glyphIndex"/> に 0 を設定した場合、登録解除。</remarks>
		public void SetGlyphId(int baseChar, int varSelector, ushort glyphIndex)
		{
			SortedList<int, ushort> mapTable;

			_ivsToGlyphIndex.TryGetValue(baseChar, out mapTable);

			if (glyphIndex != 0)
			{
				if (mapTable == null)
				{
					_ivsToGlyphIndex[baseChar]
						= (mapTable = new SortedList<int, ushort>());
				}

				mapTable[varSelector] = glyphIndex;
			}
			else if (mapTable != null && mapTable.ContainsKey(varSelector))
			{
				mapTable.Remove(varSelector);

				if (mapTable.Count == 0)
				{
					_ivsToGlyphIndex.Remove(baseChar);
				}
			}
		}

		#endregion

		#region プロテクト メソッド

		/// <summary>
		/// テーブル依存のエントリデータを読み込みます。
		/// </summary>
		/// <param name="input">テーブルデータストリーム。</param>
		/// <param name="length">テーブル長（バイト単位）。</param>
		protected override void LoadEntries(BigEndianBinaryReader input, long length)
		{
			long tableTopPos = input.Position;

			Version = input.ReadUInt16();

			ushort numTables = input.ReadUInt16();

			long format4Offset = -1;
			long format12Offset = -1;
			long format14Offset = -1;

			for (int i=0; i<numTables; i++)
			{
				ushort platformId = input.ReadUInt16();
				ushort encodingId = input.ReadUInt16();
				uint tableOffset = input.ReadUInt32();

				if (platformId == 0)
				{
					if (encodingId == 5)
					{
						format14Offset = tableTopPos + tableOffset;
					}
				}
				else if (platformId == 3)
				{
					if (encodingId == 1)
					{
						format4Offset = tableTopPos + tableOffset;
					}
					else if (encodingId == 10)
					{
						format12Offset = tableTopPos + tableOffset;
					}
				}
			}

			// format 12 優先
			//  : format 12 (UCS-4) は format 4 (UCS-2) の内容を含むため
			if (format12Offset > 0)
			{
				LoadFormat12Subtable(input, format12Offset);
			}
			else if (format4Offset > 0)
			{
				LoadFormat4Subtable(input, format4Offset);
			}

			// format 14（異体字セレクタ）
			if (format14Offset > 0)
			{
				LoadFormat14Subtable(input, format14Offset);
			}
		}

		/// <summary>
		/// テーブル依存のエントリデータを書き込みます。
		/// </summary>
		/// <param name="output">書き込み先のストリーム。</param>
		protected override void WriteEntries(BigEndianBinaryWriter output)
		{
			// サブテーブルの出力有無
			bool needFormat12 = _charToGlyphIndex.Keys.Any(v => v > 0xFFFF);
			bool needFormat14 = _ivsToGlyphIndex.Count > 0;

			// version
			output.Write(Version);

			// numTables
			ushort numTables = 1;

			if (needFormat12) numTables++;
			if (needFormat14) numTables++;

			output.Write(numTables);

			// tableRecord
			//  : 領域だけ確保、テーブル内容書き込み時に出力
			long tableRecordPos = output.Position;
			output.Position += 8 * numTables;

			// format 14 サブテーブル
			if (needFormat14)
			{
				// テーブルレコード
				long currentPos = output.Position;
				output.Position = tableRecordPos;
				output.Write((ushort)0);
				output.Write((ushort)5);
				output.Write((uint)currentPos);
				tableRecordPos += 8;
				output.Position = currentPos;

				// サブテーブル
				WriteFormat14Subtable(output);
			}

			// format 4 サブテーブル（必須）
			{
				// テーブルレコード
				long currentPos = output.Position;
				output.Position = tableRecordPos;
				output.Write((ushort)3);
				output.Write((ushort)1);
				output.Write((uint)currentPos);
				tableRecordPos += 8;
				output.Position = currentPos;

				// サブテーブル
				WriteFormat4Subtable(output);	
			}

			// format 12 サブテーブル
			if (needFormat12)
			{
				// テーブルレコード
				long currentPos = output.Position;
				output.Position = tableRecordPos;
				output.Write((ushort)3);
				output.Write((ushort)10);
				output.Write((uint)currentPos);
				output.Position = currentPos;

				// サブテーブル
				WriteFormat12Subtable(output);	
			}
		}

		#endregion

		#region プライベート メソッド

		/// <summary>
		/// format 4 サブテーブルを読み込みます。
		/// </summary>
		/// <param name="input">読み込み元ストリーム。</param>
		/// <param name="tableOffset">ストリーム先頭から format 4 サブテーブル先頭までのオフセット量（バイト単位）。</param>
		private void LoadFormat4Subtable(BigEndianBinaryReader input, long tableOffset)
		{
			// format 4 サブテーブル先頭に移動
			//  : format（使用しないので読み飛ばす）
			input.Position = tableOffset + 2;

			// length
			ushort length = input.ReadUInt16();

			// language（使用しないので読み飛ばす）
			input.Position += 2;

			// segCount:ushort
			int segCount = input.ReadUInt16() / 2;

			// 以下のエントリは使用しないため読み飛ばす
			//   searchRange:ushort, entrySelector:ushort, rangeShift:ushort
			input.Position += 6;

			// endCount
			ushort[] endCount = new ushort[segCount];

			for (int i=0; i<endCount.Length; i++)
			{
				endCount[i] = input.ReadUInt16();
			}

			// reservedPad:ushort
			input.Position += 2;

			// startCount
			ushort[] startCount = new ushort[segCount];

			for (int i=0; i<startCount.Length; i++)
			{
				startCount[i] = input.ReadUInt16();
			}

			// idDelta
			short[] idDelta = new short[segCount];

			for (int i=0; i<idDelta.Length; i++)
			{
				idDelta[i] = input.ReadInt16();
			}

			// idRangeOffset
			ushort[] idRangeOffset = new ushort[segCount];

			for (int i=0; i<idRangeOffset.Length; i++)
			{
				idRangeOffset[i] = input.ReadUInt16();
			}

			// glyphIdArray
			ushort[] glyphIdArray = new ushort[(length - (16 + 4 * (2 * segCount))) / 2];

			for (int i=0; i<glyphIdArray.Length; i++)
			{
				glyphIdArray[i] = input.ReadUInt16();
			}			

			// マッピングテーブル作成
			for (int i=0; i<segCount; i++)
			{
				if (endCount[i] == 0xFFFF)
				{
					break;
				}

				if (idRangeOffset[i] == 0)
				{
					for (ushort charCode=startCount[i]; charCode<=endCount[i]; charCode++)
					{
						_charToGlyphIndex[charCode] = (ushort)((charCode + idDelta[i]) & 0xFFFF);
					}				
				}
				else
				{
					int index = (idRangeOffset[i] - 2 * (segCount - i)) / 2;

					for (ushort charCode=startCount[i]; charCode<=endCount[i]; charCode++)
					{
						int glyphId = glyphIdArray[index + charCode - startCount[i]];
						_charToGlyphIndex[charCode] = (ushort)((glyphId + idDelta[i]) & 0xFFFF);
					}
				}
			}
		}

		/// <summary>
		/// format 4 サブテーブル形式のセグメント情報を取得します。
		/// </summary>
		/// <returns>セグメント情報コレクション。</returns>
		private List<Format4Segment> GetFormat4SubtableSegments()
		{
			List<Format4Segment> segments = new List<Format4Segment>();

			// 基本多言語面の文字のみを抽出
			List<ushort> charCodes = new List<ushort>();
			{
				foreach (long charCode in _charToGlyphIndex.Keys)
				{
					if (charCode >= 0x10000)
					{
						break;
					}
					charCodes.Add((ushort)(charCode & 0xFFFF));
				}
			}
			
			// セグメント分割
			List<ushort> subSegmentLengths = new List<ushort>();

			for (int i=0; i<charCodes.Count; i++)
			{
				ushort startCount = charCodes[i];

				subSegmentLengths.Clear();

				// サブセグメント分割
				//  : グリフ番号の連続性が途切れる箇所の抽出
				//     subSegmentLengths には、グリフ番号の連続数を格納
				if (i == charCodes.Count-1)
				{
					subSegmentLengths.Add(1);
				}
				else
				{
					ushort subSegLength = 1;

					for (int j=i+1; j<charCodes.Count; j++)
					{
						ushort code0 = charCodes[j-1];
						ushort code1 = charCodes[j];

						// 文字コードの連続性が途切れたところでループ終了
						if (code1 != code0 + 1)
						{
							subSegmentLengths.Add(subSegLength);
							subSegLength = 0;
							break;
						}

						// グリフ番号の連続性が途切れた箇所を記録
						if (_charToGlyphIndex[code1] == _charToGlyphIndex[code0] + 1)
						{
							subSegLength++;
						}
						else
						{
							subSegmentLengths.Add(subSegLength);
							subSegLength = 1;
						}

						i++;
					}

					if (subSegLength > 0)
					{
						subSegmentLengths.Add(subSegLength);
					}
				}

				// グリフ番号によるセグメント長変更
				//  : テーブル長を小さくするため、グリフ番号が連続する部分は idDeleta 形式、
				//    飛び飛びの部分は idRangeOffset 形式で格納する。
				if (subSegmentLengths.Count == 1)
				{
					// 全てのグリフ番号が連続していれば無条件で idDelta 形式で登録
					segments.Add(new Format4Segment(startCount, subSegmentLengths[0], _charToGlyphIndex[startCount]));
				}
				else
				{
					// idDelta 形式、idRangeOffset 形式の振り分けの閾値。
					ushort minLength = subSegmentLengths.Count == 2 ? (ushort)5 : (ushort)9;

					ushort segmentLength = 0;

					foreach (ushort subSegLength in subSegmentLengths)
					{
						if (subSegLength < minLength)
						{
							// グリフ番号の連続性が閾値未満であれば、まとめて idRangeOffset 形式
							segmentLength += subSegLength;
						}
						else if (segmentLength == 0)
						{
							// idDelta 形式で登録
							segments.Add(new Format4Segment(startCount, subSegLength, _charToGlyphIndex[startCount]));
							startCount += subSegLength;
						}
						else
						{
							// これより前のセグメントを idRangeOffset 形式で登録
							segments.Add(new Format4Segment(startCount, segmentLength));
							startCount += segmentLength;

							// 現在のセグメントを idDelta 形式で登録
							segments.Add(new Format4Segment(startCount, subSegLength, _charToGlyphIndex[startCount]));
							startCount += subSegLength;

							segmentLength = 0;
						}
					}

					// 残りの処理
					if (segmentLength != 0)
					{
						segments.Add(new Format4Segment(startCount, segmentLength));
					}					
				}
			}
				
			// 終端用セグメント（固定値）
			segments.Add(new Format4Segment(0xFFFF, 1, 0));

			return segments;
		}

		/// <summary>
		/// 現在のマッピング情報を format 4 サブテーブル形式で指定ストリームに出力します。
		/// </summary>
		/// <param name="output">出力先ストリーム。</param>
		private void WriteFormat4Subtable(BigEndianBinaryWriter output)
		{
			// セグメント情報の取得
			List<Format4Segment> segments = GetFormat4SubtableSegments();

			long tableTopPos = output.Position;

			// format
			output.Write((ushort)4);

			// length
			// : 後で再設定するのでとりあえず 0			
			output.Write((ushort)0);

			// language
			output.Write((ushort)0);

			// segCountX2
			ushort segCountX2 = (ushort)(2 * segments.Count);
			output.Write(segCountX2);
	
			// searchRange
			ushort searchRange = (ushort)(2 * Math.Pow(2, Math.Floor(Math.Log(segments.Count, 2))));
			output.Write(searchRange);

			// entrySelector
			output.Write((ushort)(Math.Log(searchRange, 2)-1));

			// rangeShift
			output.Write((ushort)(segCountX2 - searchRange));

			// endCount
			foreach (var segment in segments)
			{
				output.Write(segment.EndCount);
			}

			// reservedPad
			output.Write((ushort)0);

			// startCount
			foreach (var segment in segments)
			{
				output.Write(segment.StartCount);
			}

			// idDelta
			foreach (var segment in segments)
			{
				output.Write(segment.IdDelta);
			}

			// idRangeOffset
			ushort idRangeOffset = segCountX2;
			foreach (var segment in segments)
			{
				if (segment.IdRangeOffset == 0)
				{
					output.Write((ushort)0);	
				}
				else
				{
					output.Write(idRangeOffset);
					idRangeOffset += segment.IdRangeOffset;
				}
				idRangeOffset -= 2;
			}

			// glyphIdArray
			foreach (var segment in segments)
			{
				if (segment.IdRangeOffset != 0)
				{
					for (ushort charCode=segment.StartCount; charCode<=segment.EndCount; charCode++)
					{
						output.Write(_charToGlyphIndex[charCode]);
					}
				}
			}

			// length エントリの再設定
			ushort length = (ushort)(output.Position - tableTopPos);

			using (new StreamPositionStack(output, tableTopPos + 2))
			{
				output.Write(length);
			}
		}

		/// <summary>
		/// format 12 サブテーブルを読み込みます。
		/// </summary>
		/// <param name="input">読み込み元ストリーム。</param>
		/// <param name="tableOffset">ストリーム先頭から format 12 サブテーブル先頭までのオフセット量（バイト単位）。</param>
		private void LoadFormat12Subtable(BigEndianBinaryReader input, long tableOffset)
		{
			// format 12 サブテーブル先頭に移動
			//   なお以下のテーブルエントリは不要なため読み飛ばす
			//     format:ushort, reserved:ushort, length:uint, language:uint
			input.Position = tableOffset + 12;

			// nGroups
			uint nGroups = input.ReadUInt32();

			// Segment Group
			for (uint i=0; i<nGroups; i++)
			{
				int startCharCode = input.ReadInt32();
				int endCharCode = input.ReadInt32();
				int startGlyphId = input.ReadInt32();

				for (int charCode=startCharCode; charCode<=endCharCode; charCode++)
				{
					_charToGlyphIndex[charCode] =
						Convert.ToUInt16(startGlyphId + (charCode - startCharCode));
				}
			}
		}

		/// <summary>
		/// 現在のマッピング情報を format 12 サブテーブル形式で指定ストリームに出力します。
		/// </summary>
		/// <param name="output">出力先ストリーム。</param>
		private void WriteFormat12Subtable(BigEndianBinaryWriter output)
		{
			// 書き込み開始位置
			long tableTopPos = output.Position;

			// format
			output.Write((ushort)12);

			// reserved
			output.Write((ushort)0);

			// 以下のエントリは後で再設定するので領域のみ確保
			//   length:uint, language:uint, nGroups:uint
			output.Position += 12;

			// SegmentGroups
			uint nGroups = 0;

			for (int i=0; i<_charToGlyphIndex.Count; i++)
			{
				int startCharCode = _charToGlyphIndex.Keys[i];
				int endCharCode = startCharCode;
				int startGlyphIndex = _charToGlyphIndex[startCharCode];

				// 文字コード、グリフ番号が連続する場合には 1 セグメントにまとめる
				for (int j=1; j<_charToGlyphIndex.Count-i; j++)
				{
					int charCode = _charToGlyphIndex.Keys[i+j];

					if (charCode - startCharCode != j
					 || _charToGlyphIndex[charCode] - startGlyphIndex != j)
					{
						i += j-1;
						break;
					}

					endCharCode = charCode;
				}

				output.Write(startCharCode);
				output.Write(endCharCode);
				output.Write(startGlyphIndex);
				nGroups++;
			}

			uint length = (uint)(output.Position - tableTopPos);

			// length, language nGroups の設定
			using (new StreamPositionStack(output, tableTopPos + 4))
			{
				// length
				output.Write(length);

				// language
				output.Write((uint)0);

				// nGroups の再設定
				output.Write(nGroups);
			}
		}

		/// <summary>
		/// format 14 サブテーブルを読み込みます。
		/// </summary>
		/// <param name="input">読み込み元ストリーム。</param>
		/// <param name="tableOffset">ストリーム先頭から format 14 サブテーブル先頭までのオフセット量（バイト単位）。</param>
		private void LoadFormat14Subtable(BigEndianBinaryReader input, long tableOffset)
		{
			// format 14 サブテーブル先頭に移動
			//   なお以下のテーブルエントリは不要なため読み飛ばす
			//     format:ushort, length:uint
			input.Position = tableOffset + 6;

			// numVarSelectorRecords
			uint numVarSelectorRecords = input.ReadUInt32();

			// Variation Selector Record 解析
			for (uint index=0; index<numVarSelectorRecords; index++)
			{
				int varSelector = input.ReadInt24();
				uint defaultUvsOffset = input.ReadUInt32();
				uint nonDefaultUvsOffset = input.ReadUInt32();

				using (new StreamPositionStack(input))
				{
					// Default UVS
					if (defaultUvsOffset != 0)
					{
						input.Position = tableOffset + defaultUvsOffset;

						uint numUnicodeValueRanges = input.ReadUInt32();

						for (uint i=0; i<numUnicodeValueRanges; i++)
						{
							int unicodeValue = input.ReadInt24();
							int additionalCount = input.ReadByte();

							for (; additionalCount>=0; additionalCount--, unicodeValue++)
							{
								ushort glyphId;

								if (_charToGlyphIndex.TryGetValue(unicodeValue, out glyphId))
								{
									SetGlyphId(unicodeValue, varSelector, glyphId);
								}
							}
						}
					}

					// Non Default UVS
					if (nonDefaultUvsOffset != 0)
					{
						input.Position = tableOffset + nonDefaultUvsOffset;

						uint numUVSMappings = input.ReadUInt32();

						for (uint i=0; i<numUVSMappings; i++)
						{
							SetGlyphId(input.ReadInt24(), varSelector, input.ReadUInt16());
						}
					}
				}
			}
		}

		/// <summary>
		/// 現在のマッピング情報を format 14 サブテーブル形式で指定ストリームに出力します。
		/// </summary>
		/// <param name="output">出力先ストリーム。</param>
		private void WriteFormat14Subtable(BigEndianBinaryWriter output)
		{
			// format 14 用に並び替え
			//   キー：異体字セレクタ、値：標準基底文字リスト、非標準基底文字リスト
			var uvsMappingRecords = new SortedList<int, Tuple<List<int>, List<int>>>();

			foreach (int baseChar in _ivsToGlyphIndex.Keys)
			foreach (int varSelector in _ivsToGlyphIndex[baseChar].Keys)
			{
				// 対象レコード取得
				Tuple<List<int>, List<int>> varSelectorRecord;

				if (!uvsMappingRecords.TryGetValue(varSelector, out varSelectorRecord))
				{
					varSelectorRecord = new Tuple<List<int>, List<int>>(new List<int>(), new List<int>());
					uvsMappingRecords[varSelector] = varSelectorRecord;
				}

				// 標準、非標準の振り分け
				//  : 異体字セレクタ付き文字が基底文字と同じグリフ番号を持つものは標準
				if (_ivsToGlyphIndex[baseChar][varSelector] == _charToGlyphIndex[baseChar])
				{
					varSelectorRecord.Item1.Add(baseChar);
				}
				else
				{
					varSelectorRecord.Item2.Add(baseChar);
				}
			}

			long tableTopPos = output.Position;

			// format
			output.Write((ushort)14);

			// length
			// : 後で再設定するのでとりあえず 0
			output.Write((uint)0);

			// numVarSelectorRecords
			output.Write((uint)uvsMappingRecords.Count);

			// Variation Selector Records
			//  : 領域分だけ空けておき後で再設定
			output.Position += 11 * uvsMappingRecords.Count;

			for (int i=0; i<uvsMappingRecords.Count; i++)
			{
				int ivsChar = uvsMappingRecords.Keys[i];
				var varSelectorRecord = uvsMappingRecords[ivsChar];

				// Default UVS Table
				long defaultUVSTableOffset = 0;
				List<int> defaultUVSTableItems = varSelectorRecord.Item1;

				if (defaultUVSTableItems.Count > 0)
				{
					// 基底文字で昇順にソート
					defaultUVSTableItems.Sort();

					// Table Offset 位置計算
					defaultUVSTableOffset = output.Position - tableTopPos;

					// numUnicodeValueRanges
					// : 後で再設定するのでとりあえず 0
					uint numUnicodeValueRanges = 0;
					output.Write(numUnicodeValueRanges);

					// Unicode Value Range
					//  : 基底文字の文字コードが連続するものは 1 レコードとして書き込む
					for (int j=0; j<defaultUVSTableItems.Count; j++)
					{
						int startUnicodeValue = defaultUVSTableItems[j];
						byte additionalCount = 0;

						for (int k=j+1; k<defaultUVSTableItems.Count && additionalCount<0xFF; k++)
						{
							if (defaultUVSTableItems[k] - defaultUVSTableItems[k-1] == 1)
							{
								additionalCount++;
							}
							else
							{
								j = k - 1;
								break;
							}
						}

						numUnicodeValueRanges++;
						output.WriteUInt24((uint)startUnicodeValue);
						output.Write(additionalCount);
					}

					// numUnicodeValueRanges 再設定
					using (new StreamPositionStack(output, tableTopPos + defaultUVSTableOffset))
					{
						output.Write(numUnicodeValueRanges);
					}
				}

				// Non Default UVS Table
				long nonDefaultUVSTableOffset = 0;
				List<int> nonDefaultUVSTableItems = varSelectorRecord.Item2;

				if (nonDefaultUVSTableItems.Count > 0)
				{
					// 基底文字で昇順にソート
					nonDefaultUVSTableItems.Sort();

					// Table Offset 位置計算
					nonDefaultUVSTableOffset = output.Position - tableTopPos;

					// numUVSMappings
					output.Write((uint)nonDefaultUVSTableItems.Count);

					// UVS Mappings
					foreach (int baseChar in nonDefaultUVSTableItems)
					{
						output.WriteUInt24((uint)baseChar);
						output.Write(_ivsToGlyphIndex[baseChar][ivsChar]);
					}
				}

				// Variation Selector Records
				using (new StreamPositionStack(output, tableTopPos + 10 + 11 * i))
				{
					output.WriteUInt24((uint)ivsChar);
					output.Write((uint)defaultUVSTableOffset);
					output.Write((uint)nonDefaultUVSTableOffset);
				}
			}

			// length エントリの再設定
			uint length = (uint)(output.Position - tableTopPos);

			using (new StreamPositionStack(output, tableTopPos + 2))
			{
				output.Write(length);
			}
		}

		#endregion

		#region ヘルパー クラス

		/// <summary>
		/// fomart 4 サブテーブルのセグメント情報を表します。
		/// </summary>
		private sealed class Format4Segment
		{
			#region コンストラクタ

			/// <summary>
			/// 新しい <see cref="T:OfficeIVS.Font.Table.CMap00.Format4Segment"/> クラスのインスタンスを初期化します。
			/// </summary>
			/// <param name="startCount">セグメント先頭の文字コード。</param>
			/// <param name="segmentLength">セグメント長（文字数）。</param>
			public Format4Segment(ushort startCount, ushort segmentLength)
			{
				StartCount = startCount;
				EndCount = (ushort)(startCount + segmentLength - 1);
				IdDelta = 0;
				IdRangeOffset = (ushort)(2 * segmentLength);
			}

			/// <summary>
			/// 新しい <see cref="T:OfficeIVS.Font.Table.CMap00.Format4Segment"/> クラスのインスタンスを初期化します。
			/// </summary>
			/// <param name="startCount">セグメント先頭の文字コード。</param>
			/// <param name="segmentLength">セグメント長（文字数）。</param>
			/// <param name="startGlyphId">セグメント先頭のグリフ番号。</param>
			public Format4Segment(ushort startCount, ushort segmentLength, ushort startGlyphId)
			{
				StartCount = startCount;
				EndCount = (ushort)(startCount + segmentLength - 1);
				IdDelta = unchecked ((short)(startGlyphId - startCount));
				IdRangeOffset = 0;
			}

			#endregion

			#region プロパティ

			/// <summary>
			/// セグメント先頭の文字コードを取得または設定します。
			/// </summary>
			/// <value>セグメント先頭の文字コード。</value>
			public ushort StartCount { get; private set; }

			/// <summary>
			/// セグメント末尾の文字コードを取得または設定します。
			/// </summary>
			/// <value>セグメント末尾の文字コード。</value>
			public ushort EndCount { get; private  set; }

			/// <summary>
			/// グリフ番号との差分値を取得または設定します。
			/// </summary>
			/// <value>グリフ番号との差分値。</value>
			public short IdDelta { get; private set; }

			/// <summary>
			/// 本セグメントに対応するグリフ番号配列までのオフセット量を取得または設定します。
			/// </summary>
			/// <value>グリフ番号配列までのオフセット量（バイト単位）。</value>
			public ushort IdRangeOffset { get; private set; }

			#endregion
		}

		#endregion
	}
}