using System.IO;


namespace OfficeIVS.Font.Table
{

	/// <summary>
	/// OpenType フォントの cvt (Control Value Table) テーブルを表します。
	/// </summary>
	[OpenTypeFontTable("cvt ")]
	internal sealed class Cvt : OpenTypeFontTable
	{
		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="Cvt"/> クラスのインスタンスを初期化します。
		/// </summary>
		public Cvt() : base(TableTags.cvt)
		{
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// TrueType アウトライン命令により参照される値のコレクションを取得します。
		/// </summary>
		/// <value>TrueType アウトライン命令により参照される値のコレクション。</value>
		/// <remarks>配列サイズはテーブルサイズから算出。</remarks>
		public short[] Values { get; set; }
	
		#endregion

		#region メソッド

		/// <summary>
		/// フォントから 'cvt' テーブルを読み込みます。
		/// </summary>
		/// <param name="font">読み込み元となるフォント。</param>
		/// <returns><see cref="T:OfficeIVS.Font.Table.Cvt"/> 型のインスタンス。テーブルが存在しない場合には <c>null</c>。</returns>
		[OpenTypeFontTableFactoryMethod]
		public static OpenTypeFontTable Load(OpenTypeFont font)
		{
			Cvt table = null;

			ITableRecord tableRecord = font[TableTags.cvt];

			if (tableRecord != null)
			{
				using (Stream stream = tableRecord.GetTableStream())
				using (BigEndianBinaryReader reader = new BigEndianBinaryReader(stream))
				{
					table = new Cvt();
					table.LoadEntries(reader, tableRecord.Length);
					table.TableRecord = tableRecord;
				}
			}

			return table;
		}

		/// <summary>
		/// テーブル依存のエントリデータを読み込みます。
		/// </summary>
		/// <param name="input">テーブルデータストリーム。</param>
		/// <param name="length">テーブル長（バイト単位）。</param>
		protected override void LoadEntries(BigEndianBinaryReader input, long length)
		{
			if (length > 0)
			{
				Values = new short[length];

				for (int i=0; i<Values.Length/2; i++)
				{
					Values[i] = input.ReadInt16();
				}
			}
		}

		/// <summary>
		/// テーブル依存のエントリデータを書き込みます。
		/// </summary>
		/// <param name="output">書き込み先のストリーム。</param>
		protected override void WriteEntries(BigEndianBinaryWriter output)
		{
			if (Values != null)
			{
				foreach (ushort value in Values)
				{
					output.Write(value);
				}
			}
		}

		#endregion
	}
}