using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;


namespace OfficeIVS.Font.Table
{

	/// <summary>
	/// OpenType フォントの glyf (Glyf Data) テーブルを表します。
	/// </summary>
	[OpenTypeFontTable("glyf")]
	internal sealed class Glyf : OpenTypeFontTable
	{
		#region フィールド

		/// <summary>
		/// グリフデータ。
		/// </summary>
		/// <remarks>キー:グリフ番号、値:グリフデータ</remarks>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private readonly SortedList<ushort, GlyphDescription> _glyphDatas;

		/// <summary>
		/// 本 glyf テーブルに対応する loca テーブル。
		/// </summary>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private uint[] _offsets;

		#endregion

		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="Glyf"/> クラスのインスタンスを初期化します。
		/// </summary>
		public Glyf() : base(TableTags.glyf)
		{
			_glyphDatas= new SortedList<ushort, GlyphDescription>();
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// グリフ数を取得します。
		/// </summary>
		/// <value>グリフ数。</value>
		public ushort NumGlyphs
		{
			[DebuggerStepThrough]
			get { return (ushort)(_glyphDatas.Keys[_glyphDatas.Keys.Count-1] + 1); }
		}

		#endregion

		#region パブリック メソッド

		/// <summary>
		/// フォントから 'glyf' テーブルを読み込みます。
		/// </summary>
		/// <param name="font">読み込み元となるフォント。</param>
		/// <returns><see cref="T:OfficeIVS.Font.Table.Glyf"/> 型のインスタンス。テーブルが存在しない場合には <c>null</c>。</returns>
		[OpenTypeFontTableFactoryMethod]
		public static OpenTypeFontTable Load(OpenTypeFont font)
		{
			Glyf table = null;

			ITableRecord tableRecord = font[TableTags.glyf];
			Loca loca = Loca.Load(font) as Loca;

			if (tableRecord != null && loca != null)
			{
				// glyph data table
				table = new Glyf();
				table.TableRecord = tableRecord;
				table._offsets = loca.Offsets;
			}

			return table;
		}

		/// <summary>
		/// 指定グリフ番号のグリフ定義データを取得します。
		/// </summary>
		/// <param name="glyphId">グリフ番号。</param>
		/// <returns>グリフ定義データ。存在しない場合には null。</returns>
		public GlyphDescription GetGlyphDescription(ushort glyphId)
		{
			GlyphDescription glyphData = null;

			if (_glyphDatas.ContainsKey(glyphId))
			{
				glyphData = _glyphDatas[glyphId];
			}
			else if (_offsets != null && glyphId < _offsets.Length-1)
			{
				int offset = (int)_offsets[glyphId];
				int length = (int)_offsets[glyphId+1] - offset;

				if (length > 0)
				{
					byte[] glyphRowData = new byte[length];

					using (Stream s = TableRecord.GetTableStream(offset, length))
					{
						s.Read(glyphRowData, 0, length);
					}

					_glyphDatas.Add(glyphId, (glyphData = new GlyphDescription(glyphRowData)));
				}
			}

			return glyphData;
		}

		/// <summary>
		/// 指定グリフ定義を登録済みグリフの末尾に登録します。
		/// </summary>
		/// <param name="glyphData">グリフデータ。</param>
		/// <remarks>割り当てられたグリフ番号。</remarks>
		public ushort AddGlyphDescription(GlyphDescription glyphData)
		{
			ushort glyphId = _glyphDatas.Keys[_glyphDatas.Keys.Count-1];

			SetGlyphDescription(++glyphId, glyphData);

			return glyphId;
		}

		/// <summary>
		/// 指定グリフ番号にグリフ定義データを設定します。
		/// </summary>
		/// <param name="glyphId">グリフ番号。</param>
		/// <param name="glyphData">グリフデータ。</param>
		public void SetGlyphDescription(ushort glyphId, GlyphDescription glyphData)
		{
			_glyphDatas[glyphId] = glyphData;
		}

		/// <summary>
		/// 現在の 'glyf' テーブル内容から 'loca' テーブルレコードを作成します。
		/// </summary>
		/// <returns>'loca' テーブルレコード。現在の 'glyf' テーブルが無効な場合には null。</returns>
		public ITableRecord CreateLocaTableRecord()
		{
			ITableRecord tableRecord = null;

			if (_glyphDatas.Count > 0)
			{
				ushort maxGlyphId = _glyphDatas.Keys[_glyphDatas.Count-1];

				using (MemoryStream ms = new MemoryStream(4 * (maxGlyphId + 1)))
				using (BigEndianBinaryWriter writer = new BigEndianBinaryWriter(ms))
				{
					// 先頭は必ず 0
					writer.Write((uint)0);

					// 2 文字目以降は先頭からのオフセット量
					uint offset = 0;

					for (ushort glyphId=0; glyphId<=maxGlyphId; glyphId++)
					{
						GlyphDescription glyphData;

						if (_glyphDatas.TryGetValue(glyphId, out glyphData))
						{
							if (glyphData != null)
							{
								offset += (uint)((glyphData.GlyphData.Length + 3) & ~3);
							}
						}

						writer.Write(offset);
					}
					
					// テーブルレコード作成
					ms.Flush();
					tableRecord = new EditableTableRecord(TableTags.loca, ms.ToArray());
				}
			}

			return tableRecord;
		}

		/// <summary>
		/// 現在の 'glyf' テーブル内の全グリフバウンディングボックスの最小最大値を取得します。
		/// </summary>
		/// <param name="xMin">X 方向の最小値を受け取る変数。</param>
		/// <param name="yMin">Y 方向の最小値を受け取る変数。</param>
		/// <param name="xMax">X 方向の最大値を受け取る変数。</param>
		/// <param name="yMax">Y 方向の最大値を受け取る変数。</param>
		public void GetGlyphBoundingBox(out short xMin, out short yMin, out short xMax, out short yMax)
		{
			if (_glyphDatas.Count == 0)
			{
				xMin = 0;
				yMin = 0;
				xMax = 0;
				yMax = 0;
			}
			else
			{
				xMin = Int16.MaxValue;
				yMin = Int16.MaxValue;
				xMax = Int16.MinValue;
				yMax = Int16.MinValue;

				foreach (ushort glyphId in _glyphDatas.Keys)
				{
					GlyphDescription glyphData = _glyphDatas[glyphId];

					if (glyphData != null)
					{
						xMin = Math.Min(xMin, glyphData.XMin);
						yMin = Math.Min(yMin, glyphData.YMin);
						xMax = Math.Max(xMax, glyphData.XMax);
						yMax = Math.Max(yMax, glyphData.YMax);
					}
				}
			}
		}

		#endregion

		#region プロテクト メソッド

		/// <summary>
		/// テーブル依存のエントリデータを読み込みます。
		/// </summary>
		/// <param name="input">テーブルデータストリーム。</param>
		/// <param name="length">テーブル長（バイト単位）。</param>
		protected override void LoadEntries(BigEndianBinaryReader input, long length)
		{
		}

		/// <summary>
		/// テーブル依存のエントリデータを書き込みます。
		/// </summary>
		/// <param name="output">書き込み先のストリーム。</param>
		protected override void WriteEntries(BigEndianBinaryWriter output)
		{
			foreach (ushort glyphId in _glyphDatas.Keys)
			{
				var glyphData = _glyphDatas[glyphId];

				if (glyphData == null || glyphData.GlyphData == null)
				{
					continue;
				}

				// 個々のグリフデータの先頭は 4 バイト境界に合わせる
				output.Write(_glyphDatas[glyphId].GlyphData);
				output.Position = (output.Position + 3) & ~3;
			}

			output.BaseStream.SetLength(output.Position);
		}

		#endregion
	}


	/// <summary>
	/// TrueType アウトライン形式のフォントのグリフ定義を表します。
	/// </summary>
	/// <remarks>単一グリフ、複合グリフのグリフ定義情報。</remarks>
	internal sealed class GlyphDescription
	{
		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="GlyphDescription"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="glyphData">グリフデータ。</param>
		public GlyphDescription(byte[] glyphData)
		{
			if (glyphData == null)
			{
				throw new ArgumentNullException("glyphData");
			}

			using (MemoryStream ms = new MemoryStream(glyphData, false))
			using (BigEndianBinaryReader input = new BigEndianBinaryReader(ms))
			{
				NumberOfContours = input.ReadInt16();
				XMin = input.ReadInt16();
				YMin = input.ReadInt16();
				XMax = input.ReadInt16();
				YMax = input.ReadInt16();				
			}

			GlyphData = glyphData;
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// グリフデータを取得します。
		/// </summary>
		/// <value>グリフデータ。</value>
		public byte[] GlyphData { get; private set; }

		/// <summary>
		/// グリフの輪郭数を取得します。
		/// </summary>
		/// <value>グリフの輪郭数。</value>
		/// <remarks>本値が 0 以上であれば単一グリフ、負値であれば複合グリフを表します。</remarks>
		public short NumberOfContours { get; private set; }

		/// <summary>
		/// グリフの X 方向の最小値を取得します。
		/// </summary>
		/// <value>グリフの X 方向の最小値。</value>
		public short XMin { get; private set; }

		/// <summary>
		/// グリフの Y 方向の最小値を取得します。
		/// </summary>
		/// <value>グリフの Y 方向の最小値。</value>
		public short YMin { get; private set; }

		/// <summary>
		/// グリフの X 方向の最大値を取得します。
		/// </summary>
		/// <value>グリフの X 方向の最大値。</value>
		public short XMax { get; private set; }

		/// <summary>
		/// グリフの Y 方向の最大値を取得します。
		/// </summary>
		/// <value>グリフの Y 方向の最大値。</value>
		public short YMax { get; private set; }

		#endregion
	}
}