using System;
using System.IO;


namespace OfficeIVS.Font.Table
{

	/// <summary>
	/// OpenType フォントの hhea (Horizontal Header) テーブルを表します。
	/// </summary>
	[OpenTypeFontTable("hhea")]
	internal class HHea : OpenTypeFontTable
	{
		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.HHea"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="version">テーブルバージョン番号。</param>
		public HHea(uint version) : base(TableTags.hhea)
		{
			Version = version;
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// テーブルバージョン番号を取得または設定します。
		/// </summary>
		/// <value>テーブルバージョン番号。</value>
		/// <remarks>1.0(=0x00010000) 固定。</remarks>
		public uint Version { get; protected set; }

		#endregion

		#region メソッド

		/// <summary>
		/// フォントから 'hhea' テーブルを読み込みます。
		/// </summary>
		/// <param name="font">読み込み元となるフォント。</param>
		/// <returns><see cref="T:OfficeIVS.Font.Table.HHea"/> 型のインスタンス。テーブルが存在しない場合には <c>null</c>。</returns>
		[OpenTypeFontTableFactoryMethod]
		public static OpenTypeFontTable Load(OpenTypeFont font)
		{
			HHea table = null;

			ITableRecord tableRecord = font[TableTags.hhea];

			if (tableRecord != null)
			{
				using (Stream stream = tableRecord.GetTableStream())
				using (BigEndianBinaryReader reader = new BigEndianBinaryReader(stream))
				{
					uint version = reader.ReadUInt32();

					switch (version)
					{
					case 0x00010000:
						table = new HHea10();
						break;

					default:
						table = new HHea(version);
						break;
					}

					stream.Seek(-4, SeekOrigin.Current);
					table.LoadEntries(reader, tableRecord.Length);

					table.TableRecord = tableRecord;
				}
			}

			return table;
		}

		/// <summary>
		/// テーブル依存のエントリデータを読み込みます。
		/// </summary>
		/// <param name="input">テーブルデータストリーム。</param>
		/// <param name="length">テーブル長（バイト単位）。</param>
		protected override void LoadEntries(BigEndianBinaryReader input, long length)
		{
			Version = input.ReadUInt32();

			LoadExtendedEntries(input, length - 4);
		}

		/// <summary>
		/// テーブル依存のエントリデータを書き込みます。
		/// </summary>
		/// <param name="output">書き込み先のストリーム。</param>
		protected override void WriteEntries(BigEndianBinaryWriter output)
		{
			output.Write(Version);

			WriteExtendedEntries(output);
		}

		#endregion
	}


	/// <summary>
	/// OpenType フォントの hhea (Horizontal Header) テーブル バージョン 1.0 を表します。
	/// </summary>
	internal class HHea10 : HHea
	{
		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.HHea10"/> クラスのインスタンスを初期化します。
		/// </summary>
		public HHea10() : this(0x00010000, null)
		{
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.HHea10"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="hhea">元となる 'hhea' テーブル。</param>
		/// <exception cref="NotSupportedException"><paramref name="hhea"/> は未サポートテーブルバージョン。</exception>
		public HHea10(HHea hhea) : this(0x00010000, hhea)
		{
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.HHea10"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="version">テーブルバージョン番号。</param>
		/// <param name="hhea">元となる 'hhea' テーブル。</param>
		/// <exception cref="NotSupportedException"><paramref name="hhea"/> は未サポートテーブルバージョン。</exception>
		protected HHea10(uint version, HHea hhea) : base(version)
		{
			MetricDataFormat = 0;

			if (hhea == null)
			{
				return;
			}

			HHea10 source = hhea as HHea10;

			if (source == null)
			{
				throw new NotSupportedException();
			}

			Ascender = source.Ascender;
			Descender = source.Descender;
			LineGap = source.LineGap;
			AdvanceWidthMax = source.AdvanceWidthMax;
			MinLeftSideBearing = source.MinLeftSideBearing;
			MinRightSideBearing = source.MinRightSideBearing;
			XMaxExtent = source.XMaxExtent;
			CaretSlopeRise = source.CaretSlopeRise;
			CaretSlopeRun = source.CaretSlopeRun;
			CaretOffset = source.CaretOffset;
			MetricDataFormat = source.MetricDataFormat;
			NumberOfHMetrics = source.NumberOfHMetrics;
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// baseline から最大 ascender までの距離を取得または設定します。
		/// </summary>
		/// <value>Baseline から最大 ascender までの距離。</value>
		/// <remarks>Windows では 'OS/2' テーブルの値を参照すべき。</remarks>
		public short Ascender { get; set; }

		/// <summary>
		/// baseline から最小 descender までの距離を取得または設定します。
		/// </summary>
		/// <value>baseline から最小 descender までの距離。</value>
		/// <remarks>Windows では 'OS/2' テーブルの値を参照すべき。</remarks>
		public short Descender { get; set; }

		/// <summary>
		/// line gap を取得または設定します。
		/// </summary>
		/// <value>line gap。</value>
		/// <remarks>Windows では 'OS/2' テーブルの値を参照すべき。</remarks>
		public short LineGap { get; set; }

		/// <summary>
		/// 'hmtx' テーブル内の最大 advance width 値を取得または設定します。
		/// </summary>
		/// <value> 'hmtx' テーブル内の最大 advance width 値。</value>
		public ushort AdvanceWidthMax { get; set; }

		/// <summary>
		/// 'hmtx' テーブル内の最小 left sidebearing 値を取得または設定します。
		/// </summary>
		/// <value>'hmtx' テーブル内の最小 left sidebearing 値。</value>
		public short MinLeftSideBearing { get; set; }

		/// <summary>
		/// 最小 right sidebearing 値を取得または設定します。
		/// </summary>
		/// <value>最小 right sidebearing 値。</value>
		/// <remarks>
		/// 本値は、以下の数式により算出される。<br/>
		/// Min(advanceWidth - LeftSideBearing - (xMax - xMin))<br/>
		/// </remarks>
		public short MinRightSideBearing { get; set; }

		/// <summary>
		///
		/// </summary>
		/// <value></value>
		/// <remarks>
		/// 本値は、以下の数式により算出される。<br/>
		/// Max(LeftSideBearing + (xMax - xMin))<br/>
		/// </remarks>
		public short XMaxExtent { get; set; }

		/// <summary>
		/// カーソルの傾きを示す値を取得または設定します。
		/// </summary>
		/// <value>カーソルの傾きを示す値。</value>
		/// <remarks>(<see cref="CaretSlopeRise"/> / <see cref="CaretSlopeRun"/>) でカーソルの傾きを表す。</remarks>
		public short CaretSlopeRise { get; set; }

		/// <summary>
		/// カーソルの傾きを示す値を取得または設定します。
		/// </summary>
		/// <value>カーソルの傾きを示す値（0 の場合には垂直カーソル）。</value>
		/// <remarks>(<see cref="CaretSlopeRise"/> / <see cref="CaretSlopeRun"/>) でカーソルの傾きを表す。</remarks>
		public short CaretSlopeRun { get; set; }

		/// <summary>
		/// カーソルのオフセット量を取得または設定します。
		/// </summary>
		/// <value>カーソルのオフセット量。</value>
		/// <remarks>非傾斜の場合には 0。</remarks>
		public short CaretOffset { get; set; }

		/// <summary>
		/// グリフ水平メトリックデータ形式を取得または設定します。
		/// </summary>
		/// <value>グリフ水平メトリックデータ形式（現状 0 のみ）。</value>
		public short MetricDataFormat { get; set; }

		/// <summary>
		/// 'hmtx' テーブル内のグリフ水平メトリック要素数を取得または設定します。
		/// </summary>
		/// <value>'hmtx' テーブル内のグリフ水平メトリック要素数。</value>
		public ushort NumberOfHMetrics { get; set; }

		#endregion

		#region メソッド

		/// <summary>
		/// バージョン依存の拡張エントリデータを読み込みます。
		/// </summary>
		/// <param name="reader">読み込み元のストリーム（位置合わせ済み）。</param>
		/// <param name="length">拡張エントリデータサイズ。</param>
		protected override void LoadEntries(BigEndianBinaryReader reader, long length)
		{
			Version = reader.ReadUInt32();
			Ascender = reader.ReadInt16();
			Descender = reader.ReadInt16();
			LineGap = reader.ReadInt16();
			AdvanceWidthMax = reader.ReadUInt16();
			MinLeftSideBearing = reader.ReadInt16();
			MinRightSideBearing = reader.ReadInt16();
			XMaxExtent = reader.ReadInt16();
			CaretSlopeRise = reader.ReadInt16();
			CaretSlopeRun = reader.ReadInt16();
			CaretOffset = reader.ReadInt16();
			reader.Position += 8;	// 予約領域(short * 4)
			MetricDataFormat = reader.ReadInt16();
			NumberOfHMetrics = reader.ReadUInt16();
		}

		/// <summary>
		/// バージョン依存の拡張エントリデータを指定ストリームに書き込みます。
		/// </summary>
		/// <param name="writer">書き込み先のストリーム（位置合わせ済み）。</param>
		protected override void WriteEntries(BigEndianBinaryWriter writer)
		{
			writer.Write(Version);
			writer.Write(Ascender);
			writer.Write(Descender);
			writer.Write(LineGap);
			writer.Write(AdvanceWidthMax);
			writer.Write(MinLeftSideBearing);
			writer.Write(MinRightSideBearing);
			writer.Write(XMaxExtent);
			writer.Write(CaretSlopeRise);
			writer.Write(CaretSlopeRun);
			writer.Write(CaretOffset);
			writer.Write((long)0);		// 予約領域(short * 4)
			writer.Write(MetricDataFormat);
			writer.Write(NumberOfHMetrics);
		}

		/// <summary>
		/// 'glyf'、'hmtx' テーブルから本テーブルのエントリ値を算出し設定します。
		/// </summary>
		/// <param name="maxGlyph">グリフ数。</param>
		/// <param name="glyf">'glyf' テーブル。</param>
		/// <param name="hmtx">'hmtx' テーブル。</param>
		/// <remarks>
		/// 本テーブルの以下のエントリ値を算出し設定する。
		///  AdvanceWidthMax
		///  MinLeftSideBearing
		///  MinRightSideBearing
		///  XMaxExtent
		///  NumberOfHMetrics
		/// </remarks>
		public void SetHMetrics(ushort maxGlyph, Glyf glyf, HMtx hmtx)
		{
			// 'glyf'、'hmtx' テーブルからエントリの値を算出
			ushort maxAW = 0;
			short minLSB = Int16.MaxValue;
			short minRSB = Int16.MaxValue;
			short maxExt = Int16.MinValue;

			for (ushort glyphIndex=0; glyphIndex<maxGlyph; glyphIndex++)
			{
				GlyphDescription gd = glyf.GetGlyphDescription(glyphIndex);

				if (gd == null)
				{
					continue;
				}

				LongHMetric hm = hmtx.GetHMetric(glyphIndex);

				maxAW = Math.Max(maxAW, hm.AdvanceWidth);
				minLSB = Math.Min(minLSB, hm.LeftSideBearing);
				minRSB = Math.Min(minRSB, (short)(hm.AdvanceWidth - hm.LeftSideBearing - (gd.XMax - gd.XMin)));
				maxExt = Math.Max(maxExt, (short)(hm.LeftSideBearing + (gd.XMax - gd.XMin)));
			}
				
			// 'hhea' テーブルに計算値を設定
			AdvanceWidthMax = maxAW;
			MinLeftSideBearing = minLSB;
			MinRightSideBearing = minRSB;
			XMaxExtent = maxExt;
			NumberOfHMetrics = (ushort)hmtx.HMetrics.Length;
		}

		#endregion
	}
}