using System;
using System.IO;


namespace OfficeIVS.Font.Table
{

	/// <summary>
	/// OpenType フォント hmtx (Horizontal Metrics) テーブルを表します。
	/// </summary>
	[OpenTypeFontTable("hmtx")]
	internal sealed class HMtx : OpenTypeFontTable
	{
		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="HMtx"/> クラスのインスタンスを初期化します。
		/// </summary>
		public HMtx() : base(TableTags.hmtx)
		{
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// 各グリフの水平フォントメトリックコレクションを取得または設定します。
		/// </summary>
		/// <value>グリフの水平フォントメトリック。</value>
		/// <remarks>
		/// 要素数は 'hhea' テーブルの numOfHMetrics 参照。<br/>
		/// 等幅フォントの場合には 1 つのみ。最後のエントリは後続の全てのグリフに適用。<br/>
		/// 本プロパティの要素数と <see cref="LeftSideBearing"/> の要素数の加算結果が
		/// 'maxp' テーブルの numGlyphs と同じ値となる。
		/// </remarks>
		public LongHMetric[] HMetrics { get; set; }

		/// <summary>
		/// 各グリフの left sidebearing 値を取得または設定します。
		/// </summary>
		/// <value>グリフの left sidebearing 値。</value>
		/// <remarks>
		/// グリフのアドバンス幅は <see cref="HMetrics"/> の最終エントリを参照。<br/>
		/// 本プロパティの要素数と <see cref="LeftSideBearing"/> の要素数の加算結果が
		/// 'maxp' テーブルの numGlyphs と同じ値となる。
		/// </remarks>
		public short[] LeftSideBearing { get; set; }

		#endregion

		#region メソッド

		/// <summary>
		/// 指定番号のグリフの水平メトリック情報を取得します。
		/// </summary>
		/// <param name="glyphIndex">グリフ番号。</param>
		/// <returns>水平メトリック情報。</returns>
		/// <exception cref="InvalidOperationException"><see cref="HMetrics"/> が <c>null</c>。</exception>
		/// <exception cref="ArgumentOutOfRangeException"><paramref name="glyphIndex"/> が範囲外。</exception>
		public LongHMetric GetHMetric(ushort glyphIndex)
		{
			if (HMetrics == null)
			{
				throw new InvalidOperationException();
			}

			int numGlyphs = HMetrics.Length;

			if (LeftSideBearing != null)
			{
				numGlyphs += LeftSideBearing.Length;
			}

			if (glyphIndex >= numGlyphs)
			{
				throw new ArgumentOutOfRangeException("glyphIndex");	
			}

			LongHMetric result = new LongHMetric();

			if (glyphIndex < HMetrics.Length)
			{
				result.AdvanceWidth = HMetrics[glyphIndex].AdvanceWidth;
				result.LeftSideBearing = HMetrics[glyphIndex].LeftSideBearing;
			}
			else
			{
				result.AdvanceWidth = HMetrics[HMetrics.Length-1].AdvanceWidth;
				result.LeftSideBearing = LeftSideBearing[glyphIndex-HMetrics.Length];
			}

			return result;
		}

		/// <summary>
		/// フォントから 'hmtx' テーブルを読み込みます。
		/// </summary>
		/// <param name="font">読み込み元となるフォント。</param>
		/// <returns><see cref="T:OfficeIVS.Font.Table.HMtx"/> 型のインスタンス。テーブルが存在しない場合には <c>null</c>。</returns>
		[OpenTypeFontTableFactoryMethod]
		public static OpenTypeFontTable Load(OpenTypeFont font)
		{
			HMtx table = null;

			ITableRecord tableRecord = font[TableTags.hmtx];

			if (tableRecord != null)
			{
				HHea10 hhea = font.GetTable(TableTags.hhea) as HHea10;
				MaxP05 maxp = font.GetTable(TableTags.maxp) as MaxP05;

				if (hhea != null && maxp != null)
				{
					using (Stream stream = tableRecord.GetTableStream())
					using (BigEndianBinaryReader reader = new BigEndianBinaryReader(stream))
					{
						table = new HMtx();
						table.HMetrics = new LongHMetric[hhea.NumberOfHMetrics];
						table.LeftSideBearing = new short[Math.Max(0, maxp.NumGlyphs - hhea.NumberOfHMetrics)];
						table.LoadEntries(reader, tableRecord.Length);
						table.TableRecord = tableRecord;
					}
				}
			}

			return table;
		}

		/// <summary>
		/// テーブル依存のエントリデータを読み込みます。
		/// </summary>
		/// <param name="input">テーブルデータストリーム。</param>
		/// <param name="length">テーブル長（バイト単位）。</param>
		protected override void LoadEntries(BigEndianBinaryReader input, long length)
		{
			// HMetrics
			for (int i=0; i<HMetrics.Length; i++)
			{
				HMetrics[i] =
					new LongHMetric
					{
						AdvanceWidth = input.ReadUInt16(),
						LeftSideBearing = input.ReadInt16(),
					};
			}

			// LeftSideBearing
			for (int i=0; i<LeftSideBearing.Length; i++)
			{
				LeftSideBearing[i] = input.ReadInt16();
			}
		}

		/// <summary>
		/// テーブル依存のエントリデータを書き込みます。
		/// </summary>
		/// <param name="output">書き込み先のストリーム。</param>
		protected override void WriteEntries(BigEndianBinaryWriter output)
		{
			foreach (LongHMetric hMetric in HMetrics)
			{
				output.Write(hMetric.AdvanceWidth);
				output.Write(hMetric.LeftSideBearing);
			}

			if (LeftSideBearing != null && LeftSideBearing.Length > 0)
			{
				foreach (short lsb in LeftSideBearing)
				{
					output.Write(lsb);
				}
			}
		}

		#endregion
	}


	/// <summary>
	/// グリフの水平メトリック情報を表します。
	/// </summary>
	internal sealed class LongHMetric
	{
		#region プロパティ

		/// <summary>
		/// グリフのアドバンス幅を取得または設定します。
		/// </summary>
		/// <value>グリフのアドバンス幅。</value>
		public ushort AdvanceWidth { get; set; }

		/// <summary>
		/// グリフの left sidebearing 値を取得または設定します。
		/// </summary>
		/// <value>グリフの left sidebearing 値。</value>
		public short LeftSideBearing { get; set; }

		#endregion
	}
}