using System;
using System.IO;


namespace OfficeIVS.Font.Table
{

	/// <summary>
	/// OpenType フォントの head (Font Header) テーブルを表します。
	/// </summary>
	[OpenTypeFontTable("head")]
	internal class Head : OpenTypeFontTable
	{
		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="Head"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="version">テーブルバージョン番号。</param>
		public Head(uint version) : base(TableTags.head)
		{
			Version = version;
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// テーブルバージョン番号を取得または設定します。
		/// </summary>
		/// <value>テーブルバージョン番号。</value>
		/// <remarks>1.0 であれば 0x00010000。</remarks>
		public uint Version { get; protected set; }

		#endregion

		#region メソッド

		/// <summary>
		/// フォントから 'head' テーブルを読み込みます。
		/// </summary>
		/// <param name="font">読み込み元となるフォント。</param>
		/// <returns><see cref="T:OfficeIVS.Font.Table.Head"/> 型のインスタンス。テーブルが存在しない場合には <c>null</c>。</returns>
		[OpenTypeFontTableFactoryMethod]
		public static OpenTypeFontTable Load(OpenTypeFont font)
		{
			Head table = null;

			ITableRecord tableRecord = font[TableTags.head];

			if (tableRecord != null)
			{
				using (Stream stream = tableRecord.GetTableStream())
				using (BigEndianBinaryReader reader = new BigEndianBinaryReader(stream))
				{
					uint version = reader.ReadUInt32();

					switch (version)
					{
					case 0x00010000:
						table = new Head10();
						break;

					default:
						table = new Head(version);
						break;
					}

					stream.Seek(-4, SeekOrigin.Current);
					table.LoadEntries(reader, tableRecord.Length);

					table.TableRecord = tableRecord;
				}
			}

			return table;
		}

		/// <summary>
		/// テーブル依存のエントリデータを読み込みます。
		/// </summary>
		/// <param name="input">テーブルデータストリーム。</param>
		/// <param name="length">テーブル長（バイト単位）。</param>
		protected override void LoadEntries(BigEndianBinaryReader input, long length)
		{
			Version = input.ReadUInt32();

			LoadExtendedEntries(input, length - 4);
		}

		/// <summary>
		/// テーブル依存のエントリデータを書き込みます。
		/// </summary>
		/// <param name="output">書き込み先のストリーム。</param>
		protected override void WriteEntries(BigEndianBinaryWriter output)
		{
			output.Write(Version);

			WriteExtendedEntries(output);
		}

		#endregion
	}


	/// <summary>
	/// OpenType フォントの head (Font Header) テーブル バージョン 1.0 を表します。
	/// </summary>
	internal class Head10 : Head
	{
		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="Head10"/> クラスのインスタンスを初期化します。
		/// </summary>
		public Head10() : this(0x00010000)
		{
		}

		/// <summary>
		/// 新しい <see cref="Head10"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="version">テーブルバージョン番号。</param>
		protected Head10(uint version) : base(version)
		{
			MagicNumber = 0x5F0F3CF5;
			FontDirectionHint = 2;
			GlyphDataFormat = 0;
		}

		/// <summary>
		/// 新しい <see cref="Head10"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="head">元となる情報。</param>
		/// <exception cref="NotSupportedException"><paramref name="head"/> は未サポートテーブルバージョン。</exception>
		public Head10(Head head) : this()
		{
			Head10 source = head as Head10;

			if (source == null)
			{
				throw new NotSupportedException();
			}

			FontRevision = source.FontRevision;
			CheckSumAdjustment = 0;
			MagicNumber = source.MagicNumber;
			Flags = source.Flags;
			UnitsPerEm = source.UnitsPerEm;
			Created = source.Created;
			Modified = source.Modified;
			XMin = source.XMin;
			YMin = source.YMin;
			XMax = source.XMax;
			YMax = source.YMax;
			MacStyle = source.MacStyle;
			LowestRecPPEM = source.LowestRecPPEM;
			FontDirectionHint = source.FontDirectionHint;
			IndexToLocFormat = source.IndexToLocFormat;
			GlyphDataFormat = source.GlyphDataFormat;
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// フォントリビジョン番号を取得または設定します。
		/// </summary>
		/// <value>フォントリビジョン番号。</value>
		/// <remarks>
		/// 固定小数点（16.16）形式。<br/>
		/// 歴史的な理由により、Windows では本値は使用されない（'name' テーブルの ID5 文字列を参照）。
		/// </remarks>
		public uint FontRevision { get; set; }

		/// <summary>
		/// フォントファイルのチェックサム値を取得または設定します。
		/// </summary>
		/// <value>フォントファイルのチェックサム値。</value>
		/// <remarks>
		/// 0xB1B0AFBA からフォントファイル全体のチェックサム値を引いた値。
		/// なお、テーブルチェックサム計算時には、本プロパティ値を 0 として計算すること。
		/// </remarks>
		public uint CheckSumAdjustment { get; set; }

		/// <summary>
		/// マジックナンバーを取得または設定します。
		/// </summary>
		/// <value>マジックナンバー（0x5F0F3CF5 固定）。</value>
		public uint MagicNumber { get; private set; }

		/// <summary>
		/// フォント情報を取得または設定します。
		/// </summary>
		/// <value>フォント情報。</value>
		/// <remarks>
		/// 【値】<br/>
		/// Bit 0: Baseline for font at y=0.<br/>
		/// Bit 1: Left sidebearing point at x=0.<br/>
		/// Bit 2: Instructions may depend on point size.<br/>
		/// Bit 3: Force ppem to integer values for all internal scaler math;<br/>
		///        may use fractional ppem sizes if this bit is clear.<br/>
		/// Bit 4: Instructions may alter advance width.<br/>
		/// Bit 5-10: They are not implemented in OpenType. <br/>
		/// Bit11: Font data is 'lossless,' as a result of having been compressed<br/>
		///        and decompressed with the Agfa MicroType Express engine.<br/>
		/// Bit12: Font converted (produce compatible metrics).<br/>
		/// Bit13: Font optimized for ClearType.<br/>
		/// Bit14: Last Resort font.<br/>
		/// Bit15: Reserved, set to 0.<br/>
		/// </remarks>
		public ushort Flags { get; set; }

		/// <summary>
		/// 1 em 当りの Units 値を取得または設定します。
		/// </summary>
		/// <value>1 em 当りの Units 値（16～16384）。</value>
		/// <remarks>本値は 2 の冪乗でなければいけない。</remarks>
		public ushort UnitsPerEm { get; set; }

		/// <summary>
		/// 作成日時を取得または設定します。
		/// </summary>
		/// <value>作成日時。</value>
		/// <remarks>1904/1/1 0:00:00 からの経過秒数。</remarks>
		public long Created { get; set; }

		/// <summary>
		/// 編集日時を取得または設定します。
		/// </summary>
		/// <value>編集日時。</value>
		/// <remarks>1904/1/1 0:00:00 からの経過秒数。</remarks>
		public long Modified { get; set; }

		/// <summary>
		/// 全グリフのバウンディングボックスの X 方向の最小値を取得または設定します。
		/// </summary>
		/// <value>全グリフのバウンディングボックスの X 方向の最小値。</value>
		public short XMin { get; set; }

		/// <summary>
		/// 全グリフのバウンディングボックスの Y 方向の最小値を取得または設定します。
		/// </summary>
		/// <value>全グリフのバウンディングボックスの Y 方向の最小値。</value>
		public short YMin { get; set; }

		/// <summary>
		/// 全グリフのバウンディングボックスの X 方向の最大値を取得または設定します。
		/// </summary>
		/// <value>全グリフのバウンディングボックスの X 方向の最大値。</value>
		public short XMax { get; set; }

		/// <summary>
		/// 全グリフのバウンディングボックスの Y 方向の最大値を取得または設定します。
		/// </summary>
		/// <value>全グリフのバウンディングボックスの Y 方向の最大値。</value>
		public short YMax { get; set; }

		/// <summary>
		/// フォントスタイル情報を取得または設定します。
		/// </summary>
		/// <value>フォントスタイル情報。</value>
		/// <remarks>
		/// 本値は 'OS/2' テーブルの fsSelection と一致させる必要がある。<br/>
		/// なお、Windows では 'OS/2' テーブルの fsSelection が使用される。<br/>
		/// <br/>
		/// 【値】<br/>
		/// Bit 0: Bold (if set to 1)<br/>
		/// Bit 1: Italic (if set to 1)<br/>
		/// Bit 2: Underline (if set to 1)<br/>
		/// Bit 3: Outline (if set to 1)<br/>
		/// Bit 4: Shadow (if set to 1)<br/>
		/// Bit 5: Condensed (if set to 1)<br/>
		/// Bit 6: Extended (if set to 1)<br/>
		/// Bit 7: Reserved (set to 0)<br/>
		/// </remarks>
		public ushort MacStyle { get; set; }

		/// <summary>
		/// 可視可能な最小 em 単位当りのピクセル数を取得または設定します。
		/// </summary>
		/// <value>可視可能な最小 em 単位当りのピクセル数。</value>
		public ushort LowestRecPPEM { get; set; }

		/// <summary>
		/// フォント方向ヒントの取得または設定します 。
		/// </summary>
		/// <value>フォント方向ヒント。</value>
		/// <remarks>
		/// OpenType フォントでは使用されない（2 固定）。<br/>
		/// <br/>
		/// 【値】<br/>
		///  0 ... Fully mixed directional glyphs.<br/>
		///  1 ... Only strongly left to right.<br/>
		///  2 ... Like 1 but also contains neutrals.<br/>
		/// -1 ... Only strongly right to left.<br/>
		/// -2 ... Like -1 but also contains neutrals.<br/>
		/// </remarks>
		public short FontDirectionHint { get; set; }

		/// <summary>
		/// 'loca' テーブル書式を取得または設定します。
		/// </summary>
		/// <value>'loca' テーブルの書式を表す値（0:short 1:long）。</value>
		/// <remarks>
		/// 【値】<br/>
		/// 0 ... short offset format.<br/>
		/// 1 ... long offset format.<br/>
		/// </remarks>
		public short IndexToLocFormat { get; set; }

		/// <summary>
		/// グリフのデータ形式を取得または設定します。
		/// </summary>
		/// <value>グリフのデータ形式（現状 0 固定）。</value>
		public short GlyphDataFormat { get; set; }

		#endregion

		#region メソッド

		/// <summary>
		/// テーブル依存のエントリデータを読み込みます。
		/// </summary>
		/// <param name="input">テーブルデータストリーム。</param>
		/// <param name="length">テーブル長（バイト単位）。</param>
		protected override void LoadEntries(BigEndianBinaryReader input, long length)
		{
			Version = input.ReadUInt32();
			FontRevision = input.ReadUInt32();
			CheckSumAdjustment = input.ReadUInt32();
			MagicNumber = input.ReadUInt32();
			Flags = input.ReadUInt16();
			UnitsPerEm = input.ReadUInt16();
			Created = input.ReadInt64();
			Modified = input.ReadInt64();
			XMin = input.ReadInt16();
			YMin = input.ReadInt16();
			XMax = input.ReadInt16();
			YMax = input.ReadInt16();
			MacStyle = input.ReadUInt16();
			LowestRecPPEM = input.ReadUInt16();
			FontDirectionHint = input.ReadInt16();
			IndexToLocFormat = input.ReadInt16();
			GlyphDataFormat = input.ReadInt16();
		}

		/// <summary>
		/// テーブル依存のエントリデータを書き込みます。
		/// </summary>
		/// <param name="output">書き込み先のストリーム。</param>
		protected override void WriteEntries(BigEndianBinaryWriter output)
		{
			output.Write(Version);
			output.Write(FontRevision);
			output.Write(CheckSumAdjustment);
			output.Write(MagicNumber);
			output.Write(Flags);
			output.Write(UnitsPerEm);
			output.Write(Created);
			output.Write(Modified);
			output.Write(XMin);
			output.Write(YMin);
			output.Write(XMax);
			output.Write(YMax);
			output.Write(MacStyle);
			output.Write(LowestRecPPEM);
			output.Write(FontDirectionHint);
			output.Write(IndexToLocFormat);
			output.Write(GlyphDataFormat);
		}

		/// <summary>
		/// 指定 'glyf' テーブルを元にグリフバウンディングボックスの最小最大値を設定します。
		/// </summary>
		/// <param name="glyf">'glyf' テーブル。</param>
		public void SetGlyphBoundingBox(Glyf glyf)
		{
			if (glyf == null)
			{
				return;
			}

			// 'glyf' テーブル内の全グリフバウンディングボックスの最小最大値を取得
			short xMin, yMin, xMax, yMax;
			glyf.GetGlyphBoundingBox(out xMin, out yMin, out xMax, out yMax);

			// テーブルに設定
			XMin = xMin;
			YMin = yMin;
			XMax = xMax;
			YMax = yMax;
		}

		#endregion		
	}
}