using System;
using System.IO;


namespace OfficeIVS.Font.Table
{

	/// <summary>
	/// TrueType フォントの loca (Index To Location) テーブルを表します。
	/// </summary>
	[OpenTypeFontTable("loca")]
	internal sealed class Loca : OpenTypeFontTable
	{
		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="Loca"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="format">オフセット形式（0:short 1:long）。</param>
		public Loca(short format=(short)1) : base(TableTags.loca)
		{
			Format = format == 0 ? (short)0 : (short)1;
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// オフセット保存形式を取得または設定します。
		/// </summary>
		/// <value>オフセット保存形式（0:short、1:long[既定]）</value>
		public short Format { get; set; }

		/// <summary>
		/// グリフデータの 'glyf' テーブル先頭からのオフセット量コレクションを取得または設定します。
		/// </summary>
		/// <value>グリフデータの 'glyf' テーブル先頭からのオフセット量コレクション。</value>
		/// <remarks>
		/// ・要素数は 'maxp' テーブルの numGlyphs + 1（末尾グリフのサイズ計算のため余分なエントリあり）。<br/>
		/// ・コレクションは常にオフセット値で昇順でソートされていなければならない。<br/>
		/// ・本テーブルは ushort 版、uint 版が存在する。<br/>
		///   ushort 版か uint 版かは 'head' テーブル indexToLocFormat の値により区別する。<br/>
		///   なお ushort 版の場合、値は実オフセット量の 1/2 が格納されている。<br/>
		/// </remarks>
		public uint[] Offsets { get; set; }

		#endregion

		#region パブリック メソッド

		/// <summary>
		/// フォントから 'loca' テーブルを読み込みます。
		/// </summary>
		/// <param name="font">読み込み元となるフォント。</param>
		/// <returns><see cref="T:OfficeIVS.Font.Table.Loca"/> 型のインスタンス。テーブルが存在しない場合には <c>null</c>。</returns>
		[OpenTypeFontTableFactoryMethod]
		public static OpenTypeFontTable Load(OpenTypeFont font)
		{
			Loca table = null;

			ITableRecord tableRecord = font[TableTags.loca];

			if (tableRecord != null)
			{
				Head10 head = font.GetTable(TableTags.head) as Head10;
				MaxP05 maxp = font.GetTable(TableTags.maxp) as MaxP05;

				if (head != null && maxp != null)
				{
					using (Stream stream = tableRecord.GetTableStream())
					using (BigEndianBinaryReader reader = new BigEndianBinaryReader(stream))
					{
						table = new Loca(head.IndexToLocFormat);
						table.Offsets = new uint[maxp.NumGlyphs+1];
						table.LoadEntries(reader, tableRecord.Length);
						table.TableRecord = tableRecord;
					}
				}
			}

			return table;
		}

		/// <summary>
		/// 指定グリフ番号をもつグリフデータのサイズを取得します。
		/// </summary>
		/// <param name="glyphId">グリフ番号。</param>
		/// <returns>グリフデータのサイズ（バイト単位）。</returns>
		/// <exception cref="InvalidOperationException"><see cref="Offsets"/> が null。</exception>
		/// <exception cref="ArgumentOutOfRangeException"><paramref name="glyphId"/> が範囲外。</exception>
		public uint GetGlyphDataLength(ushort glyphId)
		{
			if (Offsets == null)
			{
				throw new InvalidOperationException();
			}

			if (Offsets.Length-1 < glyphId)
			{
				throw new ArgumentOutOfRangeException("glyphId");
			}

			return Offsets[glyphId+1] - Offsets[glyphId];
		}

		#endregion

		#region プロテクト メソッド

		/// <summary>
		/// テーブル依存のエントリデータを読み込みます。
		/// </summary>
		/// <param name="input">テーブルデータストリーム。</param>
		/// <param name="length">テーブル長（バイト単位）。</param>
		protected override void LoadEntries(BigEndianBinaryReader input, long length)
		{
			if (Format == 0)
			{
				for (int i=0; i<Offsets.Length; i++)
				{
					ushort value = input.ReadUInt16();
					Offsets[i] = Convert.ToUInt32(2 * value);
				}
			}
			else
			{
				for (int i=0; i<Offsets.Length; i++)
				{
					Offsets[i] = input.ReadUInt32();
				}						
			}
		}

		/// <summary>
		/// テーブル依存のエントリデータを書き込みます。
		/// </summary>
		/// <param name="output">書き込み先のストリーム。</param>
		protected override void WriteEntries(BigEndianBinaryWriter output)
		{
			if (Offsets == null || Offsets.Length == 0)
			{
				return;
			}

			if (Format == 0)
			{
				foreach (uint offset in Offsets)
				{
					output.Write(Convert.ToUInt16(offset/2));
				}
			}
			else
			{
				foreach (uint offset in Offsets)
				{
					output.Write(offset);
				}
			}
		}

		#endregion
	}
}