using System;
using System.IO;


namespace OfficeIVS.Font.Table
{

	/// <summary>
	/// OpenType フォントの maxp(Maximum Profile) テーブルを表します。
	/// </summary>
	[OpenTypeFontTable("maxp")]
	internal class MaxP : OpenTypeFontTable
	{
		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="MaxP"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <value>テーブルバージョン番号。</value>
		public MaxP(uint tableVersion) : base(TableTags.maxp)
		{
			Version = tableVersion;
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// テーブルバージョン番号を取得または設定します。
		/// </summary>
		/// <value>テーブルバージョン番号。</value>
		/// <remarks>
		/// 0.5(=0x00005000) : CFF<br/>
		/// 1.0(=0x00010000) : TrueType<br/>
		/// </remarks>
		public uint Version { get; protected set; }

		#endregion

		#region メソッド

		/// <summary>
		/// フォントから 'maxp' テーブルを読み込みます。
		/// </summary>
		/// <param name="font">読み込み元となるフォント。</param>
		/// <returns><see cref="T:OfficeIVS.Font.Table.MaxP"/> 型のインスタンス。テーブルが存在しない場合には <c>null</c>。</returns>
		[OpenTypeFontTableFactoryMethod]
		public static OpenTypeFontTable Load(OpenTypeFont font)
		{
			MaxP table = null;

			ITableRecord tableRecord = font[TableTags.maxp];

			if (tableRecord != null)
			{
				using (Stream stream = tableRecord.GetTableStream())
				using (BigEndianBinaryReader reader = new BigEndianBinaryReader(stream))
				{
					uint version = reader.ReadUInt32();

					switch (version)
					{
					case 0x00005000: 
						table = new MaxP05();
						break;

					case 0x00010000:
						table = new MaxP10();
						break;

					default:
						table = new MaxP(version);
						break;
					}

					stream.Seek(-4, SeekOrigin.Current);
					table.LoadEntries(reader, tableRecord.Length);

					table.TableRecord = tableRecord;
				}
			}

			return table;
		}

		/// <summary>
		/// テーブル依存のエントリデータを読み込みます。
		/// </summary>
		/// <param name="input">テーブルデータストリーム。</param>
		/// <param name="length">テーブル長（バイト単位）。</param>
		protected override void LoadEntries(BigEndianBinaryReader input, long length)
		{
			Version = input.ReadUInt32();

			LoadExtendedEntries(input, length - 4);
		}

		/// <summary>
		/// テーブル依存のエントリデータを書き込みます。
		/// </summary>
		/// <param name="output">書き込み先のストリーム。</param>
		protected override void WriteEntries(BigEndianBinaryWriter output)
		{
			output.Write(Version);

			WriteExtendedEntries(output);
		}

		#endregion
	}


	/// <summary>
	/// OpenType フォントの maxp (Maximum Profile) テーブル（バージョン 0.5）を表します。
	/// </summary>
	internal class MaxP05 : MaxP
	{
		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.MaxP05"/> クラスのインスタンスを初期化します。
		/// </summary>
		public MaxP05() : this(0x00005000)
		{
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.MaxP05"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="version">テーブルバージョン番号。</param>
		protected MaxP05(uint version) : base(version)
		{
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.MaxP05"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="maxp">元となる情報。</param>
		/// <exception cref="NotSupportedException"><paramref name="maxp"/> は未サポートテーブルバージョン。</exception>
		public MaxP05(MaxP maxp) : this()
		{
			MaxP05 source = maxp as MaxP05;

			if (source == null)
			{
				throw new NotSupportedException();
			}

			NumGlyphs = source.NumGlyphs;
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// フォントに含まれるグリフ数を取得または設定します。
		/// </summary>
		/// <value>グリフ数。</value>
		public ushort NumGlyphs { get; set; }

		#endregion

		#region メソッド

		/// <summary>
		/// テーブル依存のエントリデータを読み込みます。
		/// </summary>
		/// <param name="input">テーブルデータストリーム。</param>
		/// <param name="length">テーブル長（バイト単位）。</param>
		protected override void LoadEntries(BigEndianBinaryReader input, long length)
		{
			Version = input.ReadUInt32();
			NumGlyphs = input.ReadUInt16();
		}

		/// <summary>
		/// テーブル依存のエントリデータを書き込みます。
		/// </summary>
		/// <param name="output">書き込み先のストリーム。</param>
		protected override void WriteEntries(BigEndianBinaryWriter output)
		{
			output.Write(Version);
			output.Write(NumGlyphs);
		}

		#endregion
	}


	/// <summary>
	/// OpenType フォントの maxp (Maximum Profile) テーブル（バージョン 1.0）を表します。
	/// </summary>
	internal class MaxP10 : MaxP05
	{
		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="MaxP10"/> クラスのインスタンスを初期化します。
		/// </summary>
		public MaxP10() : base(0x00010000)
		{
			MaxZones = 2;
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.MaxP10"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="maxp">元となる情報。</param>
		/// <exception cref="NotSupportedException"><paramref name="maxp"/> は未サポートテーブルバージョン。</exception>
		public MaxP10(MaxP maxp) : this()
		{
			MaxP10 source = maxp as MaxP10;

			if (source == null)
			{
				throw new NotSupportedException();
			}

			NumGlyphs = source.NumGlyphs;
			MaxPoints = source.MaxPoints;
			MaxContours = source.MaxContours;
			MaxCompositePoints = source.MaxCompositePoints;
			MaxCompositeContours = source.MaxCompositeContours;
			MaxZones = source.MaxZones;
			MaxTwilightPoints = source.MaxTwilightPoints;
			MaxStorage = source.MaxStorage;
			MaxFunctionDefs = source.MaxFunctionDefs;
			MaxInstructionDefs = source.MaxInstructionDefs;
			MaxStackElements = source.MaxStackElements;
			MaxSizeOfInstructions = source.MaxSizeOfInstructions;
			MaxComponentElements = source.MaxComponentElements;
			MaxComponentDepth = source.MaxComponentDepth;
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// 非複合グリフのポイント数の最大値を取得または設定します。
		/// </summary>
		/// <value>非複合グリフのポイント数の最大値。</value>
		public ushort MaxPoints { get; set; }

		/// <summary>
		/// 非複合グリフの輪郭数の最大値を取得または設定します。
		/// </summary>
		/// <value>非複合グリフの輪郭数の最大値。</value>
		public ushort MaxContours { get; set; }

		/// <summary>
		/// 複合グリフのポイント数の最大値を取得または設定します。
		/// </summary>
		/// <value>複合グリフのポイント数の最大値。</value>
		public ushort MaxCompositePoints { get; set; }

		/// <summary>
		/// 複合グリフの輪郭数の最大値を取得または設定します。
		/// </summary>
		/// <value>複合グリフの輪郭数の最大値。</value>
		public ushort MaxCompositeContours { get; set; }

		/// <summary>
		/// 最大ゾーン数を取得または設定します。
		/// </summary>
		/// <value>最大ゾーン数。</value>
		/// <remarks>twilight zone (Z0) を使用しない場合には 1、使用する場合には 2。通常 2 を設定。</remarks>
		public ushort MaxZones { get; set; }

		/// <summary>
		/// グリフの twilight ポイント数の最大値を取得または設定します。
		/// </summary>
		/// <value>twilight ポイント数の最大値。</value>
		public ushort MaxTwilightPoints { get; set; }

		/// <summary>
		/// Strage Area Location の最大数を取得または設定します。
		/// </summary>
		/// <value>Strage Area Location の最大数。</value>
		public ushort MaxStorage { get; set; }

		/// <summary>
		/// グリフの最大ファンクション数を取得または設定します。
		/// </summary>
		/// <value>グリフの最大ファンクション数。</value>
		public ushort MaxFunctionDefs { get; set; }

		/// <summary>
		/// グリフの最大命令数を取得または設定します。
		/// </summary>
		/// <value>グリフの最大ファンクション数。</value>
		public ushort MaxInstructionDefs { get; set; }

		/// <summary>
		/// グリフの最大スタック数を取得または設定します。
		/// </summary>
		/// <value>最大スタック数。</value>
		public ushort MaxStackElements { get; set; }

		/// <summary>
		/// グリフの最大命令サイズを取得または設定します。
		/// </summary>
		/// <value>グリフの最大命令サイズ（バイト単位）。</value>
		public ushort MaxSizeOfInstructions { get; set; }

		/// <summary>
		/// 複合グリフを構成する単一グリフの最大数を取得または設定します。
		/// </summary>
		/// <value>複合グリフを構成する単一グリフの最大数。</value>
		public ushort MaxComponentElements { get; set; }

		/// <summary>
		/// 複合グリフを構築するのに必要な最大再帰レベルを取得または設定します。
		/// </summary>
		/// <value>複合グリフを構築するのに必要な最大再帰レベル。</value>
		public ushort MaxComponentDepth { get; set; }

		#endregion

		#region メソッド

		/// <summary>
		/// テーブル依存のエントリデータを読み込みます。
		/// </summary>
		/// <param name="input">テーブルデータストリーム。</param>
		/// <param name="length">テーブル長（バイト単位）。</param>
		protected override void LoadEntries(BigEndianBinaryReader input, long length)
		{
			base.LoadEntries(input, length);

			MaxPoints = input.ReadUInt16();
			MaxContours = input.ReadUInt16();
			MaxCompositePoints = input.ReadUInt16();
			MaxCompositeContours = input.ReadUInt16();
			MaxZones = input.ReadUInt16();
			MaxTwilightPoints = input.ReadUInt16();
			MaxStorage = input.ReadUInt16();
			MaxFunctionDefs = input.ReadUInt16();
			MaxInstructionDefs = input.ReadUInt16();
			MaxStackElements = input.ReadUInt16();
			MaxSizeOfInstructions = input.ReadUInt16();
			MaxComponentElements = input.ReadUInt16();
			MaxComponentDepth = input.ReadUInt16();
		}

		/// <summary>
		/// テーブル依存のエントリデータを書き込みます。
		/// </summary>
		/// <param name="output">書き込み先のストリーム。</param>
		protected override void WriteEntries(BigEndianBinaryWriter output)
		{
			base.WriteEntries(output);

			output.Write(MaxPoints);
			output.Write(MaxContours);
			output.Write(MaxCompositePoints);
			output.Write(MaxCompositeContours);
			output.Write(MaxZones);
			output.Write(MaxTwilightPoints);
			output.Write(MaxStorage);
			output.Write(MaxFunctionDefs);
			output.Write(MaxInstructionDefs);
			output.Write(MaxStackElements);
			output.Write(MaxSizeOfInstructions);
			output.Write(MaxComponentElements);
			output.Write(MaxComponentDepth);
		}

		#endregion
	}
}