using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Text;


namespace OfficeIVS.Font.Table
{

	/// <summary>
	/// OpenType フォントの name (Naming) テーブルを表します。
	/// </summary>
	[OpenTypeFontTable("name")]
	internal class Name : OpenTypeFontTable
	{
		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.Name"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="format">テーブルフォーマット番号。</param>
		public Name(ushort format) : base(TableTags.name)
		{
			Format = format;
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// テーブルのフォーマット番号を取得します。
		/// </summary>
		/// <value>テーブルのフォーマット番号。</value>
		public ushort Format { get; protected set; }

		#endregion

		#region パブリック メソッド

		/// <summary>
		/// フォントから 'name' テーブルを読み込みます。
		/// </summary>
		/// <param name="font">読み込み元となるフォント。</param>
		/// <returns><see cref="T:OfficeIVS.Font.Table.Name"/> 型のインスタンス。テーブルが存在しない場合には <c>null</c>。</returns>
		[OpenTypeFontTableFactoryMethod]
		public static OpenTypeFontTable Load(OpenTypeFont font)
		{
			Name table = null;

			ITableRecord tableRecord = font[TableTags.name];

			if (tableRecord != null)
			{
				using (Stream stream = tableRecord.GetTableStream())
				using (BigEndianBinaryReader reader = new BigEndianBinaryReader(stream))
				{
					ushort format = reader.ReadUInt16();

					switch (format)
					{
					case 0x0000:
						table = new Name00();
						break;

					case 0x0001:
						table = new Name10();
						break;

					default:
						table = new Name(format);
						break;
					}

					stream.Seek(-2, SeekOrigin.Current);
					table.LoadEntries(reader, tableRecord.Length);

					table.TableRecord = tableRecord;
				}
			}

			return table;
		}

		#endregion

		#region プロテクト メソッド

		/// <summary>
		/// テーブル依存のエントリデータを読み込みます。
		/// </summary>
		/// <param name="input">テーブルデータストリーム。</param>
		/// <param name="length">テーブル長（バイト単位）。</param>
		protected override void LoadEntries(BigEndianBinaryReader input, long length)
		{
			Format = input.ReadUInt16();

			LoadExtendedEntries(input, length - 2);
		}

		/// <summary>
		/// テーブル依存のエントリデータを書き込みます。
		/// </summary>
		/// <param name="output">書き込み先のストリーム。</param>
		protected override void WriteEntries(BigEndianBinaryWriter output)
		{
			output.Write(Format);

			WriteExtendedEntries(output);
		}

		#endregion
	}


	/// <summary>
	/// OpenType フォントの name (Naming) テーブル フォーマット 0 を表します。
	/// </summary>
	internal class Name00 : Name
	{
		#region フィールド

		/// <summary>
		/// Windows プラットフォーム用の名前テーブル。
		/// </summary>
		/// <remarks>言語タグ、NameID, 文字列のマッピング。</remarks>
		protected readonly Dictionary<string, Dictionary<ushort, string>> _windowsNameTables;

		#endregion

		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.Name00"/> クラスのインスタンスを初期化します。
		/// </summary>
		public Name00() : this(0)
		{
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.Name00"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="format">フォーマット番号。</param>
		protected Name00(ushort format) : base(format)
		{
			_windowsNameTables = new Dictionary<string, Dictionary<ushort, string>>();
		}

		#endregion

		#region パブリック メソッド

		/// <summary>
		/// 指定言語の指定項目に設定されている文字列を取得します。
		/// </summary>
		/// <param name="languageTag">
		/// IETF 言語タグ（RFC 3066）。ja-JPなど。
		/// null 指定の場合には最初に見つかった <paramref name="nameId"/> の値を返します。
		/// </param>
		/// <param name="nameId">名前ID。</param>
		/// <returns>文字列。未設定の場合には null。</returns>
		public string GetNameValue(string languageTag, ushort nameId)
		{
			string nameValue = null;

			if (languageTag == null)
			{
				foreach (string langTag in _windowsNameTables.Keys)
				{
					if (_windowsNameTables[langTag].TryGetValue(nameId, out nameValue))
					{
						break;
					}
				}
			}
			else if (_windowsNameTables.ContainsKey(languageTag))
			{
				_windowsNameTables[languageTag].TryGetValue(nameId, out nameValue);
			}

			return nameValue;
		}

		/// <summary>
		/// 指定言語の指定項目に文字列を設定します。
		/// </summary>
		/// <param name="languageTag">IETF 言語タグ（RFC 3066）。ja-JPなど。</param>
		/// <param name="nameId">名前ID。</param>
		/// <param name="value">文字列。項目削除の場合には null。</param>
		public void SetNameValue(string languageTag, ushort nameId, string value)
		{
			Dictionary<ushort, string> valueTable;
			_windowsNameTables.TryGetValue(languageTag, out valueTable);

			if (value != null)
			{
				if (valueTable == null)
				{
					_windowsNameTables[languageTag] =
						(valueTable = new Dictionary<ushort, string>());
				}

				valueTable[nameId] = value;
			}
			else if (valueTable != null)
			{
				valueTable.Remove(nameId);

				if (valueTable.Count == 0)
				{
					_windowsNameTables.Remove(languageTag);
				}
			}
		}

		#endregion

		#region プロテクト メソッド

		/// <summary>
		/// テーブル依存のエントリデータを読み込みます。
		/// </summary>
		/// <param name="input">テーブルデータストリーム。</param>
		/// <param name="length">テーブル長（バイト単位）。</param>
		protected override void LoadEntries(BigEndianBinaryReader input, long length)
		{
			long tableTopPos = input.Position;

			Format = input.ReadUInt16();
			ushort count = input.ReadUInt16();
			long stringStorageOffset = tableTopPos + input.ReadUInt16();

			// 文字列ストレージ読み込み
			byte[] stringStorage = new byte[length-stringStorageOffset];
			using (new StreamPositionStack(input))
			{
				input.Position = stringStorageOffset;
				input.Read(stringStorage, 0, stringStorage.Length);
			}

			// レコード読み込み
			for (int i=0; i<count; i++)
			{
				ushort platformId = input.ReadUInt16();
				ushort encodingId = input.ReadUInt16();

				// platformId=windows  かつ encodingId=unicode 以外は無視
				if (platformId != 3 || encodingId != 1)
				{
					input.Position += 8;
					continue;
				}

				string languageTag = GetLanguageTag(input.ReadUInt16());
				
				// 言語タグ変換できないない場合には無視
				if (languageTag == null)
				{
					continue;
				}

				ushort nameId = input.ReadUInt16();
				ushort stringLength = input.ReadUInt16();
				ushort stringOffset = input.ReadUInt16();

				try
				{
					if (!_windowsNameTables.ContainsKey(languageTag))
					{
						_windowsNameTables[languageTag] = new Dictionary<ushort, string>();
					}

					_windowsNameTables[languageTag][nameId] =
						Encoding.BigEndianUnicode.GetString(stringStorage, stringOffset, stringLength);
				}
				catch
				{
					continue;
				}
			}
		}

		/// <summary>
		/// テーブル依存のエントリデータを書き込みます。
		/// </summary>
		/// <param name="output">書き込み先のストリーム。</param>
		protected override void WriteEntries(BigEndianBinaryWriter output)
		{
			// format
			output.Write(Format);

			// count
			int count = 0;
			foreach (var nameTable in _windowsNameTables.Values)
			{
				count += nameTable.Count;
			}
			output.Write((ushort)count);

			// stringOffset
			output.Write((ushort)(6 + count * 12));

			MemoryStream stringStorage = new MemoryStream();

			foreach (string languageTag in _windowsNameTables.Keys)
			{
				var nameTables = _windowsNameTables[languageTag];

				ushort languageId = GetLanguageId(languageTag);

				if (languageId == 0)
				{
					continue;
				}

				foreach (ushort nameId in nameTables.Keys)
				{
					// platformId, encodingId, languageId, nameId
					output.Write((ushort)3);		
					output.Write((ushort)1);
					output.Write(languageId);
					output.Write(nameId);

					// length, offset
					byte[] nameValue = Encoding.BigEndianUnicode.GetBytes(nameTables[nameId]);
					output.Write((ushort)nameValue.Length);
					output.Write((ushort)stringStorage.Position);
					stringStorage.Write(nameValue, 0, nameValue.Length);
				}
			}

			// string storage
			stringStorage.Flush();
			output.Write(stringStorage.ToArray());
			stringStorage.Close();
		}

		#endregion

		#region プライベート メソッド

		/// <summary>
		/// 言語 ID から IETF 言語タグ（RFC 3066）を取得します。
		/// </summary>
		/// <param name="languageId">言語 ID。</param>
		/// <returns>IETF 言語タグ。対応する言語タグが存在しない場合には null。</returns>
		private static string GetLanguageTag(ushort languageId)
		{
			string langTag = null;

			try
			{
				if (languageId < 0x8000)
				{
					langTag = CultureInfo.GetCultureInfo(languageId).Name;
				}
			}
			catch (CultureNotFoundException)
			{
			}

			return langTag;
		}

		/// <summary>
		/// IETF 言語タグから言語 ID を取得します。
		/// </summary>
		/// <param name="languageTag">IETF 言語タグ（RFC 3066）。</param>
		/// <returns>Windows 言語 ID。対応する言語 ID が存在しない場合には 0。</returns>
		private static ushort GetLanguageId(string languageTag)
		{
			ushort languageId = 0;

			try
			{
				languageId = (ushort)CultureInfo.CreateSpecificCulture(languageTag).LCID;
			}
			catch (CultureNotFoundException)
			{
			}

			return languageId;
		}

		#endregion
	}


	/// <summary>
	/// OpenType フォントの name (Naming) テーブル フォーマット 1 を表します。
	/// </summary>
	internal class Name10 : Name00
	{
		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.Name10"/> クラスのインスタンスを初期化します。
		/// </summary>
		public Name10() : base(1)
		{
		}

		#endregion

		#region プロテクト メソッド

		/// <summary>
		/// テーブル依存のエントリデータを読み込みます。
		/// </summary>
		/// <param name="input">テーブルデータストリーム。</param>
		/// <param name="length">テーブル長（バイト単位）。</param>
		protected override void LoadEntries(BigEndianBinaryReader input, long length)
		{
			long tableTopPos = input.Position;

			// format
			Format = input.ReadUInt16();

			// count
			ushort count = input.ReadUInt16();

			// stringStorageOffset
			long stringStorageOffset = tableTopPos + input.ReadUInt16();

			long currentPos = input.Position;

			// 文字列ストレージ読み込み
			byte[] stringStorage = new byte[length-stringStorageOffset];
			input.Position = stringStorageOffset;
			input.Read(stringStorage, 0, stringStorage.Length);

			// 言語タグレコード読み込み
			input.Position = currentPos + 12 * count;
			ushort langTagCount = input.ReadUInt16();
			string[] languageTagRecords = new string[langTagCount];

			for (int i=0; i<langTagCount; i++)
			{
				ushort stringLength = input.ReadUInt16();
				ushort stringOffset = input.ReadUInt16();
				languageTagRecords[i] =
					Encoding.BigEndianUnicode.GetString(stringStorage, stringOffset, stringLength);
			}

			input.Position = currentPos;

			// レコード読み込み
			for (int i=0; i<count; i++)
			{
				ushort platformId = input.ReadUInt16();
				ushort encodingId = input.ReadUInt16();

				// platformId=windows  かつ encodingId=unicode 以外は無視
				if (platformId != 3 || encodingId != 1)
				{
					input.Position += 8;
					continue;
				}

				string languageTag = GetLanguageTag(input.ReadUInt16(), languageTagRecords);
				
				// 言語タグ変換できないない場合には無視
				if (languageTag == null)
				{
					continue;
				}

				ushort nameId = input.ReadUInt16();
				ushort stringLength = input.ReadUInt16();
				ushort stringOffset = input.ReadUInt16();

				try
				{
					if (!_windowsNameTables.ContainsKey(languageTag))
					{
						_windowsNameTables[languageTag] = new Dictionary<ushort, string>();
					}

					_windowsNameTables[languageTag][nameId] =
						Encoding.BigEndianUnicode.GetString(stringStorage, stringOffset, stringLength);
				}
				catch
				{
					continue;
				}
			}
		}

		/// <summary>
		/// テーブル依存のエントリデータを書き込みます。
		/// </summary>
		/// <param name="output">書き込み先のストリーム。</param>
		protected override void WriteEntries(BigEndianBinaryWriter output)
		{
			// format
			output.Write(Format);

			// count
			int count = 0;
			foreach (var nameTable in _windowsNameTables.Values)
			{
				count += nameTable.Count;
			}
			output.Write((ushort)count);

			// stringOffset
			output.Write((ushort)(count * 12 + 4));

			List<string> languageTagRecords = new List<string>();
			MemoryStream stringStorage = new MemoryStream();

			foreach (string languageTag in _windowsNameTables.Keys)
			{
				var nameTables = _windowsNameTables[languageTag];

				ushort languageId = GetLanguageId(languageTag, languageTagRecords);

				foreach (ushort nameId in nameTables.Keys)
				{
					// platformId, encodingId, languageId, nameId
					output.Write((ushort)3);		
					output.Write((ushort)1);
					output.Write(languageId);
					output.Write(nameId);

					// length, offset
					byte[] nameValue = Encoding.BigEndianUnicode.GetBytes(nameTables[nameId]);
					output.Write((ushort)nameValue.Length);
					output.Write((ushort)stringStorage.Position);
					stringStorage.Write(nameValue, 0, nameValue.Length);
				}
			}

			// langTagCount
			output.Write((ushort)languageTagRecords.Count);

			// langTagRecord
			foreach (string langTag in languageTagRecords)
			{
				byte[] langTagValue = Encoding.BigEndianUnicode.GetBytes(langTag);
				output.Write((ushort)langTagValue.Length);
				output.Write((ushort)stringStorage.Position);
				stringStorage.Write(langTagValue, 0, langTagValue.Length);
			}

			// string storage
			stringStorage.Flush();
			output.Write(stringStorage.ToArray());
			stringStorage.Close();
		}

		#endregion

		#region プライベート メソッド

		/// <summary>
		/// 言語 ID から IETF 言語タグ（RFC 3066）を取得します。
		/// </summary>
		/// <param name="languageId">言語 ID。</param>
		/// <param name="languageTagRecords">言語タグレコード配列。</param>
		/// <returns>IETF 言語タグ。対応する言語タグが存在しない場合には null。</returns>
		private static string GetLanguageTag(ushort languageId, IList<string> languageTagRecords)
		{
			string langTag = null;

			try
			{
				if (languageId < 0x8000)
				{
					langTag = CultureInfo.GetCultureInfo(languageId).Name;
				}
				else if (languageTagRecords != null)
				{
					languageId -= 0x8000;

					if (languageId < languageTagRecords.Count)
					{
						langTag = languageTagRecords[languageId];
					}
				}
			}
			catch (CultureNotFoundException)
			{
			}

			return langTag;
		}

		/// <summary>
		/// IETF 言語タグから言語 ID を取得します。
		/// </summary>
		/// <param name="languageTag">IETF 言語タグ（RFC 3066）。</param>
		/// <param name="languageTagRacords">言語タグレコードコレクション。</param>
		/// <returns>Windows 言語 ID。対応する言語 ID が存在しない場合には 0。</returns>
		private static ushort GetLanguageId(string languageTag, List<string> languageTagRacords)
		{
			ushort languageId;

			try
			{
				languageId = (ushort)CultureInfo.CreateSpecificCulture(languageTag).LCID;
			}
			catch (CultureNotFoundException)
			{
				int index = languageTagRacords.IndexOf(languageTag);

				if (index == -1)
				{
					languageTagRacords.Add(languageTag);
					index = 0x8000 + languageTagRacords.Count - 1;
				}

				languageId = (ushort)index;
			}

			return languageId;
		}

		#endregion
	}
}