using System;
using System.IO;


namespace OfficeIVS.Font.Table
{

	/// <summary>
	/// OpenType フォントの OS/2 (OS/2 and Windows Metrics) テーブルを表します。
	/// </summary>
	[OpenTypeFontTable("OS/2")]
	internal class OS2 : OpenTypeFontTable
	{
		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="OS2"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="version">テーブルバージョン番号。</param>
		public OS2(ushort version) : base(TableTags.OS2)
		{
			Version = version;
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// 'OS/2' テーブルバージョン番号を取得または設定します。
		/// </summary>
		/// <value>テーブルバージョン番号。</value>
		/// <remarks>
		/// 現状、以下のバージョンが存在する。<br/>
		/// 0x0000 ... version 0<br/>
		/// 0x0001 ... version 1<br/>
		/// 0x0002 ... version 2<br/>
		/// 0x0003 ... version 3<br/>
		/// 0x0004 ... version 4<br/>
		/// </remarks>
		public ushort Version { get; protected set; }

		#endregion

		#region メソッド

		/// <summary>
		/// フォントから 'OS/2' テーブルを読み込みます。
		/// </summary>
		/// <param name="font">読み込み元となるフォント。</param>
		/// <returns><see cref="T:OfficeIVS.Font.Table.OS2"/> 型のインスタンス。テーブルが存在しない場合には <c>null</c>。</returns>
		[OpenTypeFontTableFactoryMethod]
		public static OpenTypeFontTable Load(OpenTypeFont font)
		{
			OS2 table = null;

			ITableRecord tableRecord = font[TableTags.OS2];

			if (tableRecord != null)
			{
				using (Stream stream = tableRecord.GetTableStream())
				using (BigEndianBinaryReader input = new BigEndianBinaryReader(stream))
				{
					ushort version = input.ReadUInt16();

					switch (version)
					{
					case 0x0000:
						table = new OS2_00();
						break;

					case 0x0001:
						table = new OS2_10();
						break;

					case 0x0002:
						table = new OS2_20();
						break;

					case 0x0003:
						table = new OS2_30();
						break;

					case 0x0004:
						table = new OS2_40();
						break;

					default:
						table = new OS2(version);
						break;
					}

					stream.Seek(-2, SeekOrigin.Current);
					table.LoadEntries(input, tableRecord.Length);

					table.TableRecord = tableRecord;
				}
			}

			return table;
		}

		/// <summary>
		/// テーブル依存のエントリデータを読み込みます。
		/// </summary>
		/// <param name="input">テーブルデータストリーム。</param>
		/// <param name="length">テーブル長（バイト単位）。</param>
		protected override void LoadEntries(BigEndianBinaryReader input, long length)
		{
			Version = input.ReadUInt16();

			LoadExtendedEntries(input, length - 2);
		}

		/// <summary>
		/// テーブル依存のエントリデータを書き込みます。
		/// </summary>
		/// <param name="output">書き込み先のストリーム。</param>
		protected override void WriteEntries(BigEndianBinaryWriter output)
		{
			output.Write(Version);

			WriteExtendedEntries(output);
		}

		#endregion
	}


	/// <summary>
	/// OpenType フォントの OS/2 (OS/2 and Windows Metrics) テーブル バージョン 0 を表します。
	/// </summary>
	internal sealed class OS2_00 : OS2
	{
		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="OS2_00"/> クラスのインスタンスを初期化します。
		/// </summary>
		public OS2_00() : base(0x0000)
		{
			CharRange = new uint[4];
			Panose = new byte[10];
			AchVendID = new byte[4];
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.OS2_00"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="os2">元となる 'OS/2' テーブル。</param>
		/// <exception cref="NotSupportedException"><paramref name="os2"/> は未サポートテーブルバージョン。</exception>
		public OS2_00(OS2 os2) : this()
		{
			OS2_00 source = os2 as OS2_00;

			if (source == null)
			{
				throw new NotSupportedException();
			}

			XAvgCharWidth = source.XAvgCharWidth;
			WeightClass = source.WeightClass;
			WidthClass = source.WidthClass;
			Type = source.Type;
			SubscriptXSize = source.SubscriptXSize;
			SubscriptYSize = source.SubscriptYSize;
			SubscriptXOffset = source.SubscriptXOffset;
			SubscriptYOffset = source.SubscriptYOffset;
			SuperscriptXSize = source.SuperscriptXSize;
			SuperscriptYSize = source.SuperscriptYSize;
			SuperscriptXOffset = source.SuperscriptXOffset;
			SuperscriptYOffset = source.SuperscriptYOffset;
			StrikeoutSize = source.StrikeoutSize;
			StrikeoutPosition = source.StrikeoutPosition;
			FamilyClass = source.FamilyClass;
			Panose[0] = source.Panose[0];
			Panose[1] = source.Panose[1];
			Panose[2] = source.Panose[2];
			Panose[3] = source.Panose[3];
			Panose[4] = source.Panose[4];
			Panose[5] = source.Panose[5];
			Panose[6] = source.Panose[6];
			Panose[7] = source.Panose[7];
			Panose[8] = source.Panose[8];
			Panose[9] = source.Panose[9];
			CharRange[0] = source.CharRange[0];
			CharRange[1] = source.CharRange[1];
			CharRange[2] = source.CharRange[2];
			CharRange[3] = source.CharRange[3];
			AchVendID[0] = source.AchVendID[0];
			AchVendID[1] = source.AchVendID[1];
			AchVendID[2] = source.AchVendID[2];
			AchVendID[3] = source.AchVendID[3];
			Selection = source.Selection;
			FirstCharIndex = source.FirstCharIndex;
			LastCharIndex = source.LastCharIndex;
			TypoAscender = source.TypoAscender;
			TypoDescender = source.TypoDescender;
			TypoLineGap = source.TypoLineGap;
			WinAscent = source.WinAscent;
			WinDescent = source.WinDescent;

		}

		#endregion

		#region プロパティ

		/// <summary>
		/// 平均文字幅を取得または設定します。
		/// </summary>
		/// <value>平均文字幅（単位： Pels/em）。</value>
		/// <remarks>フォント内の 0 幅のグリフを除いた全てののグリフの算術平均値。</remarks>
		public short XAvgCharWidth { get; set; }

		/// <summary>
		/// フォントの幅を取得または設定します。
		/// </summary>
		/// <value>フォントの幅。</value>
		/// <remarks>
		/// 【値】<br/>
		/// 100 : Thin<br/>
		/// 200 : Extra-light (Ultra-light)<br/>
		/// 300 : Light<br/>
		/// 400 : Normal (Regular)<br/>
		/// 500 : Medium<br/>
		/// 600 : Semi-bold (Demi-bold)<br/>
		/// 700 : Bold<br/>
		/// 800 : Extra-bold (Ultra-bold)<br/>
		/// 900 : Black (Heavy)<br/>
		/// </remarks>
		public ushort WeightClass { get; set; }

		/// <summary>
		/// フォントのアスペクト比を取得または設定します。
		/// </summary>
		/// <value>フォントのアスペクト比。</value>
		/// <remarks>
		/// 【値】<br/>
		/// 1 : Ultra-condensed<br/>
		/// 2 : Extra-condensed<br/>
		/// 3 : Condensed<br/>
		/// 4 : Semi-condensed<br/>
		/// 5 : Medium (normal)<br/>
		/// 6 : Semi-expanded<br/>
		/// 7 : Expanded<br/>
		/// 8 : Extra-expanded<br/>
		/// 9 : Ultra-expanded<br/>
		/// </remarks>
		public ushort WidthClass { get; set; }

		/// <summary>
		/// フォントのタイプフラグを取得または設定します。
		/// </summary>
		/// <value>フォントのタイプフラグ値。</value>
		/// <remarks>
		/// </remarks>
		public ushort Type { get; set; }

		/// <summary>
		/// 添字の X 方向の推奨サイズを取得または設定します。
		/// </summary>
		/// <value>添字の X 方向の推奨サイズ（単位： font design units）。</value>
		public short SubscriptXSize { get; set; }

		/// <summary>
		/// 添字の Y 方向の推奨サイズを取得または設定します。
		/// </summary>
		/// <value>添字の Y 方向の推奨サイズ（単位： font design units）。</value>
		public short SubscriptYSize { get; set; }

		/// <summary>
		/// 添字の X 方向の推奨オフセット量を取得または設定します。
		/// </summary>
		/// <value>添字の X 方向の推奨オフセット量（単位： font design units）。</value>
		public short SubscriptXOffset { get; set; }

		/// <summary>
		/// 添字の Y 方向の推奨オフセット量を取得または設定します。
		/// </summary>
		/// <value>添字の Y 方向の推奨オフセット量（単位： font design units）。</value>
		public short SubscriptYOffset { get; set; }

		/// <summary>
		/// 上付き文字の X 方向の推奨サイズを取得または設定します。
		/// </summary>
		/// <value>上付き文字の X 方向の推奨サイズ（単位： font design units）。</value>
		public short SuperscriptXSize { get; set; }

		/// <summary>
		/// 上付き文字の Y 方向の推奨サイズを取得または設定します。
		/// </summary>
		/// <value>上付き文字の Y 方向の推奨サイズ（単位： font design units）。</value>
		public short SuperscriptYSize { get; set; }

		/// <summary>
		/// 上付き文字の X 方向の推奨オフセット量を取得または設定します。
		/// </summary>
		/// <value>上付き文字の X 方向の推奨オフセット量（単位： font design units）。</value>
		public short SuperscriptXOffset { get; set; }

		/// <summary>
		/// 上付き文字の Y 方向の推奨オフセット量を取得または設定します。
		/// </summary>
		/// <value>上付き文字の Y 方向の推奨オフセット量（単位： font design units）。</value>
		public short SuperscriptYOffset { get; set; }

		/// <summary>
		/// 取り消し線のストローク幅を取得または設定します。
		/// </summary>
		/// <value>取り消し線のストローク幅（単位： font design units）。</value>
		public short StrikeoutSize { get; set; }

		/// <summary>
		/// ベースラインを基準として取り消し線のストローク上端の位置を取得または設定します。
		/// </summary>
		/// <value>ベースラインを基準として取り消し線のストローク上端の位置（単位： font design units）。</value>
		public short StrikeoutPosition { get; set; }

		/// <summary>
		/// フォントファミリーの分類を表す値を取得または設定します。
		/// </summary>
		/// <value>フォントファミリーの分類を表す値。</value>
		/// <remarks></remarks>
		public short FamilyClass { get; set; }

		/// <summary>
		/// 非ラテン文字セットを分類する PANOSE 値を取得または設定します。
		/// </summary>
		/// <value>PANOSE 値。</value>
		/// <remarks>10 byte 固定。</remarks>
		public byte[] Panose { get; private set; }

		/// <summary>
		/// フォント内で定義されている Unicode 文字の範囲を表すフラグ値を取得または設定します。
		/// </summary>
		/// <value>Unicode 文字の範囲を表すフラグ値。</value>
		/// <remarks>要素数 4。</remarks>
		public uint[] CharRange { get; private set; }

		/// <summary>
		/// フォントベンダー ID を取得または設定します。
		/// </summary>
		/// <value>フォントベンダー ID。</value>
		/// <remarks>ASCII 4 文字。</remarks>
		public byte[] AchVendID { get; private set; }

		/// <summary>
		/// フォントスタイル情報を取得または設定します。
		/// </summary>
		/// <value>フォントスタイル情報。</value>
		/// <remarks>
		/// 【値】<br/>
		/// Bit 0 : ITALIC
		/// Bit 1 : UNDERSCORE
		/// Bit 2 : NEGATIVE
		/// Bit 3 : OUTLINED
		/// Bit 4 : STRIKEOUT
		/// Bit 5 : BOLD
		/// Bit 6 : REGULAR
		/// Bit 7 : USE_TYPO_METRICS
		/// Bit 8 : WWS
		/// Bit 9 : OBLIQUE
		/// </remarks>
		public ushort Selection { get; set; }

		/// <summary>
		/// フォント内の最小 Unicode コードポイントを取得または設定します。
		/// </summary>
		/// <value>フォント内の最小 Unicode コードポイント（通常 U+0020）。</value>
		/// <remarks>最小の Unicode 文字が補助面の文字の場合には 0xFFFF。</remarks>
		public ushort FirstCharIndex { get; set; }

		/// <summary>
		/// フォント内の最大 Unicode コードポイントを取得または設定します。
		/// </summary>
		/// <value>フォント内の最大 Unicode コードポイント。</value>
		/// <remarks>最大の Unicode 文字が補助面の文字の場合には 0xFFFF。</remarks>
		public ushort LastCharIndex { get; set; }

		/// <summary>
		/// フォントのタイポグラフィアセンダを取得または設定します。
		/// </summary>
		/// <value>フォントのタイポグラフィアセンダ。</value>
		public short TypoAscender { get; set; }

		/// <summary>
		/// フォントのタイポグラフィディセンダーを取得または設定します。
		/// </summary>
		/// <value>フォントのタイポグラフィディセンダー。</value>
		public short TypoDescender { get; set; }

		/// <summary>
		/// フォントのタイポグラフィラインギャップを取得または設定します。
		/// </summary>
		/// <value>フォントのタイポグラフィラインギャップ。</value>
		public short TypoLineGap { get; set; }

		/// <summary>
		/// フォントの Windows 用アセンダを取得または設定します。
		/// </summary>
		/// <value>フォントのWindows 用アセンダ。</value>
		public ushort WinAscent { get; set; }

		/// <summary>
		/// フォントの Windows 用ディセントを取得または設定します。
		/// </summary>
		/// <value>フォントの Windows 用ディセント。</value>
		public ushort WinDescent { get; set; }

		#endregion

		#region メソッド

		/// <summary>
		/// テーブル依存のエントリデータを読み込みます。
		/// </summary>
		/// <param name="input">テーブルデータストリーム。</param>
		/// <param name="length">テーブル長（バイト単位）。</param>
		protected override void LoadEntries(BigEndianBinaryReader input, long length)
		{
			Version = input.ReadUInt16();
			XAvgCharWidth = input.ReadInt16();
			WeightClass = input.ReadUInt16();
			WidthClass = input.ReadUInt16();
			Type = input.ReadUInt16();
			SubscriptXSize = input.ReadInt16();
			SubscriptYSize = input.ReadInt16();
			SubscriptXOffset = input.ReadInt16();
			SubscriptYOffset = input.ReadInt16();
			SuperscriptXSize = input.ReadInt16();
			SuperscriptYSize = input.ReadInt16();
			SuperscriptXOffset = input.ReadInt16();
			SuperscriptYOffset = input.ReadInt16();
			StrikeoutSize = input.ReadInt16();
			StrikeoutPosition = input.ReadInt16();
			FamilyClass = input.ReadInt16();
			input.Read(Panose, 0, Panose.Length);
			CharRange[0] = input.ReadUInt32();
			CharRange[1] = input.ReadUInt32();
			CharRange[2] = input.ReadUInt32();
			CharRange[3] = input.ReadUInt32();
			input.Read(AchVendID, 0, AchVendID.Length);
			Selection = input.ReadUInt16();
			FirstCharIndex = input.ReadUInt16();
			LastCharIndex = input.ReadUInt16();
			TypoAscender = input.ReadInt16();
			TypoDescender = input.ReadInt16();
			TypoLineGap = input.ReadInt16();
			WinAscent = input.ReadUInt16();
			WinDescent = input.ReadUInt16();
		}

		/// <summary>
		/// テーブル依存のエントリデータを書き込みます。
		/// </summary>
		/// <param name="output">書き込み先のストリーム。</param>
		protected override void WriteEntries(BigEndianBinaryWriter output)
		{
			output.Write(Version);
			output.Write(XAvgCharWidth);
			output.Write(WeightClass);
			output.Write(WidthClass);
			output.Write(Type);
			output.Write(SubscriptXSize);
			output.Write(SubscriptYSize);
			output.Write(SubscriptXOffset);
			output.Write(SubscriptYOffset);
			output.Write(SuperscriptXSize);
			output.Write(SuperscriptYSize);
			output.Write(SuperscriptXOffset);
			output.Write(SuperscriptYOffset);
			output.Write(StrikeoutSize);
			output.Write(StrikeoutPosition);
			output.Write(FamilyClass);
			output.Write(Panose);
			output.Write(CharRange[0]);
			output.Write(CharRange[1]);
			output.Write(CharRange[2]);
			output.Write(CharRange[3]);
			output.Write(AchVendID);
			output.Write(Selection);
			output.Write(FirstCharIndex);
			output.Write(LastCharIndex);
			output.Write(TypoAscender);
			output.Write(TypoDescender);
			output.Write(TypoLineGap);
			output.Write(WinAscent);
			output.Write(WinDescent);
		}

		#endregion
	}


	/// <summary>
	/// OpenType フォントの OS/2 (OS/2 and Windows Metrics) テーブル バージョン 1 を表します。
	/// </summary>
	internal class OS2_10 : OS2
	{
		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.OS2_10"/> クラスのインスタンスを初期化します。
		/// </summary>
		public OS2_10() : this(0x0001, null)
		{
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.OS2_10"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="os2">元となる 'OS/2' テーブル。</param>
		/// <exception cref="NotSupportedException"><paramref name="os2"/> は未サポートテーブルバージョン。</exception>
		public OS2_10(OS2 os2) : this(0x0001, os2)
		{
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.OS2_10"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="version">テーブルバージョン番号。</param>
		/// <param name="os2">元となる 'OS/2' テーブル。</param>
		/// <exception cref="NotSupportedException"><paramref name="os2"/> は未サポートテーブルバージョン。</exception>
		protected OS2_10(ushort version, OS2 os2) : base(version)
		{
			Panose = new byte[10];
			AchVendID = new byte[4];

			if (os2 == null)
			{
				return;
			}

			OS2_10 source = os2 as OS2_10;

			if (source == null)
			{
				throw new NotSupportedException();
			}

			XAvgCharWidth = source.XAvgCharWidth;
			WeightClass = source.WeightClass;
			WidthClass = source.WidthClass;
			Type = source.Type;
			SubscriptXSize = source.SubscriptXSize;
			SubscriptYSize = source.SubscriptYSize;
			SubscriptXOffset = source.SubscriptXOffset;
			SubscriptYOffset = source.SubscriptYOffset;
			SuperscriptXSize = source.SuperscriptXSize;
			SuperscriptYSize = source.SuperscriptYSize;
			SuperscriptXOffset = source.SuperscriptXOffset;
			SuperscriptYOffset = source.SuperscriptYOffset;
			StrikeoutSize = source.StrikeoutSize;
			StrikeoutPosition = source.StrikeoutPosition;
			FamilyClass = source.FamilyClass;
			Panose[0] = source.Panose[0];
			Panose[1] = source.Panose[1];
			Panose[2] = source.Panose[2];
			Panose[3] = source.Panose[3];
			Panose[4] = source.Panose[4];
			Panose[5] = source.Panose[5];
			Panose[6] = source.Panose[6];
			Panose[7] = source.Panose[7];
			Panose[8] = source.Panose[8];
			Panose[9] = source.Panose[9];
			UnicodeRange1 = source.UnicodeRange1;
			UnicodeRange2 = source.UnicodeRange2;
			UnicodeRange3 = source.UnicodeRange3;
			UnicodeRange4 = source.UnicodeRange4;
			AchVendID[0] = source.AchVendID[0];
			AchVendID[1] = source.AchVendID[1];
			AchVendID[2] = source.AchVendID[2];
			AchVendID[3] = source.AchVendID[3];
			Selection = source.Selection;
			FirstCharIndex = source.FirstCharIndex;
			LastCharIndex = source.LastCharIndex;
			TypoAscender = source.TypoAscender;
			TypoDescender = source.TypoDescender;
			TypoLineGap = source.TypoLineGap;
			WinAscent = source.WinAscent;
			WinDescent = source.WinDescent;
			CodePageRange1 = source.CodePageRange1;
			CodePageRange2 = source.CodePageRange2;
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// 平均文字幅を取得または設定します。
		/// </summary>
		/// <value>平均文字幅（単位： Pels/em）。</value>
		/// <remarks>フォント内の 0 幅のグリフを除いた全てののグリフの算術平均値。</remarks>
		public short XAvgCharWidth { get; set; }

		/// <summary>
		/// フォントの幅を取得または設定します。
		/// </summary>
		/// <value>フォントの幅。</value>
		/// <remarks>
		/// 【値】<br/>
		/// 100 : Thin<br/>
		/// 200 : Extra-light (Ultra-light)<br/>
		/// 300 : Light<br/>
		/// 400 : Normal (Regular)<br/>
		/// 500 : Medium<br/>
		/// 600 : Semi-bold (Demi-bold)<br/>
		/// 700 : Bold<br/>
		/// 800 : Extra-bold (Ultra-bold)<br/>
		/// 900 : Black (Heavy)<br/>
		/// </remarks>
		public ushort WeightClass { get; set; }

		/// <summary>
		/// フォントのアスペクト比を取得または設定します。
		/// </summary>
		/// <value>フォントのアスペクト比。</value>
		/// <remarks>
		/// 【値】<br/>
		/// 1 : Ultra-condensed<br/>
		/// 2 : Extra-condensed<br/>
		/// 3 : Condensed<br/>
		/// 4 : Semi-condensed<br/>
		/// 5 : Medium (normal)<br/>
		/// 6 : Semi-expanded<br/>
		/// 7 : Expanded<br/>
		/// 8 : Extra-expanded<br/>
		/// 9 : Ultra-expanded<br/>
		/// </remarks>
		public ushort WidthClass { get; set; }

		/// <summary>
		/// フォントのタイプフラグを取得または設定します。
		/// </summary>
		/// <value>フォントのタイプフラグ値。</value>
		/// <remarks>
		/// </remarks>
		public ushort Type { get; set; }

		/// <summary>
		/// 添字の X 方向の推奨サイズを取得または設定します。
		/// </summary>
		/// <value>添字の X 方向の推奨サイズ（単位： font design units）。</value>
		public short SubscriptXSize { get; set; }

		/// <summary>
		/// 添字の Y 方向の推奨サイズを取得または設定します。
		/// </summary>
		/// <value>添字の Y 方向の推奨サイズ（単位： font design units）。</value>
		public short SubscriptYSize { get; set; }

		/// <summary>
		/// 添字の X 方向の推奨オフセット量を取得または設定します。
		/// </summary>
		/// <value>添字の X 方向の推奨オフセット量（単位： font design units）。</value>
		public short SubscriptXOffset { get; set; }

		/// <summary>
		/// 添字の Y 方向の推奨オフセット量を取得または設定します。
		/// </summary>
		/// <value>添字の Y 方向の推奨オフセット量（単位： font design units）。</value>
		public short SubscriptYOffset { get; set; }

		/// <summary>
		/// 上付き文字の X 方向の推奨サイズを取得または設定します。
		/// </summary>
		/// <value>上付き文字の X 方向の推奨サイズ（単位： font design units）。</value>
		public short SuperscriptXSize { get; set; }

		/// <summary>
		/// 上付き文字の Y 方向の推奨サイズを取得または設定します。
		/// </summary>
		/// <value>上付き文字の Y 方向の推奨サイズ（単位： font design units）。</value>
		public short SuperscriptYSize { get; set; }

		/// <summary>
		/// 上付き文字の X 方向の推奨オフセット量を取得または設定します。
		/// </summary>
		/// <value>上付き文字の X 方向の推奨オフセット量（単位： font design units）。</value>
		public short SuperscriptXOffset { get; set; }

		/// <summary>
		/// 上付き文字の Y 方向の推奨オフセット量を取得または設定します。
		/// </summary>
		/// <value>上付き文字の Y 方向の推奨オフセット量（単位： font design units）。</value>
		public short SuperscriptYOffset { get; set; }

		/// <summary>
		/// 取り消し線のストローク幅を取得または設定します。
		/// </summary>
		/// <value>取り消し線のストローク幅（単位： font design units）。</value>
		public short StrikeoutSize { get; set; }

		/// <summary>
		/// ベースラインを基準として取り消し線のストローク上端の位置を取得または設定します。
		/// </summary>
		/// <value>ベースラインを基準として取り消し線のストローク上端の位置（単位： font design units）。</value>
		public short StrikeoutPosition { get; set; }

		/// <summary>
		/// フォントファミリーの分類を表す値を取得または設定します。
		/// </summary>
		/// <value>フォントファミリーの分類を表す値。</value>
		/// <remarks></remarks>
		public short FamilyClass { get; set; }

		/// <summary>
		/// 非ラテン文字セットを分類する PANOSE 値を取得または設定します。
		/// </summary>
		/// <value>PANOSE 値。</value>
		/// <remarks>10 byte 固定。</remarks>
		public byte[] Panose { get; private set; }

		/// <summary>
		/// フォント内で定義されている Unicode 文字の範囲を表すフラグ値を取得または設定します。
		/// </summary>
		/// <value>Unicode 文字の範囲を表すフラグ値。</value>
		public uint UnicodeRange1 { get; set; }

		/// <summary>
		/// フォント内で定義されている Unicode 文字の範囲を表すフラグ値を取得または設定します。
		/// </summary>
		/// <value>Unicode 文字の範囲を表すフラグ値。</value>
		public uint UnicodeRange2 { get; set; }

		/// <summary>
		/// フォント内で定義されている Unicode 文字の範囲を表すフラグ値を取得または設定します。
		/// </summary>
		/// <value>Unicode 文字の範囲を表すフラグ値。</value>
		public uint UnicodeRange3 { get; set; }

		/// <summary>
		/// フォント内で定義されている Unicode 文字の範囲を表すフラグ値を取得または設定します。
		/// </summary>
		/// <value>Unicode 文字の範囲を表すフラグ値。</value>
		public uint UnicodeRange4 { get; set; }

		/// <summary>
		/// フォントベンダー ID を取得または設定します。
		/// </summary>
		/// <value>フォントベンダー ID。</value>
		/// <remarks>4 byte 固定</remarks>
		public byte[] AchVendID { get; set; }

		/// <summary>
		/// フォントスタイル情報を取得または設定します。
		/// </summary>
		/// <value>フォントスタイル情報。</value>
		/// <remarks>
		/// 【値】<br/>
		/// Bit 0 : ITALIC
		/// Bit 1 : UNDERSCORE
		/// Bit 2 : NEGATIVE
		/// Bit 3 : OUTLINED
		/// Bit 4 : STRIKEOUT
		/// Bit 5 : BOLD
		/// Bit 6 : REGULAR
		/// Bit 7 : USE_TYPO_METRICS
		/// Bit 8 : WWS
		/// Bit 9 : OBLIQUE
		/// </remarks>
		public ushort Selection { get; set; }

		/// <summary>
		/// フォント内の最小 Unicode コードポイントを取得または設定します。
		/// </summary>
		/// <value>フォント内の最小 Unicode コードポイント（通常 U+0020）。</value>
		/// <remarks>最小の Unicode 文字が補助面の文字の場合には 0xFFFF。</remarks>
		public ushort FirstCharIndex { get; set; }

		/// <summary>
		/// フォント内の最大 Unicode コードポイントを取得または設定します。
		/// </summary>
		/// <value>フォント内の最大 Unicode コードポイント。</value>
		/// <remarks>最大の Unicode 文字が補助面の文字の場合には 0xFFFF。</remarks>
		public ushort LastCharIndex { get; set; }

		/// <summary>
		/// フォントのタイポグラフィアセンダを取得または設定します。
		/// </summary>
		/// <value>フォントのタイポグラフィアセンダ。</value>
		public short TypoAscender { get; set; }

		/// <summary>
		/// フォントのタイポグラフィディセンダーを取得または設定します。
		/// </summary>
		/// <value>フォントのタイポグラフィディセンダー。</value>
		public short TypoDescender { get; set; }

		/// <summary>
		/// フォントのタイポグラフィラインギャップを取得または設定します。
		/// </summary>
		/// <value>フォントのタイポグラフィラインギャップ。</value>
		public short TypoLineGap { get; set; }

		/// <summary>
		/// フォントの Windows 用アセンダを取得または設定します。
		/// </summary>
		/// <value>フォントのWindows 用アセンダ。</value>
		public ushort WinAscent { get; set; }

		/// <summary>
		/// フォントの Windows 用ディセントを取得または設定します。
		/// </summary>
		/// <value>フォントの Windows 用ディセント。</value>
		public ushort WinDescent { get; set; }

		/// <summary>
		/// 'cmap' テーブル が包含するコードページを表すフラグ値を取得または設定します。
		/// </summary>
		/// <value>'cmap' テーブル が包含するコードページを表すフラグ値。</value>
		/// <remarks>互換性： v1 以降</remarks>
		public uint CodePageRange1 { get; set; }

		/// <summary>
		/// 'cmap' テーブル が包含するコードページを表すフラグ値を取得または設定します。
		/// </summary>
		/// <value>'cmap' テーブル が包含するコードページを表すフラグ値。</value>
		/// <remarks>互換性： v1 以降</remarks>
		public uint CodePageRange2 { get; set; }

		#endregion

		#region メソッド

		/// <summary>
		/// テーブル依存のエントリデータを読み込みます。
		/// </summary>
		/// <param name="input">テーブルデータストリーム。</param>
		/// <param name="length">テーブル長（バイト単位）。</param>
		protected override void LoadEntries(BigEndianBinaryReader input, long length)
		{
			Version = input.ReadUInt16();
			XAvgCharWidth = input.ReadInt16();
			WeightClass = input.ReadUInt16();
			WidthClass = input.ReadUInt16();
			Type = input.ReadUInt16();
			SubscriptXSize = input.ReadInt16();
			SubscriptYSize = input.ReadInt16();
			SubscriptXOffset = input.ReadInt16();
			SubscriptYOffset = input.ReadInt16();
			SuperscriptXSize = input.ReadInt16();
			SuperscriptYSize = input.ReadInt16();
			SuperscriptXOffset = input.ReadInt16();
			SuperscriptYOffset = input.ReadInt16();
			StrikeoutSize = input.ReadInt16();
			StrikeoutPosition = input.ReadInt16();
			FamilyClass = input.ReadInt16();
			input.Read(Panose, 0, Panose.Length);
			UnicodeRange1 = input.ReadUInt32();
			UnicodeRange2 = input.ReadUInt32();
			UnicodeRange3 = input.ReadUInt32();
			UnicodeRange4 = input.ReadUInt32();
			input.Read(AchVendID, 0, AchVendID.Length);
			Selection = input.ReadUInt16();
			FirstCharIndex = input.ReadUInt16();
			LastCharIndex = input.ReadUInt16();
			TypoAscender = input.ReadInt16();
			TypoDescender = input.ReadInt16();
			TypoLineGap = input.ReadInt16();
			WinAscent = input.ReadUInt16();
			WinDescent = input.ReadUInt16();
			CodePageRange1 = input.ReadUInt32();
			CodePageRange2 = input.ReadUInt32();
		}

		/// <summary>
		/// テーブル依存のエントリデータを書き込みます。
		/// </summary>
		/// <param name="output">書き込み先のストリーム。</param>
		protected override void WriteEntries(BigEndianBinaryWriter output)
		{
			output.Write(Version);
			output.Write(XAvgCharWidth);
			output.Write(WeightClass);
			output.Write(WidthClass);
			output.Write(Type);
			output.Write(SubscriptXSize);
			output.Write(SubscriptYSize);
			output.Write(SubscriptXOffset);
			output.Write(SubscriptYOffset);
			output.Write(SuperscriptXSize);
			output.Write(SuperscriptYSize);
			output.Write(SuperscriptXOffset);
			output.Write(SuperscriptYOffset);
			output.Write(StrikeoutSize);
			output.Write(StrikeoutPosition);
			output.Write(FamilyClass);
			output.Write(Panose);
			output.Write(UnicodeRange1);
			output.Write(UnicodeRange2);
			output.Write(UnicodeRange3);
			output.Write(UnicodeRange4);
			output.Write(AchVendID);
			output.Write(Selection);
			output.Write(FirstCharIndex);
			output.Write(LastCharIndex);
			output.Write(TypoAscender);
			output.Write(TypoDescender);
			output.Write(TypoLineGap);
			output.Write(WinAscent);
			output.Write(WinDescent);
			output.Write(CodePageRange1);
			output.Write(CodePageRange2);
		}

		#endregion
	}


	/// <summary>
	/// OpenType フォントの OS/2 (OS/2 and Windows Metrics) テーブル バージョン 2 を表します。
	/// </summary>
	internal class OS2_20 : OS2_10
	{
		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.OS2_20"/> クラスのインスタンスを初期化します。
		/// </summary>
		public OS2_20() : this(0x0002, null)
		{
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.OS2_20"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="os2">元となる 'OS/2' テーブル。</param>
		/// <exception cref="NotSupportedException"><paramref name="os2"/> は未サポートテーブルバージョン。</exception>
		public OS2_20(OS2 os2) : this(0x0002, os2)
		{
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.OS2_10"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="os2">元となる 'OS/2' テーブル。</param>
		/// <param name="version">テーブルバージョン番号。</param>
		/// <exception cref="NotSupportedException"><paramref name="os2"/> は未サポートテーブルバージョン。</exception>
		protected OS2_20(ushort version, OS2 os2) : base(version, os2)
		{
			if (os2 == null)
			{
				return;
			}

			if (os2.Version == 0)
			{
			}
			else if (os2.Version >= 2)
			{
				OS2_20 source = os2 as OS2_20;

				if (source == null)
				{
					throw new NotSupportedException();
				}

				XHeight = source.XHeight;
				CapHeight = source.CapHeight;
				DefaultChar = source.DefaultChar;
				BreakChar = source.BreakChar;
				MaxContext = source.MaxContext;
			}
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// 小文字英字のベースライン基準のおおよその高さを取得または設定します。
		/// </summary>
		/// <value>ベースライン基準の高さ（単位 FUnit）。</value>
		public short XHeight { get; set; }

		/// <summary>
		/// 大文字英字のベースライン基準のおおよその高さを取得または設定します。
		/// </summary>
		/// <value>ベースライン基準の高さ（単位 FUnit）。</value>
		public short CapHeight { get; set; }

		/// <summary>
		/// Windows で文字欠落時に使用する文字を取得または設定します。
		/// </summary>
		/// <value>文字欠落時に使用する文字の文字コード。ただし 0 の場合には glyphId 0 を表示。</value>
		public ushort DefaultChar { get; set; }

		/// <summary>
		/// Windows で区切り文字として扱う文字を取得または設定します。
		/// </summary>
		/// <value>区切り文字の文字コード（通常は半角スペース。）。</value>
		public ushort BreakChar { get; set; }

		/// <summary>
		/// 機能対象となるグリフの最大長を取得または設定します。
		/// </summary>
		/// <value>グリフの最大長。</value>
		public ushort MaxContext { get; set; }

		#endregion

		#region メソッド

		/// <summary>
		/// テーブル依存のエントリデータを読み込みます。
		/// </summary>
		/// <param name="input">テーブルデータストリーム。</param>
		/// <param name="length">テーブル長（バイト単位）。</param>
		protected override void LoadEntries(BigEndianBinaryReader input, long length)
		{
			base.LoadEntries(input, length);

			XHeight = input.ReadInt16();
			CapHeight = input.ReadInt16();
			DefaultChar = input.ReadUInt16();
			BreakChar = input.ReadUInt16();
			MaxContext = input.ReadUInt16();
		}

		/// <summary>
		/// テーブル依存のエントリデータを書き込みます。
		/// </summary>
		/// <param name="output">書き込み先のストリーム。</param>
		protected override void WriteEntries(BigEndianBinaryWriter output)
		{
			base.WriteEntries(output);

			output.Write(XHeight);
			output.Write(CapHeight);
			output.Write(DefaultChar);
			output.Write(BreakChar);
			output.Write(MaxContext);
		}

		#endregion
	}


	/// <summary>
	/// OpenType フォントの OS/2 (OS/2 and Windows Metrics) テーブル バージョン 3 を表します。
	/// </summary>
	internal class OS2_30 : OS2_20
	{
		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.OS2_30"/> クラスのインスタンスを初期化します。
		/// </summary>
		public OS2_30() : base(0x0003, null)
		{
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.OS2_30"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="os2">元となる 'OS/2' テーブル。</param>
		/// <exception cref="NotSupportedException"><paramref name="os2"/> は未サポートテーブルバージョン。</exception>
		public OS2_30(OS2 os2) : base(0x0003, os2)
		{
		}

		#endregion
	}


	/// <summary>
	/// OpenType フォントの OS/2 (OS/2 and Windows Metrics) テーブル バージョン 4 を表します。
	/// </summary>
	internal class OS2_40 : OS2_20
	{
		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.OS2_40"/> クラスのインスタンスを初期化します。
		/// </summary>
		public OS2_40() : base(0x0004, null)
		{
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.OS2_40"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="os2">元となる 'OS/2' テーブル。</param>
		/// <exception cref="NotSupportedException"><paramref name="os2"/> は未サポートテーブルバージョン。</exception>
		public OS2_40(OS2 os2) : base(0x0004, os2)
		{
		}

		#endregion
	}
}