using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Reflection;


namespace OfficeIVS.Font.Table
{


	/// <summary>
	/// OpenType フォントのテーブルを表します。
	/// </summary>
	[DebuggerDisplay("{Tag}")]
	internal class OpenTypeFontTable
	{
		#region フィールド

		/// <summary>
		/// テーブル生成メソッドマッピング。
		/// </summary>
		/// <remarks>キーは、テーブル識別子。</remarks>
		private static readonly Dictionary<uint, MethodInfo> TableFactoryMethods;

		/// <summary>
		/// テーブル識別子。
		/// </summary>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private readonly Tag _tag;

		/// <summary>
		/// テーブルレコード。
		/// </summary>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private ITableRecord _tableRecord;

		#endregion

		#region コンストラクタ

		/// <summary>
		/// 静的コンストラクタ。
		/// </summary>
		static OpenTypeFontTable()
		{
			TableFactoryMethods = new Dictionary<uint, MethodInfo>();
			
			Assembly assembly = Assembly.GetAssembly(typeof(OpenTypeFontTable));

			foreach (Type type in assembly.GetTypes())
			{
				if (!type.IsClass || !type.IsSubclassOf(typeof(OpenTypeFontTable)))
				{
					continue;
				}

				OpenTypeFontTableAttribute otfAttr =
					OpenTypeFontTableAttribute.GetAttribute(type);

				if (otfAttr == null)
				{
					continue;
				}

				MethodInfo factoryMethod = 
					OpenTypeFontTableFactoryMethodAttribute.FindFactoryMethod(type);

				if (factoryMethod != null)
				{
					TableFactoryMethods[otfAttr.Tag] = factoryMethod;
				}
			}
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.OpenTypeFontTable"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="tag">テーブル識別子。</param>
		protected OpenTypeFontTable(Tag tag)
		{
			_tag = tag;
			_tableRecord = null;
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.OpenTypeFontTable"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="tableRecord">テーブルレコード。</param>
		protected OpenTypeFontTable(ITableRecord tableRecord)
		{
			if (tableRecord == null)
			{
				throw new ArgumentNullException("tableRecord");
			}
			_tag = tableRecord.Tag;
			_tableRecord = tableRecord;
		}

		#endregion

		#region パブリック プロパティ

		/// <summary>
		/// テーブル識別子を取得します。
		/// </summary>
		/// <value>テーブル識別子。</value>
		public Tag Tag
		{
			[DebuggerStepThrough]
			get { return _tag; }
		}

		/// <summary>
		/// 拡張エントリーデータを取得または設定します。
		/// </summary>
		/// <value>拡張エントリーデータ。</value>
		/// <remarks>未対応バージョンの形式不明なデータなど。</remarks>
		public byte[] ExtendedEntries { get; set; }

		#endregion

		#region プロテクト プロパティ

		/// <summary>
		/// 現在設定されているテーブルレコードを取得または設定します。
		/// </summary>
		/// <value>テーブルレコード。</value>
		protected ITableRecord TableRecord
		{
			[DebuggerStepThrough]
			get { return _tableRecord; }

			[DebuggerStepThrough]
			set { _tableRecord = value; }
		}

		#endregion

		#region パブリック メソッド

		/// <summary>
		/// 本テーブルのテーブルレコードを取得します。
		/// </summary>
		/// <returns>テーブルレコード。</returns>
		public ITableRecord GetTableRecord()
		{
			if (_tableRecord == null)
			{
				EditableTableRecord tableRecord = new EditableTableRecord(Tag);

				using (MemoryStream ms = new MemoryStream())
				{
					WriteTo(ms);
					ms.Flush();

					tableRecord.SetTableData(ms.ToArray());
				}

				_tableRecord = tableRecord;
			}

			return _tableRecord;
		}

		/// <summary>
		/// OpenType フォントから指定されたテーブルを読み込みます。
		/// </summary>
		/// <param name="font">テーブルレコードを所有するフォント。</param>
		/// <param name="tableTag">読み込むテーブルのテーブル識別子。。</param>
		/// <returns>OpenType フォントテーブル。指定されたテーブルが含まれていない場合には <c>null</c>。</returns>
		/// <exception cref="ArgumentNullException"><paramref name="font"/> が <c>null</c>。</exception>
		public static OpenTypeFontTable Load(OpenTypeFont font, Tag tableTag)
		{
			if (font == null)
			{
				throw new ArgumentNullException("font");
			}

			// テーブルオブジェクトの生成
			OpenTypeFontTable otfTable = null;

			MethodInfo factoryMethod;
			TableFactoryMethods.TryGetValue(tableTag.Value, out factoryMethod);

			if (factoryMethod != null)
			{
				otfTable = factoryMethod.Invoke(null, new object[]{font}) as OpenTypeFontTable;
			}
			else
			{
				ITableRecord tableRecord = font[tableTag];

				if (tableRecord != null)
				{
					using (Stream tableStream = tableRecord.GetTableStream())
					using (BigEndianBinaryReader reader = new BigEndianBinaryReader(tableStream))
					{
						otfTable = new OpenTypeFontTable(tableRecord);
						otfTable.LoadEntries(reader, tableRecord.Length);
					}
				}
			}

			return otfTable;
		}

		/// <summary>
		/// 指定ストリームにテーブルデータを書き込みます。
		/// </summary>
		/// <param name="output">書き込み先のストリーム。</param>
		/// <returns>書き込んだバイト数。</returns>
		/// <remarks>本メソッドはテーブル長を 4 バイト境界に合わせません。</remarks>
		/// <exception cref="ArgumentException"><paramref name="output"/> が読み取りをサポートしていない、もしくは既に閉じられている。</exception>
		/// <exception cref="ArgumentNullException"><paramref name="output"/> が <c>null</c>。</exception>
		public uint WriteTo(Stream output)
		{
			long startPos = output.Position;

			using (BigEndianBinaryWriter writer = new BigEndianBinaryWriter(output))
			{
				// テーブルデータの書き込み
				WriteEntries(writer);
			}

			return Convert.ToUInt32(output.Position - startPos);
		}

		#endregion

		#region プロテクト メソッド

		/// <summary>
		/// テーブル依存のエントリデータを読み込みます。
		/// </summary>
		/// <param name="input">テーブルデータストリーム。</param>
		/// <param name="length">テーブル長（バイト単位）。</param>
		protected virtual void LoadEntries(BigEndianBinaryReader input, long length)
		{
			LoadExtendedEntries(input, length);
		}

		/// <summary>
		/// テーブルのバージョン依存のエントリデータを読み込みます。
		/// </summary>
		/// <param name="input">テーブルデータストリーム。</param>
		/// <param name="length">残りのテーブル長（バイト単位）。</param>
		protected void LoadExtendedEntries(BigEndianBinaryReader input, long length)
		{
			if (length > 0)
			{
				ExtendedEntries = new byte[length];
				input.Read(ExtendedEntries, 0, ExtendedEntries.Length);
			}
		}

		/// <summary>
		/// テーブル依存のエントリデータを書き込みます。
		/// </summary>
		/// <param name="output">書き込み先のストリーム。</param>
		protected virtual void WriteEntries(BigEndianBinaryWriter output)
		{
			WriteExtendedEntries(output);
		}

		/// <summary>
		/// テーブル依存のエントリデータを書き込みます。
		/// </summary>
		/// <param name="output">書き込み先のストリーム。</param>
		protected void WriteExtendedEntries(BigEndianBinaryWriter output)
		{
			if (ExtendedEntries != null&& ExtendedEntries.Length > 0)
			{
				output.Write(ExtendedEntries, 0, ExtendedEntries.Length);
			}
		}

		#endregion
	}
}