﻿using System;
using System.Diagnostics;
using System.Reflection;


namespace OfficeIVS.Font.Table
{

	/// <summary>
	/// OpenType フォントのテーブルを表すクラスであることを示す属性です。
	/// </summary>
	[AttributeUsage(AttributeTargets.Class)]
	internal sealed class OpenTypeFontTableAttribute : Attribute
	{
		#region フィールド

		/// <summary>
		/// テーブル識別子。
		/// </summary>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private readonly uint _tag;

		#endregion

		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.OpenTypeFontTableAttribute"/> のインスタンスを初期化します。
		/// </summary>
		/// <param name="tag">テーブル識別子。</param>
		public OpenTypeFontTableAttribute(string tag)
		{
			_tag = Font.Tag.FromName(tag).Value;
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// テーブル識別子を取得します。
		/// </summary>
		/// <value>テーブル識別子。</value>
		public uint Tag
		{
			[DebuggerStepThrough]
			get { return _tag; }
		}

		#endregion

		#region メソッド

		/// <summary>
		/// 指定型に設定されている <see cref="T:OfficeIVS.Font.Table.OpenTypeFontTableAttribute"/> 属性を取得します。
		/// </summary>
		/// <param name="type">検証対象の型。</param>
		/// <returns><see cref="T:OfficeIVS.Font.Table.OpenTypeFontTableAttribute"/> 属性。存在しない場合には <c>null</c>。</returns>
		public static OpenTypeFontTableAttribute GetAttribute(Type type)
		{
			OpenTypeFontTableAttribute result = null;

			object[] attributes = type.GetCustomAttributes(typeof(OpenTypeFontTableAttribute), false);

			if (attributes.Length > 0)
			{
				result = (OpenTypeFontTableAttribute)attributes[0];
			}

			return result;
		}

		#endregion
	}


	/// <summary>
	/// メソッドが OpenType フォントテーブルの生成メソッドであることを示す属性です。
	/// </summary>
	/// <remarks>
	/// 生成メソッドは、静的かつ以下の条件に一致する必要があります。
	///  引数型： OpenTypeFont
	///  戻値型： OpenTypeFontTable
	/// </remarks>
	[AttributeUsage(AttributeTargets.Method)]
	internal sealed class OpenTypeFontTableFactoryMethodAttribute : Attribute
	{
		#region メソッド

		/// <summary>
		/// 指定型の中から生成メソッドを検索し、最初に見つかった生成メソッドのメタ情報を返します。
		/// </summary>
		/// <param name="type">検索対象の型。</param>
		/// <returns>生成メソッドのメタデータ。存在しない場合には <c>null</c>。</returns>
		public static MethodInfo FindFactoryMethod(Type type)
		{
			MethodInfo factoryMethod = null;

			if (type != null && type.IsClass)
			{
				foreach (MethodInfo methodInfo in type.GetMethods(BindingFlags.Static | BindingFlags.Public | BindingFlags.NonPublic))
				{
					// 属性有無のチェック
					if (!methodInfo.IsDefined(typeof(OpenTypeFontTableFactoryMethodAttribute), false))
					{
						continue;
					}

					// メソッド引数の型チェック
					ParameterInfo[] paramInfos = methodInfo.GetParameters();

					if (paramInfos.Length != 1 
					 || paramInfos[0].ParameterType != typeof(OpenTypeFont))
					{
						continue;
					}

					// メソッド戻値の型チェック
					if (methodInfo.ReturnType != typeof(OpenTypeFontTable))
					{
						continue;	
					}

					factoryMethod = methodInfo;
					break;
				}
			}

			return factoryMethod;
		}

		#endregion
	}
}