using System.IO;


namespace OfficeIVS.Font.Table
{

	/// <summary>
	/// OpenType フォントの prep (Control Value Program : Pre Program) テーブルを表します。
	/// </summary>
	[OpenTypeFontTable("prep")]
	internal sealed class PreP : OpenTypeFontTable
	{
		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="PreP"/> クラスのインスタンスを初期化します。
		/// </summary>
		public PreP() : base(TableTags.prep)
		{
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// フォント、フォントサイズ変更時の命令セットを取得または設定します。
		/// </summary>
		/// <value>命令セット。</value>
		/// <remarks>配列サイズはテーブル長と一致。</remarks>
		public byte[] Instructions { get; set; }

		#endregion

		#region メソッド

		/// <summary>
		/// フォントから 'prep' テーブルを読み込みます。
		/// </summary>
		/// <param name="font">読み込み元となるフォント。</param>
		/// <returns><see cref="T:OfficeIVS.Font.Table.PreP"/> 型のインスタンス。テーブルが存在しない場合には <c>null</c>。</returns>
		[OpenTypeFontTableFactoryMethod]
		public static OpenTypeFontTable Load(OpenTypeFont font)
		{
			PreP table = null;

			ITableRecord tableRecord = font[TableTags.prep];

			if (tableRecord != null)
			{
				using (Stream stream = tableRecord.GetTableStream())
				using (BigEndianBinaryReader reader = new BigEndianBinaryReader(stream))
				{
					table = new PreP();
					table.LoadEntries(reader, tableRecord.Length);
					table.TableRecord = tableRecord;
				}
			}

			return table;
		}

		/// <summary>
		/// テーブル依存のエントリデータを読み込みます。
		/// </summary>
		/// <param name="input">テーブルデータストリーム。</param>
		/// <param name="length">テーブル長（バイト単位）。</param>
		protected override void LoadEntries(BigEndianBinaryReader input, long length)
		{
			if (length > 0)
			{
				Instructions = new byte[length];
				input.Read(Instructions, 0, Instructions.Length);
			}
		}

		/// <summary>
		/// テーブル依存のエントリデータを書き込みます。
		/// </summary>
		/// <param name="output">書き込み先のストリーム。</param>
		protected override void WriteEntries(BigEndianBinaryWriter output)
		{
			if (Instructions != null && Instructions.Length > 0)
			{
				output.Write(Instructions);
			}
		}

		#endregion
	}
}