using System;
using System.IO;


namespace OfficeIVS.Font.Table
{

	/// <summary>
	/// OpenType フォントの vhea (Vertical Metrics Header) テーブルを表します。
	/// </summary>
	[OpenTypeFontTable("vhea")]
	internal class VHea : OpenTypeFontTable
	{
		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.VHea"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="version">テーブルバージョン番号。</param>
		public VHea(uint version) : base(TableTags.vhea)
		{
			Version = version;
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// テーブルバージョン番号を取得または設定します。
		/// </summary>
		/// <value>テーブルバージョン番号。</value>
		/// <remarks>1.0(=0x00010000) 固定。</remarks>
		public uint Version { get; protected set; }

		#endregion

		#region メソッド

		/// <summary>
		/// フォントから 'vhea' テーブルを読み込みます。
		/// </summary>
		/// <param name="font">読み込み元となるフォント。</param>
		/// <returns><see cref="T:OfficeIVS.Font.Table.VHea"/> 型のインスタンス。テーブルが存在しない場合には <c>null</c>。</returns>
		[OpenTypeFontTableFactoryMethod]
		public static OpenTypeFontTable Load(OpenTypeFont font)
		{
			VHea table = null;

			ITableRecord tableRecord = font[TableTags.vhea];

			if (tableRecord != null)
			{
				using (Stream stream = tableRecord.GetTableStream())
				using (BigEndianBinaryReader reader = new BigEndianBinaryReader(stream))
				{
					uint version = reader.ReadUInt32();

					switch (version)
					{
					case 0x00010000:
						table = new VHea10();
						break;

					case 0x00011000:
						table = new VHea11();
						break;

					default:
						table = new VHea(version);
						break;
					}

					stream.Seek(-4, SeekOrigin.Current);
					table.LoadEntries(reader, tableRecord.Length);

					table.TableRecord = tableRecord;
				}
			}

			return table;
		}

		/// <summary>
		/// テーブル依存のエントリデータを読み込みます。
		/// </summary>
		/// <param name="input">テーブルデータストリーム。</param>
		/// <param name="length">テーブル長（バイト単位）。</param>
		protected override void LoadEntries(BigEndianBinaryReader input, long length)
		{
			Version = input.ReadUInt32();

			LoadExtendedEntries(input, length - 4);
		}

		/// <summary>
		/// テーブル依存のエントリデータを書き込みます。
		/// </summary>
		/// <param name="output">書き込み先のストリーム。</param>
		protected override void WriteEntries(BigEndianBinaryWriter output)
		{
			output.Write(Version);

			WriteExtendedEntries(output);
		}

		#endregion
	}


	/// <summary>
	/// OpenType フォントの vhea (Horizontal Header) テーブル バージョン 1.0 を表します。
	/// </summary>
	internal class VHea10 : VHea
	{
		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.VHea10"/> クラスのインスタンスを初期化します。
		/// </summary>
		public VHea10() : this(0x00010000, null)
		{
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.VHea10"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="vhea">元となる 'hhea' テーブル。</param>
		/// <exception cref="NotSupportedException"><paramref name="vhea"/> は未サポートテーブルバージョン。</exception>
		public VHea10(VHea vhea) : this(0x00010000, vhea)
		{
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.VHea10"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="version">テーブルバージョン番号。</param>
		/// <param name="vhea">元となる 'vhea' テーブル。</param>
		/// <exception cref="NotSupportedException"><paramref name="vhea"/> は未サポートテーブルバージョン。</exception>
		protected VHea10(uint version, VHea vhea) : base(version)
		{
			LineGap = 0;
			MetricDataFormat = 0;

			if (vhea == null)
			{
				return;
			}

			VHea10 source = vhea as VHea10;

			if (source == null)
			{
				throw new NotSupportedException();
			}

			Ascent = source.Ascent;
			Descent = source.Descent;
			LineGap = source.LineGap;
			AdvanceHeightMax = source.AdvanceHeightMax;
			MinTopSideBearing = source.MinTopSideBearing;
			MinBottomSideBearing = source.MinBottomSideBearing;
			YMaxExtent = source.YMaxExtent;
			CaretSlopeRise = source.CaretSlopeRise;
			CaretSlopeRun = source.CaretSlopeRun;
			CaretOffset = source.CaretOffset;
			MetricDataFormat = source.MetricDataFormat;
			NumOfLongVerMetrics = source.NumOfLongVerMetrics;
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// 中心線から前行の descent までの距離を取得または設定します。
		/// </summary>
		/// <value>中心線から前行の descent までの距離（FUNIT 単位）。</value>
		public short Ascent { get; set; }

		/// <summary>
		/// 中心線から次行の ascent までの距離を取得または設定します。
		/// </summary>
		/// <value>中心線から次行の ascent までの距離（FUNIT 単位）。</value>
		public short Descent { get; set; }

		/// <summary>
		/// ラインギャップを取得または設定します。
		/// </summary>
		/// <value>ラインギャップ（0 固定）。</value>
		public short LineGap { get; set; }

		/// <summary>
		/// フォント内のグリフの最大アドバンス高さを取得または設定します。
		/// </summary>
		/// <value>グリフの最大アドバンス高さ（FUNIT 単位）。</value>
		/// <remarks>'vmtx' テーブルの内の値と同じでなければならない。</remarks>
		public ushort AdvanceHeightMax { get; set; }

		/// <summary>
		/// フォント内のグリフの最小 top sidebearing 値を取得または設定します。
		/// </summary>
		/// <value>グリフの最小 top sidebearing 値（FUNIT 単位）。</value>
		/// <remarks>'vmtx' テーブルの内の値と同じでなければならない。</remarks>
		public short MinTopSideBearing { get; set; }

		/// <summary>
		/// フォント内のグリフの最小 bottom sidebearing 値を取得または設定します。
		/// </summary>
		/// <value>グリフの最小 top sidebearing 値（FUNIT 単位）。</value>
		/// <remarks>'vmtx' テーブルの内の値と同じでなければならない。</remarks>
		public short MinBottomSideBearing { get; set; }

		/// <summary>
		/// yMaxExtent 値を取得または設定します。
		/// </summary>
		/// <value>yMaxExtent 値。</value>
		/// <remarks>yMaxExtent=minTopSideBearing+(yMax-yMin)</remarks>
		public short YMaxExtent { get; set; }

		/// <summary>
		/// キャレットの傾きパラメータを取得または設定します。
		/// </summary>
		/// <value>キャレットの傾きパラメータ。</value>
		/// <remarks>キャレットの傾き=(rise/run)。rise=1、run=0 の場合、水平とする。</remarks>
		public short CaretSlopeRise { get; set; }

		/// <summary>
		/// キャレットの傾きパラメータを取得または設定します。
		/// </summary>
		/// <value>キャレットの傾きパラメータ。</value>
		/// <remarks>キャレットの傾き=(rise/run)。rise=1、run=0 の場合、水平とする。</remarks>
		public short CaretSlopeRun { get; set; }

		/// <summary>
		/// キャレットのオフセット量を取得または設定します。
		/// </summary>
		/// <value>キャレットのオフセット量。</value>
		/// <remarks>非傾斜の場合には 0。</remarks>
		public short CaretOffset { get; set; }

		/// <summary>
		/// グリフ垂直メトリックデータ形式を取得または設定します。
		/// </summary>
		/// <value>グリフ垂直メトリックデータ形式（現状 0 のみ）。</value>
		public short MetricDataFormat { get; set; }

		/// <summary>
		/// 'vmtx' テーブル内のグリフ垂直メトリック要素数を取得または設定します。
		/// </summary>
		/// <value>'hmtx' テーブル内のグリフ垂直メトリック要素数。</value>
		public ushort NumOfLongVerMetrics { get; set; }

		#endregion

		#region メソッド

		/// <summary>
		/// バージョン依存の拡張エントリデータを読み込みます。
		/// </summary>
		/// <param name="reader">読み込み元のストリーム（位置合わせ済み）。</param>
		/// <param name="length">拡張エントリデータサイズ。</param>
		protected override void LoadEntries(BigEndianBinaryReader reader, long length)
		{
			Version = reader.ReadUInt32();
			Ascent = reader.ReadInt16();
			Descent = reader.ReadInt16();
			LineGap = reader.ReadInt16();
			AdvanceHeightMax = reader.ReadUInt16();
			MinTopSideBearing = reader.ReadInt16();
			MinBottomSideBearing = reader.ReadInt16();
			YMaxExtent = reader.ReadInt16();
			CaretSlopeRise = reader.ReadInt16();
			CaretSlopeRun = reader.ReadInt16();
			CaretOffset = reader.ReadInt16();
			reader.Position += 8;	// 予約領域(short * 4)
			MetricDataFormat = reader.ReadInt16();
			NumOfLongVerMetrics = reader.ReadUInt16();
		}

		/// <summary>
		/// バージョン依存の拡張エントリデータを指定ストリームに書き込みます。
		/// </summary>
		/// <param name="writer">書き込み先のストリーム（位置合わせ済み）。</param>
		protected override void WriteEntries(BigEndianBinaryWriter writer)
		{
			writer.Write(Version);
			writer.Write(Ascent);
			writer.Write(Descent);
			writer.Write(LineGap);
			writer.Write(AdvanceHeightMax);
			writer.Write(MinTopSideBearing);
			writer.Write(MinBottomSideBearing);
			writer.Write(YMaxExtent);
			writer.Write(CaretSlopeRise);
			writer.Write(CaretSlopeRun);
			writer.Write(CaretOffset);
			writer.Write((long)0);		// 予約領域(short * 4)
			writer.Write(MetricDataFormat);
			writer.Write(NumOfLongVerMetrics);
		}

		/// <summary>
		/// 'glyf'、'vmtx' テーブルから本テーブルのエントリ値を算出し設定します。
		/// </summary>
		/// <param name="maxGlyph">グリフ数。</param>
		/// <param name="glyf">'glyf' テーブル。</param>
		/// <param name="vmtx">'vmtx' テーブル。</param>
		/// <remarks>
		/// 本テーブルの以下のエントリ値を算出し設定する。
		///  AdvanceHeightMax
		///  MinTopSideBearing
		///  MinBottomSideBearing
		///  YMaxExtent
		///  NumOfLongVerMetrics
		/// </remarks>
		public void SetVMetrics(ushort maxGlyph, Glyf glyf, VMtx vmtx)
		{
			// 'glyf'、'hmtx' テーブルからエントリの値を算出
			ushort maxAH = 0;
			short minTSB = Int16.MaxValue;
			short minBSB = Int16.MaxValue;
			short maxExt = Int16.MinValue;

			for (ushort glyphIndex=0; glyphIndex<maxGlyph; glyphIndex++)
			{
				GlyphDescription gd = glyf.GetGlyphDescription(glyphIndex);

				if (gd == null)
				{
					continue;
				}

				LongVMetric vm = vmtx.GetVMetric(glyphIndex);

				maxAH = Math.Max(maxAH, vm.AdvanceHeight);
				minTSB = Math.Min(minTSB, vm.TopSideBearing);
				minBSB = Math.Min(minBSB, (short)(vm.AdvanceHeight - vm.TopSideBearing - (gd.YMax - gd.YMin)));
				maxExt = Math.Max(maxExt, (short)(vm.TopSideBearing + (gd.YMax - gd.YMin)));
			}
				
			// 'hhea' テーブルに計算値を設定
			AdvanceHeightMax = maxAH;
			MinTopSideBearing = minTSB;
			MinBottomSideBearing = minBSB;
			YMaxExtent = maxExt;
			NumOfLongVerMetrics = (ushort)vmtx.VMetrics.Length;
		}

		#endregion
	}


	/// <summary>
	/// OpenType フォントの vhea (Horizontal Header) テーブル バージョン 1.1 を表します。
	/// </summary>
	internal class VHea11 : VHea10
	{
		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="VHea11"/> クラスのインスタンスを初期化します。
		/// </summary>
		public VHea11() : base(0x00011000, null)
		{
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.VHea11"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="vhea">元となる 'hhea' テーブル。</param>
		/// <exception cref="NotSupportedException"><paramref name="vhea"/> は未サポートテーブルバージョン。</exception>
		public VHea11(VHea vhea) : this(0x00011000, vhea)
		{
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.Table.VHea11"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="version">テーブルバージョン番号。</param>
		/// <param name="vhea">元となる 'vhea' テーブル。</param>
		/// <exception cref="NotSupportedException"><paramref name="vhea"/> は未サポートテーブルバージョン。</exception>
		protected VHea11(uint version, VHea vhea) : base(version, vhea)
		{
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// 垂直方向のタイポグラフィック ascender を取得または設定します。
		/// </summary>
		/// <value>垂直方向のタイポグラフィック ascender。</value>
		/// <remarks>通常は (head.unitsPerEm)/2 を設定。</remarks>
		public short VertTypoAscender
		{
			get { return Ascent; }
			set { Ascent = value; }
		}

		/// <summary>
		/// 垂直方向のタイポグラフィック descender を取得または設定します。
		/// </summary>
		/// <value>垂直方向のタイポグラフィック descender。</value>
		/// <remarks>通常は (head.unitsPerEm)/2 を設定。</remarks>
		public short VertTypoDescender
		{
			get { return Descent; }
			set { Descent = value; }
		}

		/// <summary>
		/// 垂直方向のタイポグラフィック gap を取得または設定します。
		/// </summary>
		/// <value>垂直方向のタイポグラフィック gap。</value>
		public short VertTypoLineGap
		{
			get { return LineGap; }
			set { LineGap = value; }
		}

		#endregion
	}
}