using System;
using System.IO;


namespace OfficeIVS.Font.Table
{

	/// <summary>
	/// OpenType フォントの vmtx (Vertical Metrics) テーブルを表します。
	/// </summary>
	[OpenTypeFontTable("vmtx")]
	internal sealed class VMtx : OpenTypeFontTable
	{
		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="VMtx"/> クラスのインスタンスを初期化します。
		/// </summary>
		public VMtx() : base(TableTags.vmtx)
		{
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// 各グリフの垂直フォントメトリックコレクションを取得または設定します。
		/// </summary>
		/// <value>グリフの垂直フォントメトリック。</value>
		/// <remarks>
		/// 要素数は 'vhea' テーブルの numOfLongVerMetrics 参照。<br/>
		/// 等幅フォントの場合には 1 つのみ。最後のエントリは後続の全てのグリフに適用。<br/>
		/// 本プロパティの要素数と <see cref="TopSideBearing"/> の要素数の加算結果が
		/// 'maxp' テーブルの numGlyphs と同じ値となる。
		/// </remarks>
		public LongVMetric[] VMetrics { get; set; }

		/// <summary>
		/// 各グリフの top sidebearing 値を取得または設定します。
		/// </summary>
		/// <value>グリフの top sidebearing 値。</value>
		/// <remarks>
		/// グリフのアドバンス高さは <see cref="VMetrics"/> の最終エントリを参照。<br/>
		/// 本プロパティの要素数と <see cref="TopSideBearing"/> の要素数の加算結果が
		/// 'maxp' テーブルの numGlyphs と同じ値となる。
		/// </remarks>
		public short[] TopSideBearing { get; set; }

		#endregion

		#region メソッド

		/// <summary>
		/// 指定番号のグリフの垂直メトリック情報を取得します。
		/// </summary>
		/// <param name="glyphIndex">グリフ番号。</param>
		/// <returns>垂直メトリック情報。</returns>
		/// <exception cref="InvalidOperationException"><see cref="VMetrics"/> が <c>null</c>。</exception>
		/// <exception cref="ArgumentOutOfRangeException"><paramref name="glyphIndex"/> が範囲外。</exception>
		public LongVMetric GetVMetric(ushort glyphIndex)
		{
			if (VMetrics == null)
			{
				throw new InvalidOperationException();
			}

			int numGlyphs = VMetrics.Length;

			if (TopSideBearing != null)
			{
				numGlyphs += TopSideBearing.Length;
			}

			if (glyphIndex >= numGlyphs)
			{
				throw new ArgumentOutOfRangeException("glyphIndex");	
			}

			LongVMetric result = new LongVMetric();

			if (glyphIndex < VMetrics.Length)
			{
				result.AdvanceHeight = VMetrics[glyphIndex].AdvanceHeight;
				result.TopSideBearing = VMetrics[glyphIndex].TopSideBearing;
			}
			else
			{
				result.AdvanceHeight = VMetrics[VMetrics.Length-1].AdvanceHeight;
				result.TopSideBearing = TopSideBearing[glyphIndex-VMetrics.Length];
			}

			return result;
		}

		/// <summary>
		/// フォントから 'vmtx' テーブルを読み込みます。
		/// </summary>
		/// <param name="font">読み込み元となるフォント。</param>
		/// <returns><see cref="T:OfficeIVS.Font.Table.VMtx"/> 型のインスタンス。テーブルが存在しない場合には <c>null</c>。</returns>
		[OpenTypeFontTableFactoryMethod]
		public static OpenTypeFontTable Load(OpenTypeFont font)
		{
			VMtx table = null;

			ITableRecord tableRecord = font[TableTags.vmtx];

			if (tableRecord != null)
			{
				VHea10 vhea = font.GetTable(TableTags.vhea) as VHea10;
				MaxP05 maxp = font.GetTable(TableTags.maxp) as MaxP05;

				if (vhea != null && maxp != null)
				{
					using (Stream stream = tableRecord.GetTableStream())
					using (BigEndianBinaryReader reader = new BigEndianBinaryReader(stream))
					{
						table = new VMtx();
						table.VMetrics = new LongVMetric[vhea.NumOfLongVerMetrics];
						table.TopSideBearing = new short[Math.Max(0, maxp.NumGlyphs - vhea.NumOfLongVerMetrics)];
						table.LoadEntries(reader, tableRecord.Length);
						table.TableRecord = tableRecord;
					}
				}
			}

			return table;
		}

		/// <summary>
		/// テーブル依存のエントリデータを読み込みます。
		/// </summary>
		/// <param name="input">テーブルデータストリーム。</param>
		/// <param name="length">テーブル長（バイト単位）。</param>
		protected override void LoadEntries(BigEndianBinaryReader input, long length)
		{
			// VMetrics
			for (int i=0; i<VMetrics.Length; i++)
			{
				VMetrics[i] =
					new LongVMetric
					{
						AdvanceHeight = input.ReadUInt16(),
						TopSideBearing = input.ReadInt16(),
					};
			}

			// TopSideBearing
			for (int i=0; i<TopSideBearing.Length; i++)
			{
				TopSideBearing[i] = input.ReadInt16();
			}
		}

		/// <summary>
		/// テーブル依存のエントリデータを書き込みます。
		/// </summary>
		/// <param name="output">書き込み先のストリーム。</param>
		protected override void WriteEntries(BigEndianBinaryWriter output)
		{
			foreach (LongVMetric vMetric in VMetrics)
			{
				output.Write(vMetric.AdvanceHeight);
				output.Write(vMetric.TopSideBearing);
			}

			if (TopSideBearing != null)
			{
				foreach (short lsb in TopSideBearing)
				{
					output.Write(lsb);
				}
			}
		}

		#endregion
	}

	/// <summary>
	/// グリフの垂直メトリック情報を表します。
	/// </summary>
	internal sealed class LongVMetric
	{
		#region プロパティ

		/// <summary>
		/// グリフのアドバンス高さを取得または設定します。
		/// </summary>
		/// <value>グリフのアドバンス高さ。</value>
		public ushort AdvanceHeight { get; set; }

		/// <summary>
		/// グリフの top sidebearing 値を取得または設定します。
		/// </summary>
		/// <value>グリフの top sidebearing 値。</value>
		public short TopSideBearing { get; set; }

		#endregion
	}
}