using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.IO.MemoryMappedFiles;

using OfficeIVS.Font.Table;


namespace OfficeIVS.Font
{

	/// <summary>
	/// OpenType フォント Offset テーブルの TableRecord を表します。
	/// </summary>
	internal interface ITableRecord : IComparable<ITableRecord>
	{
		#region プロパティ

		/// <summary>
		/// テーブル識別子を取得します。
		/// </summary>
		/// <value>テーブル識別子。</value>
		Tag Tag { get; }

		/// <summary>
		/// テーブルのチェックサム値を取得します。
		/// </summary>
		/// <value>テーブルのチェックサム値。</value>
		uint CheckSum { get; }

		/// <summary>
		/// テーブル長を取得します。
		/// </summary>
		/// <value>テーブル長（バイト単位）。</value>
		uint Length { get; }

		#endregion

		#region メソッド

		/// <summary>
		/// フォントファイルのテーブルストリームを返します。
		/// </summary>
		/// <returns>フォントファイルのテーブルストリーム。</returns>
		Stream GetTableStream();

		/// <summary>
		/// フォントファイルテーブルの指定部分ストリームを返します。
		/// </summary>
		/// <param name="offset">テーブル先頭からのオフセット量（バイト単位）。</param>
		/// <param name="length">ストリームサイズ（バイト単位）。</param>
		/// <returns>フォントファイルのテーブルの部分ストリーム。</returns>
		Stream GetTableStream(int offset, int length);

		#endregion
	}


	/// <summary>
	/// OpenType フォント Offset テーブルの TableRecord を表します。
	/// </summary>
	internal abstract class TableRecord : ITableRecord
	{
		#region フィールド

		/// <summary>
		/// テーブル優先度情報。
		/// </summary>
		/// <remarks>キー:テーブル識別子、値：優先度（小さい値ほど優先度が高い）</remarks>
		private static readonly Dictionary<uint, int> TablePriorities;

		/// <summary>
		/// テーブルの識別子。
		/// </summary>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private readonly Tag _tag;

		#endregion

		#region コンストラクタ

		/// <summary>
		/// クラスを初期化します。
		/// </summary>
		static TableRecord()
		{
			Tag[] tableTags = new[]
			{
				TableTags.head, TableTags.hhea, TableTags.maxp, TableTags.OS2,
				TableTags.hmtx, TableTags.LTSH, TableTags.VDMX, TableTags.hdmx,
				TableTags.cmap, TableTags.fpgm, TableTags.prep, TableTags.cvt,
				TableTags.loca, TableTags.glyf, TableTags.kern, TableTags.name,
				TableTags.post, TableTags.CFF,  TableTags.gasp, TableTags.PCLT,
				TableTags.DSIG,
			};

			TablePriorities = new Dictionary<uint, int>();

			int priority = 1;

			foreach (Tag tableTag in tableTags)
			{
				TablePriorities[tableTag.Value] = priority++;
			}
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.TableRecord"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="tableTag">テーブル識別子。</param>
		protected TableRecord(Tag tableTag)
		{
			_tag = tableTag;
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// テーブル識別子を取得します。
		/// </summary>
		/// <value>テーブル識別子。</value>
		public Tag Tag
		{
			get { return _tag; }
		}

		/// <summary>
		/// テーブルのチェックサム値を取得します。
		/// </summary>
		/// <value>テーブルのチェックサム値。</value>
		public abstract uint CheckSum { get; }

		/// <summary>
		/// テーブル長を取得します。
		/// </summary>
		/// <value>テーブル長（バイト単位）。</value>
		public abstract uint Length { get; }

		#endregion

		#region メソッド

		/// <summary>
		/// フォントファイルのテーブルストリームを返します。
		/// </summary>
		/// <returns>フォントファイルのテーブルストリーム。</returns>
		public abstract Stream GetTableStream();

		/// <summary>
		/// フォントファイルテーブルの指定部分ストリームを返します。
		/// </summary>
		/// <param name="offset">テーブル先頭からのオフセット量（バイト単位）。</param>
		/// <param name="length">ストリームサイズ（バイト単位）。</param>
		/// <returns>フォントファイルのテーブルの部分ストリーム。</returns>
		public abstract Stream GetTableStream(int offset, int length);

		/// <summary>
		/// 現在のオブジェクトを同じ型の別のオブジェクトと比較します。
		/// </summary>
		/// <returns>
		/// 比較対象オブジェクトの相対順序を示す値。<br/>
		/// 0 より小さい : このオブジェクトが <paramref name="other"/> パラメーターより小さいことを意味します。<br/>
		/// 0            : このオブジェクトが <paramref name="other"/> と等しいことを意味します。<br/>
		/// 0 より大きい : このオブジェクトが <paramref name="other"/> よりも大きいことを意味します。<br/>
		/// </returns>
		/// <param name="other">このオブジェクトと比較するオブジェクト。</param>
		public int CompareTo(ITableRecord other)
		{
			if (ReferenceEquals(this, other))
			{
				return 0;
			}

			if (ReferenceEquals(other, null))
			{
				return 1;
			}

			return GetTablePriority(Tag) - GetTablePriority(other.Tag);
		}

		/// <summary>
		/// テーブルの優先度を取得します。
		/// </summary>
		/// <param name="tag">テーブル識別子。</param>
		/// <returns>テーブル優先度（値が小さいほど優先度が高い）。</returns>
		private static int GetTablePriority(Tag tag)
		{
			int priority;

			if (!TablePriorities.TryGetValue(tag.Value, out priority))
			{
				priority = Int32.MaxValue;
			}

			return priority;
		}

		#endregion
	}


	/// <summary>
	/// OpenType フォント Offset テーブルの編集可能 TableRecord を表します。
	/// </summary>
	internal sealed class EditableTableRecord : TableRecord
	{
		#region フィールド

		/// <summary>
		/// テーブルチェックサム値が計算済みかどうか。
		/// </summary>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private bool _checkSumCalculated;
	
		/// <summary>
		/// テーブルチェックサム値。
		/// </summary>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private uint _checkSum;

		/// <summary>
		/// テーブルデータ。
		/// </summary>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private byte[] _tableData;

		#endregion

		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.EditableTableRecord"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="tableTag">テーブル識別子。</param>
		public EditableTableRecord(Tag tableTag)
			: base(tableTag)
		{
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.EditableTableRecord"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="tableTag">テーブル識別子。</param>
		/// <param name="tableData">テーブルデータ。</param>
		public EditableTableRecord(Tag tableTag, byte[] tableData)
			: this(tableTag)
		{
			_tableData = tableData;
		}

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.EditableTableRecord"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="table">テーブル。</param>
		/// <param name="minTableSize">最小テーブルサイズ。</param>
		public EditableTableRecord(OpenTypeFontTable table, long minTableSize=0)
			: this(table.Tag)
		{
			using (MemoryStream tableData = new MemoryStream())
			{
				table.WriteTo(tableData);

				if (minTableSize > 0)
				{
					if (tableData.Length < minTableSize)
					{
						tableData.SetLength(minTableSize);
					}
				}

				tableData.Flush();

				SetTableData(tableData.ToArray());
			}
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// テーブルのチェックサム値を取得します。
		/// </summary>
		/// <value>テーブルのチェックサム値。</value>
		public override uint CheckSum
		{
			[DebuggerStepThrough]
			get
			{
				if (!_checkSumCalculated)
				{
					UpdateTableCheckSum();
				}
				return _checkSum;
			}
		}

		/// <summary>
		/// テーブル長を取得します。
		/// </summary>
		/// <value>テーブル長（バイト単位）。</value>
		public override uint Length
		{
			[DebuggerStepThrough]
			get { return (uint)_tableData.Length; }
		}

		#endregion

		#region メソッド
	
		/// <summary>
		/// フォントファイルのテーブルストリームを返します。
		/// </summary>
		/// <returns>フォントファイルのテーブルストリーム。</returns>
		public override Stream GetTableStream()
		{
			return new MemoryStream(_tableData, false);
		}

		/// <summary>
		/// フォントファイルテーブルの指定部分ストリームを返します。
		/// </summary>
		/// <param name="offset">テーブル先頭からのオフセット量（バイト単位）。</param>
		/// <param name="length">ストリームサイズ（バイト単位）。</param>
		/// <returns>フォントファイルのテーブルの部分ストリーム。</returns>
		public override Stream GetTableStream(int offset, int length)
		{
			return new MemoryStream(_tableData, offset, length, false);
		}

		/// <summary>
		/// フォントファイルの書き込み可能なテーブルストリームを返します。
		/// </summary>
		/// <param name="resetCheckSum">編集終了後にチェックサムの再計算を行うかどうか。</param>
		/// <returns>書き込み可能なテーブルストリーム。</returns>
		public Stream GetWriteableTableStream(bool resetCheckSum=true)
		{
			if (resetCheckSum)
			{
				_checkSumCalculated = false;
			}
			return new MemoryStream(_tableData, true);
		}

		/// <summary>
		/// このテーブルレコードに対応するテーブルデータを設定します。
		/// </summary>
		/// <param name="tableData">テーブルデータ。</param>
		/// <remarks>テーブルデータの 4 byte アライメントは不要。純粋なテーブルデータのみを設定すること。</remarks>
		/// <exception cref="ArgumentNullException"><paramref name="tableData"/> が <c>null</c>。</exception>
		public void SetTableData(byte[] tableData)
		{
			if (tableData == null)
			{
				throw new ArgumentNullException("tableData");
			}

			_checkSumCalculated = false;
			_tableData = tableData;
		}
		
		/// <summary>
		/// テーブルチェックサム値を更新します。
		/// </summary>
		private void UpdateTableCheckSum()
		{
			uint checkSum = 0;

			if (_tableData != null)
			{
				using (MemoryStream ms = new MemoryStream(_tableData, false))
				using (BigEndianBinaryReader r = new BigEndianBinaryReader(ms))
				{
					// 対象データを符号なし 32bit 値として読み込み、順次加算
					for (int i=0; i<(_tableData.Length >> 2); i++)
					{
						checkSum = unchecked(checkSum + r.ReadUInt32());
					}

					// 32bit 値として読み込めない残りデータの計算
					int remainder = _tableData.Length & 3;

					if (remainder > 0)
					{
						byte[] buffer = new byte[4];
						ms.Read(buffer, 0, remainder);
						Array.Reverse(buffer);		// データは Big Endian なので反転
						checkSum = unchecked(checkSum + BitConverter.ToUInt32(buffer, 0));
					}
				}

				_checkSumCalculated = true;
			}

			_checkSum = checkSum;
		}

		#endregion
	}


	/// <summary>
	/// OpenType フォント Offset テーブルの読み取り専用 TableRecord を表します。
	/// </summary>
	internal sealed class ReadOnlyTableRecord : TableRecord
	{
		#region フィールド

		/// <summary>
		/// フォントファイル。
		/// </summary>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private readonly MemoryMappedFile _fontFile;

		/// <summary>
		/// テーブルチェックサム値。
		/// </summary>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private readonly uint _checkSum;

		/// <summary>
		/// テーブル位置（ファイル先頭からのバイト単位のオフセット量）。
		/// </summary>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private readonly uint _offset;

		/// <summary>
		/// テーブル長（バイト単位）。
		/// </summary>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private readonly uint _length;

		#endregion

		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="T:OfficeIVS.Font.ReadOnlyTableRecord"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="fontFile">フォントファイル。</param>
		/// <param name="tagValue">テーブル識別子。</param>
		/// <param name="checkSum">テーブルのチェックサム値。</param>
		/// <param name="offset">ファイル先頭からテーブルデータまでのオフセット量（バイト単位）。</param>
		/// <param name="length">テーブルサイズ（バイト単位）。</param>
		public ReadOnlyTableRecord(MemoryMappedFile fontFile, uint tagValue, uint checkSum, uint offset, uint length)
			: base(Tag.FromValue(tagValue))
		{
			_fontFile = fontFile;
			_checkSum = checkSum;
			_offset = offset;
			_length = length;
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// テーブルのチェックサム値を取得します。
		/// </summary>
		/// <value>テーブルのチェックサム値。</value>
		public override uint CheckSum
		{
			[DebuggerStepThrough]
			get { return _checkSum; }
		}

		/// <summary>
		/// テーブル長を取得します。
		/// </summary>
		/// <value>テーブル長（バイト単位）。</value>
		public override  uint Length
		{
			[DebuggerStepThrough]
			get { return _length; }
		}

		#endregion

		#region メソッド
	
		/// <summary>
		/// フォントファイルのテーブルストリームを返します。
		/// </summary>
		/// <returns>フォントファイルのテーブルストリーム。</returns>
		public override  Stream GetTableStream()
		{
			return _fontFile.CreateViewStream(_offset, Length, MemoryMappedFileAccess.Read);
		}

		/// <summary>
		/// フォントファイルテーブルの指定部分ストリームを返します。
		/// </summary>
		/// <param name="offset">テーブル先頭からのオフセット量（バイト単位）。</param>
		/// <param name="length">ストリームサイズ（バイト単位）。</param>
		/// <returns>フォントファイルのテーブルの部分ストリーム。</returns>
		public override  Stream GetTableStream(int offset, int length)
		{
			return _fontFile.CreateViewStream(_offset + offset, length, MemoryMappedFileAccess.Read);
		}

		#endregion
	}
}