﻿using System;
using System.Diagnostics;


namespace OfficeIVS.Font
{

	/// <summary>
	/// OpenType Format のタグ型を表します。
	/// </summary>
	[DebuggerDisplay("{Name}")]
	[DebuggerStepThrough]
	internal struct Tag : IEquatable<Tag>
	{
		#region フィールド

		/// <summary>
		/// タグ値。
		/// </summary>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private readonly uint _tagValue;

		#endregion

		#region コンストラクタ

		/// <summary>
		/// 指定したタグ値を表す新しい <see cref="Tag"/> 構造体を初期化します。
		/// </summary>
		/// <param name="tagValue">タグ値</param>
		private Tag(uint tagValue)
		{
			_tagValue = tagValue;
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// タグ名を取得します。
		/// </summary>
		/// <value>タグ名。</value>
		public string Name
		{
			get
			{
				return new string(
					new []
					{
						(char)((_tagValue >> 24) & 0x000000FF),
						(char)((_tagValue >> 16) & 0x000000FF),
						(char)((_tagValue >>  8) & 0x000000FF),
						(char)(_tagValue & 0x000000FF),
					}
				);
			}
		}

		/// <summary>
		/// タグ値を取得します。
		/// </summary>
		/// <value>タグ値。</value>
		public uint Value
		{
			get { return _tagValue; }
		}

		#endregion

		#region メソッド

		/// <summary>
		/// 指定した文字列を使用して新しい <see cref="Tag"/> 構造体を作成します。
		/// </summary>
		/// <param name="tagName">タグ文字列。</param>
		/// <returns>指定されたタグ文字列を表す <see cref="Tag"/> 構造体。</returns>
		/// <exception cref="ArgumentNullException"><paramref name="tagName"/> が null。</exception>
		/// <exception cref="ArgumentException">タグ名として使用できない文字が含まれている。</exception>
		public static Tag FromName(string tagName)
		{
			if (tagName == null)
			{
				throw new ArgumentNullException();
			}

			if (tagName.Length < 4)
			{
				tagName = tagName.PadRight(4);
			}
			
			uint tagValue = 0;

			for (int i=0; i<4; i++)
			{
				char ch = tagName[i];

				if (ch < 0x0020 || 0x007E < ch)
				{
					throw new ArgumentException(String.Format("文字 U+{0:X4} はタグ名に使用できません。", (int)ch), "tagName");
				}

				tagValue = (tagValue << 8) | ((uint)ch & 0x000000FF);
			}

			return FromValue(tagValue);
		}

		/// <summary>
		/// 指定したタグ値を使用して新しい <see cref="Tag"/> 構造体を作成します。
		/// </summary>
		/// <param name="tagValue">タグ値。</param>
		/// <returns>指定されたタグ値を表す <see cref="Tag"/> 構造体。</returns>
		public static Tag FromValue(uint tagValue)
		{
			return new Tag(tagValue);
		}

		/// <summary>
		/// 指定したオブジェクトが、現在のタグと等しいかどうかをテストします。
		/// </summary>
		/// <param name="obj">比較するオブジェクト。</param>
		/// <returns><paramref name="obj"/> が現在の <see cref="Tag"/> 構造体と等しい場合は <c>true</c>。それ以外の場合は <c>false</c>。</returns>
		public override bool Equals(object obj)
		{
			if (obj is Tag)
			{
				return Value == ((Tag)obj).Value;
			}

			return false;
		}

		/// <summary>
		/// 指定した <see cref="Tag"/> 構造体が、現在のタグと等しいかどうかをテストします。
		/// </summary>
		/// <param name="tag">比較する <see cref="Tag"/> 構造体。</param>
		/// <returns><paramref name="tag"/> が現在の <see cref="Tag"/> 構造体と等しい場合は <c>true</c>。それ以外の場合は <c>false</c>。</returns>
		public bool Equals(Tag tag)
		{
			return Value == tag.Value;
		}

		/// <summary>
		/// 現在の <see cref="Tag"/> 構造体のハッシュコードを取得します。
		/// </summary>
		/// <returns>現在の <see cref="Tag"/> 構造体のハッシュ コード。</returns>
		public override int GetHashCode()
		{
			return unchecked((int)_tagValue);
		}

		/// <summary>
		/// 現在の <see cref="Tag"/> 構造体のタグ名を返します。
		/// </summary>
		/// <returns>タグ名。</returns>
		public override string ToString()
		{
			return Name;
		}

		#endregion

		#region 演算子

		/// <summary>
		/// 2 つの <see cref="Tag"/> 構造体が等しいかどうかテストします。
		/// </summary>
		/// <param name="tag1">比較する最初の <see cref="Tag"/> 構造体。</param>
		/// <param name="tag2">比較する 2 番目の <see cref="Tag"/> 構造体。</param>
		/// <returns><paramref name="tag1"/> および <paramref name="tag2"/> が等しい場合は <c>true</c>。それ以外の場合は <c>false</c>。</returns>
		public static bool operator == (Tag tag1, Tag tag2)
		{
			return tag1.Value == tag2.Value;
		}

		/// <summary>
		/// 2 つの <see cref="Tag"/> 構造体が異なるかどうかテストします。
		/// </summary>
		/// <param name="tag1">比較する最初の <see cref="Tag"/> 構造体。</param>
		/// <param name="tag2">比較する 2 番目の <see cref="Tag"/> 構造体。</param>
		/// <returns><paramref name="tag1"/> および <paramref name="tag2"/> が等しくない場合は <c>true</c>。それ以外の場合は <c>false</c>。</returns>
		public static bool operator != (Tag tag1, Tag tag2)
		{
			return tag1.Value != tag2.Value;
		}

		#endregion
	}
}