using System;
using System.Diagnostics;
using System.Threading;
using System.Windows.Markup;
using System.Windows.Media;

using OfficeIVS.Font;


namespace OfficeIVS
{

	/// <summary>
	/// フォント情報を表します。
	/// </summary>
	[DebuggerDisplay("{FamilyName}")]
	public sealed class FontInfo
	{
		#region フィールド

		/// <summary>
		/// フォントのファイルパス。
		/// </summary>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private readonly Uri _uri;

		/// <summary>
		/// フォントファミリ名。
		/// </summary>
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		private readonly string _fontfamilyName;

		/// <summary>
		/// UI 表示用のフォントファミリ名。
		/// </summary>
		private readonly string _displayFontFamilyName;

		#endregion

		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="FontInfo"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="fontFamily">フォントファミリ。</param>
		/// <param name="fontUri">フォント URI。</param>
		/// <exception cref="ArgumentNullException"><paramref name="fontFamily"/>、もしくは <paramref name="fontUri"/> が null。</exception>
		internal FontInfo(FontFamily fontFamily, Uri fontUri)
		{
			if (fontFamily == null || fontUri == null)
			{
				throw new ArgumentNullException("fontFamily");
			}

			// フォントの物理ファイルパス
			_uri = fontUri;

			// フォントファミリ名
			//  : 以下の優先度の名称を取得する
			//      1. UI スレッドの言語
			//      2. 英語
			//      3. 最初に見つかったもの
			if (!fontFamily.FamilyNames.TryGetValue(XmlLanguage.GetLanguage(Thread.CurrentThread.CurrentUICulture.Name), out _displayFontFamilyName)
			 && !fontFamily.FamilyNames.TryGetValue(XmlLanguage.GetLanguage("en-US"), out _displayFontFamilyName))
			{
				foreach (var xmlLang in fontFamily.FamilyNames.Keys)
				{
					_displayFontFamilyName = fontFamily.FamilyNames[xmlLang];
					break;
				}
			}

			_fontfamilyName = fontFamily.ToString();
		}

		/// <summary>
		/// 新しい <see cref="FontInfo"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="font">フォント。</param>
		/// <exception cref="ArgumentNullException"><paramref name="font"/> が null。</exception>
		/// <exception cref="NotSupportedException">フォントファミリ名が取得できない。</exception>
		internal FontInfo(OpenTypeFont font)
		{
			if (font == null)
			{
				throw new ArgumentException("font");
			}

			// フォントの物理ファイルパス
			_uri = font.FontUri;

			// 表示用のフォントファミリ名
			//  : 以下の優先度の名称を取得する
			//      1. UI スレッドの言語
			//      2. 英語
			//      3. 最初に見つかったもの
			_displayFontFamilyName = font.GetFontFamilyName();

			if (_displayFontFamilyName == null)
			{
				throw new NotSupportedException();
			}

			// フォントファミリ名
			_fontfamilyName = font.GetFontFamilyName("en-US");

			if (_fontfamilyName == null)
			{
				_fontfamilyName = _displayFontFamilyName;
			}
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// フォントオブジェクト位置を取得します。
		/// </summary>
		/// <value>フォントのオブジェクト位置。</value>
		public Uri Uri
		{
			[DebuggerStepThrough]
			get { return _uri; }
		}

		/// <summary>
		/// UI 表示用のフォントファミリー名を取得します。
		/// </summary>
		/// <value>フォントファイミリー名。</value>
		public string FamilyName
		{
			[DebuggerStepThrough]
			get { return _displayFontFamilyName; }
		}

		/// <summary>
		/// フォントファミリー名（たぶん英語）を取得します。
		/// </summary>
		/// <value>フォントファイミリー名。</value>
		public string FontFamilyName
		{
			[DebuggerStepThrough]
			get { return _fontfamilyName; }
		}

		#endregion
	}
}