﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;


namespace OfficeIVS
{

	/// <summary>
	/// Office IVS 用外字フォント管理機能を提供します。
	/// </summary>
	public static class FontManager
	{
		#region フィールド

		/// <summary>
		/// IVD シーケンスコレクション。
		/// </summary>
		/// <remarks>
		/// IVD で定義済みの異体字文字情報のコレクション。<br/>
		/// 　キー : 異体字セレクタ付き文字コード（上位 32bit: 基底文字、下位 32bit: 字形選択肢）。<br/>
		/// 　値　 : コレクション名、グリフ名。<br/>
		/// </remarks>
		private static readonly Dictionary<long, Tuple<string, string>> IvdSequences;

		/// <summary>
		/// IVS 対応フォントのフォント情報コレクション。
		/// </summary>
		private static readonly List<FontInfo> FontInfos;
	
		/// <summary>
		/// IVS フォントの文字マッピング情報コレクション。
		/// </summary>
		/// <remarks>
		///   キー : フォントファミリ名。<br/>
		/// 　値　 : フォントマッピング情報。<br/>
		/// </remarks>
		private static readonly Dictionary<string, CharMap> CharMapping;

		/// <summary>
		/// 初期化済みかどうかを表す値。
		/// </summary>
		private static bool _isInitialized = false;

		#endregion

		#region コンストラクタ

		/// <summary>
		/// <see cref="FontManager"/> クラスを初期化します。
		/// </summary>
		static FontManager()
		{
			// 静的フィールド初期化
			IvdSequences = LoadIvdSequences();
			FontInfos = new List<FontInfo>();
			CharMapping = new Dictionary<string, CharMap>();
		}

		#endregion

		#region パブリック メソッド

		/// <summary>
		/// 初期化処理を行います。
		/// </summary>
		public static void Initialize()
		{
			if (!_isInitialized)
			{
				// セッションデータ接続
				int refCount = GlobalSessionData.Connect();

				// IVS フォント情報の読み込み
				foreach (CharMap charMap in CharMap.LoadAllIvsFonts(refCount == 1))
				{
					FontInfo fontInfo = charMap.FontInfo;
					FontInfos.Add(fontInfo);
					CharMapping[fontInfo.FamilyName] = charMap;
				}
			}
			_isInitialized = true;
		}

		/// <summary>
		/// 終了処理を行います。
		/// </summary>
		public static void Uninitialize()
		{
			if (_isInitialized)
			{
				// フォントマップ情報解放
				foreach (CharMap charMap in CharMapping.Values)
				{
					charMap.Dispose();	
				}
				CharMapping.Clear();

				// セッションデータ切断
				GlobalSessionData.Disconnect();
			}
			_isInitialized = false;
		}

		/// <summary>
		/// 異体字セレクタ対応フォントのフォント情報の一覧を取得します。
		/// </summary>
		/// <returns>IVS 対応フォントのフォント情報コレクション。IVS 対応フォントが存在しない場合には空コレクション。</returns>
		public static IList<FontInfo> GetIvsFontInfos()
		{
			return FontInfos;
		}

		/// <summary>
		/// 割り当て可能な私用領域の範囲を取得します。
		/// </summary>
		/// <returns>割り当て可能な私用領域の範囲。</returns>
		public static IList<CodeRange> GetEudcRange()
		{
			return GlobalSessionData.EudcCodeRanges;
		}

		/// <summary>
		/// 指定フォントの異体字セレクタ付き文字と外字文字とのマッピング情報を取得します。
		/// </summary>
		/// <param name="fontFamilyName">フォント名。</param>
		/// <returns>異体字セレクタ付き文字と外字文字とのマッピング情報。未対応のフォントが指定された場合には null。</returns>
		public static ICharMap GetCharMap(string fontFamilyName)
		{
			CharMap charMap;
			CharMapping.TryGetValue(fontFamilyName, out charMap);
			return charMap;
		}
		
		/// <summary>
		/// 指定フォント内の指定文字を基底文字とする異体字の一覧を取得します。
		/// </summary>
		/// <param name="fontName">フォント名。</param>
		/// <param name="baseChar">基底文字。</param>
		/// <returns>検索条件に一致した異体字情報。一致する文字が存在しない場合には、空コレクション。</returns>
		public static IList<IvCharInfo> FindIvChar(string fontName, string baseChar)
		{
			IList<IvCharInfo> result = new List<IvCharInfo>();

			CharMap charMap;

			if (CharMapping.TryGetValue(fontName, out charMap))
			{
				int bc = CodePoint.ToCodePoint(baseChar);

				var ivChars = charMap.FindIvChar(bc);

				foreach (int vs in ivChars.Keys)
				{
					IvCharInfo charInfo = new IvCharInfo
					{
						Code = CodePoint.FromCodePoint(bc, vs),
						GlyphIndex = ivChars[vs],
					};

					Tuple<string, string> ivdInfo;
					
					if (IvdSequences.TryGetValue(CodePoint.ToIvsCode(bc, vs), out ivdInfo))
					{
						charInfo.IvdCollectionName = ivdInfo.Item1;
						charInfo.GlyphName = ivdInfo.Item2;
					}

					result.Add(charInfo);
				}
			}

			return result;
		}

		#endregion

		#region プライベート メソッド

		/// <summary>
		/// IVD_Sequences.txt の内容で IvdSequences フィールドを初期化します。
		/// </summary>
		/// <remarks>
		///  IVD_Sequences.txt について<br/>
		///    IVD により定義済みのグリフ情報が定義されているテキストファイル。<br/>
		///    The Unicode Ideographic Variation Database (IVD) から提供されている。<br/>
		///
		///  参考 URL<br/>
		///    http://unicode.org/ivd/<br/>
		///
		/// ファイル書式<br/>
		///  行書式<br/>
		///    (上位サロゲート 下位サロゲート); (コレクション名); (グリフ名)<br/>
		///  補足<br/>
		///    先頭文字が '#' の場合はコメント行。<br/>
		///    上位サロゲート、下位サロゲートは 16 進数値（プリフィックスなし）。<br/>
		/// </remarks>
		private static Dictionary<long, Tuple<string, string>> LoadIvdSequences()
		{
			Dictionary<long, Tuple<string, string>> ivdSequences = new Dictionary<long, Tuple<string, string>>();

			using (StringReader reader = new StringReader(Properties.Resources.IVD_Sequences))
			{
				// コレクション名のキャッシュ
				//  : コレクション名の種類は少ないため（現状 Adobe-Japan-1、Hanyou-Denshi の 2 種類）、
				//    同じ文字列インスタンスを使い回すことで省メモリ化する
				Dictionary<string, string> collectionNames = new Dictionary<string, string>();

				// 行単位での解析処理
				string line;

				while ((line = reader.ReadLine()) != null)
				{
					// 行の書式
					// (上位サロゲート 下位サロゲート); (コレクション名); (グリフ名)

					line = line.Trim();

					// 空行は無視
					if (string.IsNullOrWhiteSpace(line))
					{
						continue;
					}

					// コメント行は無視
					if (line[0] == '#')
					{
						continue;
					}

					// ';' で分割
					string[] items = line.Split(new []{"; "}, StringSplitOptions.None);

					if (items.Length != 3)
					{
						continue;
					}

					// コードポイント取得
					long ivs;
					{
						string[] codePoint = items[0].Split(' ');

						if (codePoint.Length != 2)
						{
							continue;
						}

						ivs = CodePoint.ToIvsCode(Int32.Parse(codePoint[0], NumberStyles.HexNumber),
						                          Int32.Parse(codePoint[1], NumberStyles.HexNumber));
					}

					// コレクション名をキャッシュから取得
					string collectionName = items[1];

					if (collectionNames.ContainsKey(collectionName))
					{
						collectionName = items[1];
					}
					else
					{
						collectionNames[collectionName] = collectionName;
					}

					// 登録
					ivdSequences[ivs] = new Tuple<string, string>(collectionName, items[2]);
				}
			}

			return ivdSequences;
		}

		#endregion
	}
}
