﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Management;
using System.Security.Cryptography;
using System.Text;


namespace OfficeIVS
{


	/// <summary>
	/// 本 DLL がロードされている間有効なデータを表します。
	/// </summary>
	internal static class GlobalSessionData
	{
		#region フィールド

		/// <summary>
		/// セッションファイル名。
		/// </summary>
		private static readonly string SessionFileName;

		/// <summary>
		/// 割り当て可能な私用領域のコード範囲。
		/// </summary>
		private static IList<CodeRange> CodeRanges;

		#endregion

		#region コンストラクタ

		/// <summary>
		/// <see cref="GlobalSessionData"/> クラスを初期化します。
		/// </summary>
		static GlobalSessionData()
		{
			// セッションファイル名の取得
			SessionFileName = GetSessionFileName();
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// 外字割り当て可能な PUA 内の範囲を取得します。
		/// </summary>
		/// <returns>外字割り当て可能な PUA 内の範囲コレクション。</returns>
		public static IList<CodeRange> EudcCodeRanges
		{
			[DebuggerStepThrough]
			get { return CodeRanges; }
		}

		#endregion

		#region メソッド

		/// <summary>
		/// グローバルセッションデータに接続します。
		/// </summary>
		/// <remarks>参照数。</remarks>
		public static int Connect()
		{
			int refCount = 1;

			// セッションファイル読み込み
			using (new SessionFileLocker())
			{
				// セッションファイル読み込み
				if (File.Exists(SessionFileName))
				{
					// セッションファイル読み込み
					using (Stream stream = File.Open(SessionFileName, FileMode.Open, FileAccess.ReadWrite))
					using (BinaryReader reader = new BinaryReader(stream))
					using (BinaryWriter writer = new BinaryWriter(stream))
					{
						// 参照カウント更新
						{
							refCount = reader.ReadInt32() + 1;
							stream.Seek(0, SeekOrigin.Begin);
							writer.Write(refCount);
						}
					
						// コードレンジ
						{
							int count = reader.ReadInt32();

							CodeRanges = new List<CodeRange>(count);

							for (int i=0; i<count; i++)
							{
								CodeRanges.Add(new CodeRange(reader.ReadInt32(), reader.ReadInt32()));
							}
						}

					}
				}

				// ファイルが存在しなければ新規作成
				else
				{
					using (Stream stream = File.Open(SessionFileName, FileMode.Create, FileAccess.ReadWrite))
					using (BinaryWriter writer = new BinaryWriter(stream))
					{
						// 参照カウント
						{
							writer.Write(refCount);
						}

						// コードレンジ
						{
							CodeRanges = LoadCodeRanges();

							// 要素数
							writer.Write(CodeRanges.Count);

							// 範囲
							foreach (CodeRange codeRange in CodeRanges)
							{
								writer.Write(codeRange.StartCharCode);
								writer.Write(codeRange.EndCharCode);
							}
						}
					}
				}
			}

			return refCount;
		}

		/// <summary>
		/// グローバルセッションデータとの接続を解除します。
		/// </summary>
		public static void Disconnect()
		{
			if (!File.Exists(SessionFileName))
			{
				return;
			}

			using (new SessionFileLocker())
			{
				int refCount;

				// 参照カウント更新
				using (Stream stream = File.Open(SessionFileName, FileMode.Open, FileAccess.ReadWrite))
				using (BinaryReader reader = new BinaryReader(stream))
				using (BinaryWriter writer = new BinaryWriter(stream))
				{
					refCount = reader.ReadInt32() - 1;
					stream.Seek(0, SeekOrigin.Begin);
					writer.Write(refCount);
				}

				// 参照カウントが 0 になったら、セッションファイルを削除
				if (refCount <= 0)
				{
					File.Delete(SessionFileName);
				}
			}
		}

		/// <summary>
		/// セッションファイル名を取得します。
		/// </summary>
		/// <returns>セッションファイル名（フルパス）。</returns>
		/// <remarks>セッションファイル名はログイン日時を SHA-1 でハッシュ化した文字列。</remarks>
		private static string GetSessionFileName()
		{
			StringBuilder fileName = new StringBuilder();

			// ログイン時刻からファイル名を作成
			using (SHA1Managed sh1 = new SHA1Managed())
			{
				foreach (byte value in sh1.ComputeHash(Encoding.Unicode.GetBytes(GetCurrentUserLoginTime())))
				{
					fileName.Append(value.ToString("X2"));
				}
			}

			// 拡張子
			fileName.Append(".dat");

			return Path.Combine(Path.GetTempPath(), fileName.ToString());
		}

		/// <summary>
		/// 現在のユーザーのログイン時刻を取得します。
		/// </summary>
		/// <returns>現在のユーザーのログイン時刻（yyyyMMddHHmmss 形式）。</returns>
		private static string GetCurrentUserLoginTime()
		{
			try
			{
				ManagementScope scope = new ManagementScope(@"\root\cimv2");

				scope.Connect();

				ObjectQuery query = new ObjectQuery("Select StartTime from Win32_LogonSession Where LogonType=2");
				
				using (ManagementObjectSearcher searcher = new ManagementObjectSearcher(scope, query))
				using (ManagementObjectCollection collection = searcher.Get())
				{
					foreach (ManagementObject mo in collection)
					{
						return ((string)mo["StartTime"]).Substring(0, 14);
					}
				}
			}
			catch
			{
			}

			return DateTime.Now.ToString("yyyyMMddHHmmss");
		}

		/// <summary>
		/// 設定ファイルから外字登録可能なコード範囲を読み込みます。
		/// </summary>
		/// <returns>外字登録可能なコード範囲の一覧。</returns>
		/// <remarks>
		///  設定ファイル<br/>
		///    $LocalApplicationData$\EAST\Office AddIn\IVS\CodeRanges.txt<br/>
		/// 
		///  ファイル書式<br/>
		///    (下限値)-(上限値)[,(下限値)-(上限値)]<br/>
		///
		///  備考<br/>
		///    下限値、上限値はプリフィックスなしの 16 進数値。<br/>
		/// </remarks>
		private static IList<CodeRange> LoadCodeRanges()
		{
			List<CodeRange> codeRanges = new List<CodeRange>();

			string codeRangesFilePath = Path.Combine(
				Environment.GetFolderPath(Environment.SpecialFolder.LocalApplicationData),
				@"EAST\Office AddIn\IVS\CodeRanges.txt"
			);

			if (File.Exists(codeRangesFilePath))
			{
				try
				{
					// 設定ファイルから読み込み
					using (StreamReader reader = new StreamReader(codeRangesFilePath, true))
					{
						string line = reader.ReadLine();

						if (!String.IsNullOrEmpty(line))
						{
							foreach (string range in line.Split(','))
							{
								string[] code = range.Split('-');

								if (code.Length == 1)
								{
									int l = Math.Max(0xE000, Math.Min(Int32.Parse(code[0].Trim(), NumberStyles.HexNumber), 0xF8FF));
									codeRanges.Add(new CodeRange(l, l));
								}
								else if (code.Length == 2)
								{
									int l = Math.Max(0xE000, Math.Min(Int32.Parse(code[0].Trim(), NumberStyles.HexNumber), 0xF8FF));
									int h = Math.Max(0xE000, Math.Min(Int32.Parse(code[1].Trim(), NumberStyles.HexNumber), 0xF8FF));
									codeRanges.Add(new CodeRange(l, h));
								}
							}
						}
					}

					// 開始コードで並び替え
					codeRanges.Sort((v1, v2) => v1.StartCharCode - v2.StartCharCode);

					// 重複部分を削除
					for (int i=0; i<codeRanges.Count-1; i++)
					{
						if (codeRanges[i] == null)
						{
							continue;
						}

						for (int j=i+1; j<codeRanges.Count-1; j++)
						{
							if (codeRanges[i].EndCharCode <= codeRanges[j].StartCharCode)
							{
								codeRanges[i].StartCharCode = 0;
								codeRanges[i].EndCharCode = codeRanges[j].EndCharCode;
								codeRanges[j] = null;
							}
						}
					}
					codeRanges.RemoveAll(v => v == null);
				}
				catch
				{
					codeRanges.Clear();
				}
			}

			// 初期値設定
			if (codeRanges.Count == 0)
			{
				codeRanges.Add(new CodeRange(0xF100, 0xF8FF));
			}

			return codeRanges;
		}

		#endregion

		#region ヘルパークラス

		/// <summary>
		/// セッションファイルへのアクセスを同期するための補助機能を提供します。
		/// </summary>
		private sealed class SessionFileLocker : SyncAccess
		{
			#region コンストラクタ

			/// <summary>
			/// 新しい <see cref="SessionFileLocker"/> クラスのインスタンスを初期化します。
			/// </summary>
			public SessionFileLocker() : base(@"Office.AddIn.Ivs.GlobalSession")
			{
			}

			#endregion
		}

		#endregion
	}
}
