﻿using System;
using System.Diagnostics;
using System.IO;


namespace OfficeIVS
{

	/// <summary>
	/// グリフ情報を表します。
	/// </summary>
	internal sealed class GlyphInfo
	{
		#region コンストラクタ

		/// <summary>
		/// 新しい <see cref="GlyphInfo"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="bc">基底文字の Unicode コードポイント。</param>
		/// <param name="vs">字形選択肢の Unicode コードポイント。</param>
		/// <param name="oldGlyphIndex">元フォント内でのグリフ番号。</param>
		public GlyphInfo(int bc, int vs, ushort oldGlyphIndex)
		{
			GlyphIndex = 0;
			OldGlyphIndex = oldGlyphIndex;
			BaseChar = bc;
			VarSelector = vs;
			Eudc = -1;
			Count = 0;
			LastUsedTime = DateTime.MinValue;
		}

		/// <summary>
		/// 新しい <see cref="GlyphInfo"/> クラスのインスタンスを初期化します。
		/// </summary>
		/// <param name="reader">読み込み元のストリーム。</param>
		public GlyphInfo(BinaryReader reader)
		{
			Read(reader);
		}

		#endregion

		#region プロパティ

		/// <summary>
		/// フォント内でのグリフ番号を取得または設定します。
		/// </summary>
		/// <value>グリフ番号。</value>
		public ushort GlyphIndex { get; set; }

		/// <summary>
		/// 元フォント内でのグリフ番号を取得または設定します。
		/// </summary>
		/// <value>元フォント内でのグリフ番号。</value>
		public ushort OldGlyphIndex { get; set; }

		/// <summary>
		/// 異体字セレクタ付き文字を文字列形式で取得します。
		/// </summary>
		/// <value>異体字セレクタ付き文字。</value>
		public string Ivs
		{
			[DebuggerStepThrough]
			get { return CodePoint.FromCodePoint(BaseChar, VarSelector); }
		}
	
		/// <summary>
		/// 異体字セレクタ付き文字をコード形式で取得します。
		/// </summary>
		/// <value> 異体字セレクタ付き文字コード。</value>
		/// <remarks>
		/// 上位 32 bit : 基底文字の Unicode コードポイント。<br/>
		/// 下位 32 bit : 字形選択肢の Unicode コードポイント。<br/>
		/// </remarks>
		public long IvsCode
		{
			[DebuggerStepThrough]
			get
			{
				return ((long)BaseChar << 32) | unchecked((uint)VarSelector);
			}
		}

		/// <summary>
		/// 基底文字の Unicode コードポイントを取得または設定します。
		/// </summary>
		/// <value>基底文字の Unicode コードポイント。</value>
		public int BaseChar { get; private set; }

		/// <summary>
		/// 字形選択肢の Unicode コードポイントを取得または設定します。
		/// </summary>
		/// <value>字形選択肢の Unicode コードポイント。使用しない場合は -1。</value>
		public int VarSelector { get; private set; }

		/// <summary>
		/// 本グリフに割り当てられている外字コードを取得または設定します。
		/// </summary>
		/// <value>PUA 領域内の外字コード。外字割り当てされていない場合には -1。</value>
		public int Eudc { get; set; }

		/// <summary>
		/// 本グリフの使用回数を取得または設定します。
		/// </summary>
		/// <value>本グリフの使用回数。</value>
		public int Count { get; set; }

		/// <summary>
		/// 本グリフの最終使用日時を取得または設定します。
		/// </summary>
		/// <value>本グリフの最終使用日時（UTC）。</value>
		public DateTime LastUsedTime { get; set; }

		/// <summary>
		/// 本グリフの参照カウントを取得または設定します。
		/// </summary>
		/// <value>本グリフの参照カウント（使用中のプロセス数）。</value>
		public int ReferenceCount { get; set; }

		#endregion

		#region メソッド

		/// <summary>
		/// 指定ストリームからグリフ情報を読み込みます。
		/// </summary>
		/// <param name="reader">読み込み元ストリーム。</param>
		public void Read(BinaryReader reader)
		{
			BaseChar = reader.ReadInt32();
			VarSelector = reader.ReadInt32();
			Eudc = reader.ReadInt32();
			GlyphIndex = reader.ReadUInt16();
			OldGlyphIndex = reader.ReadUInt16();
			Count = reader.ReadInt32();
			LastUsedTime = DateTime.FromBinary(reader.ReadInt64());
			ReferenceCount = reader.ReadInt32();			
		}

		/// <summary>
		/// 指定ストリームにグリフ情報を出力します。
		/// </summary>
		/// <param name="writer">出力先ストリーム。</param>
		public void Write(BinaryWriter writer)
		{
			writer.Write(BaseChar);
			writer.Write(VarSelector);
			writer.Write(Eudc);
			writer.Write(GlyphIndex);
			writer.Write(OldGlyphIndex);
			writer.Write(Count);
			writer.Write(LastUsedTime.ToBinary());
			writer.Write(ReferenceCount);			
		}

		#endregion
	}
}
